/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 18/12/2006
 *
 * Last Updated :
 */

#include <iostream>
#include <string>
#include <vector>
#include <cstdlib>
#include <iomanip>
#include <fstream>

#include "core_header.hpp"
#include "solver.hpp"
#include "loss.hpp"

#include "petsc.h"
#include "petscmat.h"
#include "petscvec.h"


// solver and loss selectors
CSolver* SelectSolver(CInfo *info);
CLoss* SelectLoss(CInfo *info, CDataset *ds);


using namespace std;


int main(int argc, char **argv)
{
  ErrorCode    ec;             // BMRM error code

  CTimer       dstime;         // dataset loading time
  CTimer       totaltime;      // total time
  CTimer       solvertime;     // solver time
  CTimer       lossgradtime;   // gradient & loss time
  CTimer       presolvertime;  // update matrix and vectors
  CTimer       getsolutiontime;  //
  CTimer       misctime;       // printf's and ofstream's

  CInfo        info;           // learning & miscellaneous parameters
  CLoss        *lf;            // loss function
  CSolver      *solver;        // inner solver

  CMatVecArr   *w;             // weight (solution)
  CMatVecArr   *a;             // gradient
  
  int          iter = 0;               // iteration counter
  Real         Loss = 0.0;             // per dataset loss
  Real         xi = 0;                 // lower bound of empirical risk
  Real         approxObjVal = 0;       // lower bound of objective val
  Real         prevApproxObjVal = 0;   // previous approxObjVal
  Real         exactObjVal = 0;        // exact objective val
  Real         minExactObjVal = INFTY; // minimum exact objective value achieved so far
  Real         regVal = 0;             // value of the regularization term
  Real         prevRegVal = 0;         // previous regVal
  Real         epsilon = 0;            // duality gap (exact - approximated)
  Real         gamma = 0;              // lower bound on duality gap
  Real         w_dot_a = 0;            // tmp for <w,a>

  ofstream epsilonfile;
  ofstream xifile;
  ofstream gammafile;
  ofstream regvalfile;
  ofstream lossfile;
  ofstream approxobjfile;
  ofstream exactobjfile;

  int  dummy1 = 0;
  char **dummy2;
  PetscInitialize(&dummy1, &dummy2, (char*)0, "");

  // start working !
  totaltime.Start();
  ParseCmdLineArgTrain(argc, argv, &info);
  
  dstime.Start();
  CDataset ds(&info);
  dstime.Stop();
  
  lf     = SelectLoss(&info, &ds);  
  solver = SelectSolver(&info);
  a      = 0;
  w      = new CMatVecArr(info.dimW, info.k);

  // print parameters
  if(info.verbosity > 0)
    info.Print();
 
  if(info.convLog == 1) {
    epsilonfile.open("epsilon.dump");
    xifile.open("xi.dump");
    gammafile.open("gamma.dump");
    regvalfile.open("regval.dump");
    lossfile.open("loss.dump");
    approxobjfile.open("approxobj.dump");
    exactobjfile.open("exactobj.dump");
  }



  // optimization loop
  while(1) 
  {
    iter++;
    
    // column generation: compute loss and gradients    
    a = new CMatVecArr(info.dimW, info.k);
    Loss = 0;
    lossgradtime.Start();
    lf->ComputeLossGradient(w, &ds, Loss, a);
    lossgradtime.Stop();
        
    // update the solver with new gradients
    presolvertime.Start();
    VecDot(w->vec, a->vec, &w_dot_a);
    solver->Update(a, Loss - w_dot_a);
    presolvertime.Stop();

    // solve the problem
    solvertime.Start();    
    solver->Solve();
    solvertime.Stop();

    // construct new model parameter
    getsolutiontime.Start();
    prevApproxObjVal = approxObjVal;   // duality gap :=  Loss_{i+1} - \xi_{i} 
    prevRegVal = regVal;
    solver->GetSolution(w, xi, regVal, approxObjVal);
    getsolutiontime.Stop();

    // update some important values
    exactObjVal = info.C*Loss + prevRegVal;
    minExactObjVal = MIN(minExactObjVal, exactObjVal);

    // maintain statistics
    misctime.Start();
    epsilon = exactObjVal - approxObjVal;
    gamma = minExactObjVal - approxObjVal;
    
    if(info.convLog == 1) {
      lossfile      << Loss         << endl;
      xifile        << xi           << endl;
      regvalfile    << regVal       << endl;
      approxobjfile << approxObjVal << endl;
      exactobjfile  << exactObjVal  << endl;
      epsilonfile   << epsilon      << endl;
      gammafile     << gamma        << endl;
    }
    
    // dump statistics on screen
    if(info.verbosity > 0) {
      printf("[Iter %3d] epsilon: %.5e   gamma: %.5e   loss: %.5e   xi: %.5e   regularizer: %.5e\n",iter, epsilon, gamma, Loss, xi, regVal);
    }
    misctime.Stop();
    
    
    // stopping criterion        
    if(iter >= 2 && (gamma < info.gammaTol || epsilon < info.epsilonTol)) 
      break;
    if(iter >= info.maxIter)
    { 
      printf("\nWARNING: program exceeded maximum iteration (%d) !\n", info.maxIter);
      break;
    }
    
   
    
  }
  totaltime.Stop();

  // timing reports
  Real subtotal = lossgradtime.Total() + presolvertime.Total() + getsolutiontime.Total() + solvertime.Total();
  Real total    = totaltime.Total();

  printf("----------------------------------------------------------\n");
  printf("CPU-time in:\n");
  printf(" 1. gradient and loss : %.2f sec  (%.2f%%)\n",lossgradtime.Total(), lossgradtime.Total()*100.0/total);
  printf(" 2. prepare solver    : %.2f sec  (%.2f%%)\n",presolvertime.Total(), presolvertime.Total()*100.0/total);
  printf(" 2. get solution      : %.2f sec  (%.2f%%)\n",getsolutiontime.Total(), getsolutiontime.Total()*100.0/total);
  printf(" 3. solver            : %.2f sec  (%.2f%%)\n",solvertime.Total(), solvertime.Total()*100.0/total);
  printf("             Subtotal : %.2f sec  (%.2f%%)\n",subtotal, 100*subtotal/total);
  printf(" 4. maintaining stats : %.2f sec  (%.2f%%)\n",misctime.Total(), misctime.Total()*100.0/total);
  printf(" 5. loading dataset   : %.2f sec  (%.2f%%)\n",dstime.Total(), dstime.Total()*100.0/total);
  printf("                Total : %.2f sec\n",total);
  printf("----------------------------------------------------------\n");

  printf("%.2f [CPU-SECONDS] subtotal\n",subtotal);
  printf("%lf [OBJECTIVE VALUE] in standard SVM formulation\n", exactObjVal/info.lambda);
  printf("%lf [DUALITY GAP] upper bound on duality gap in standard SVM formulation\n", exactObjVal/info.lambda - approxObjVal/info.lambda);
  printf("Final objective value in BMRM formulation: %lf\n", exactObjVal);

  CLabel ybar;  // prediction
  CMatVecArr *f = new CMatVecArr(ds.nExample, info.k);  // the matrix Xw
  lf->Predict(w, &ds, &ybar, f);
  lf->EvalPerformance(w, &ds, &ybar, f);
  
  // dump model to model file
  ec = WriteModelFile(w, &info);
  if(ec == NOERROR && info.verbosity > 0)
    printf("Model file written : %s\n",info.modelFn.c_str());

  if(w)      delete w;
  if(f)      delete f;
  if(solver) delete solver;
  if(lf)     delete lf;


  PetscFinalize();
  return 0;
}





