/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 16/11/2006
 * Last Updated : 15/02/2007
 */

#ifndef _DATASET_CPP_
#define _DATASET_CPP_

#include <string>
#include <sstream>
#include <fstream>
#include <cassert>

#undef SEEK_SET
#undef SEEK_END
#undef SEEK_CUR
#include "mpi.h"

#include "errorcode.hpp"
#include "dataset.hpp"


/** Print the contents of CLabel into a file
 *
 *  \param filename [read] The name of the output file
 */
void CLabel::ToFile(string filename)
{
  ofstream ofp;
  // open file
  ofp.open(filename.c_str());
  if(!ofp.good()) {
    printf("ERROR: {CLabel} cannot open file <%s> !\n",filename.c_str());
    return;
  }

  // single label
  if(STL_vec.size() > 0) {
    
    for(size_t i=0; i<STL_vec.size(); i++)
      ofp << STL_vec[i] << endl;
  }
  // multiple label
  else {
    for(size_t i=0; i<STL_vvec.size(); i++) {
      for(size_t j=0; j<STL_vvec[i].size(); j++)
        ofp << STL_vvec[i][j] << " ";
      ofp << endl;
    }
  }
}


/** CDataset constructor
 */
CDataset::CDataset()
  :info(0),
   partID(0),
   nPart(1),
   datafileType(SPARSE_FILE),
   datasetType(SPARSE_DS),
   labelType(SINGLE_LABEL),
   row_ptr(0),
   idx(0),
   val(0),
   sample(0),
   dimY(1),
   dimX(0),
   effectiveDimX(-1),
   density(0),
   totalLabel(0),
   totalSample(0),
   totalExample(0),
   totalFeature(0),
   nSample(0),
   nExample(0),
   nFeature(0),
   startExample(0),
   startSample(0),
   exampleStat(new vector<int>),
   sampleStat(new vector<int>),
   labelStat(new vector<int>)
{}



/** CDataset constructor
 */
CDataset::CDataset(CInfo *_info)
  :info(_info),
   partID(0),
   nPart(1),
   datafileType(SPARSE_FILE),
   datasetType(SPARSE_DS),
   labelType(SINGLE_LABEL),
   row_ptr(0),
   idx(0),
   val(0),
   sample(0),
   dimY(1),
   dimX(0),
   effectiveDimX(-1),
   density(0),
   totalLabel(0),
   totalSample(0),
   totalExample(0),
   totalFeature(0),
   nSample(0),
   nExample(0),
   nFeature(0),
   startExample(0),
   startSample(0),
   exampleStat(new vector<int>),
   sampleStat(new vector<int>),
   labelStat(new vector<int>)

{
  /*
     NOTE:
     1. command line argument parser should:
        a. set CMiscParam::compMode
        b. set CMiscParam::datasetFn and CMiscParam::labelsetFn in distributed DS case
  */
  
  ErrorCode ec;
  sprintf(tag, "{proc:%02d:%s}",info->procID, info->hostname.c_str());

  // Do a pre-scan of dataset and collect some info.
  ScanDS();
  
  // no load balancing
  if(info->nNode == 1) {
    nSample  = totalSample;
    nExample = totalExample;
    nLabel   = totalLabel;    
    nFeature = totalFeature;
  }
  
  // Balance the load in PARALLEL CENTRALISED DATASET case
  if(info->nNode > 1 && info->compMode == PARALLEL_CENTRALISED_DS) {
    ec = BalanceLoad(); 
    if(ec != NOERROR) {
      printf("%s has error in BalanceLoad()\n",tag);
    }
  }
  
  // In PARALLEL DISTRIBUTED DATASET case, we do nothing at the moment...

  
  // create new communicator if some (extra) machines are to be excluded

  if(info->nUsedNode < info->nNode) 
  {
    int ierr;
    BMRM_Process_Ranks = new int[info->nUsedNode];
    for(int i=0; i<info->nUsedNode; i++) BMRM_Process_Ranks[i] = i;
    MPI_Comm_group(PETSC_COMM_WORLD, &PETSC_group_world);  
    ierr = MPI_Group_incl(PETSC_group_world, info->nUsedNode, BMRM_Process_Ranks, &BMRM_group_world);
    MPI_Comm_create(PETSC_COMM_WORLD, BMRM_group_world, &BMRM_COMM_WORLD);
    
    if(BMRM_COMM_WORLD == MPI_COMM_NULL)
      printf("\n%s is excluded in new communicator !\n", tag);
  }

  // start reading dataset (for some machines only)
  SOME_MACHINES(info->procID, info->nUsedNode)
  {
    if(datasetType == SPARSE_DS) {
      ec = ReadSparseDS();
      if(ec != NOERROR) {
        printf("%s has error in ReadSparseDS()\n",tag);
        exit(EXIT_FAILURE);
      }
    }
    else {
      ec = ReadDenseDS();  
      if(ec != NOERROR) {
        printf("%s has error in ReadDenseDS()\n",tag);
        exit(EXIT_FAILURE);
      }
    }
  }
  
}


/** CDataset destructor
 */
CDataset::~CDataset()
{
  //printf("%s destructing dataset...\n", tag);
  
//   if(datasetType == DENSE_DS)
//     MatDestroy(X);
    
  if(row_ptr)     delete [] row_ptr;
  if(idx)         delete [] idx;
  if(val)         delete [] val;
  if(sample)      delete [] sample;
  if(exampleStat) delete exampleStat;
  if(sampleStat)  delete sampleStat;
  if(labelStat)   delete labelStat;
}


/** Check if a string is blank.
 *
 *  \param line [read] The line to be checked if it is blank
 */
bool CDataset::IsBlankLine(string &line)
{
  int n = line.size();
  for(int i=0; i<n; i++)
    if(!isspace(line[i]))
      return false;
  return true;
}
 

/** Get rid of the trailing whitespaces.
 *
 *  \param line [r/w] The line to be trimmed.
 */
void CDataset::TrimWhitespaces(string &line)
{
  int cnt = line.size();
  while(isspace(line[cnt-1])) cnt--;
  line = line.substr(0, cnt);
}


/** Scan dataset once in order to validate dataset format and collect 
 *  information such as:
 *  1. label type (single or multiple)
 *  2. number of example
 *  3. total number of nonzero feature
 *  4. total number of sample 
 *  5. dimension of example
 *  6. 
 *
 *  Important Notes:
 *  1. Example will be appended by a new feature "1" to enable
 *     hyperplane with shift, i.e. <w,x> + b
 *
 *  Notes:
 *  1. Dataset can be in sparse (svmlight) or dense format.
 *  2. Each line corresponds to a data point.
 *  3. Labelset contains a list of label(s), ordered as in dataset.
 *  4. For multi-label case, labels of an example are delimited by whitespaces.
 *       
 */
ErrorCode CDataset::ScanDS()
{
  // local variables
  int      nLabelLine = 0;      // number of label line
  int      labelPerEx = 0;
  int      curQID = 0;          // current query id
  int      preQID = -1;         // previous query id
  int      featureCnt = 0;      // number of feature per example
  int      tmpfidx;             // tmp feature index
  Real     tmpfval;             // tmp feature value
  int      preDimX = -1;   // temporary
  int      prefidx = 0;
  int      line_cnt = 0;
  string   line = "";
  string   token = "";
  ifstream labelFp;
  ifstream dataFp;


  if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
    if(info->verbosity > 0) {
      printf("%s Scanning dataset...", tag);
      fflush(stdout);
    }
  }

  
  // scan label set
  if(info->programMode != PREDICTION_MODE) {
    labelFp.open(info->labelsetFn.c_str());
    
    if(!labelFp.good()) {
      printf("\n%s ERROR: {CDataset} can not open label file <%s> !\n", tag, info->labelsetFn.c_str());
      fflush(stdout);
      return ERROR;
    }
    
    for(nLabelLine=0; !labelFp.eof();) {
      getline(labelFp, line);
      TrimWhitespaces(line);
      if(IsBlankLine(line)) continue;  // blank line
      if(line[0] == '#') continue;     // comment
      istringstream iss(line);
      
      for(labelPerEx = 0; !iss.eof();) {
	iss >> token;
	if(sscanf(token.c_str(), "%lf", &tmpfval) == 1)
	  labelPerEx++;
	else {
	  printf("\n %s ERROR: {CDataset} label file contain invalid label at line %d !\n", tag, nLabelLine+1);       
	  printf("%s LINE : %s\n", tag, line.c_str());
	  printf("%s TOKEN : %s\n", tag, token.c_str());
	  fflush(stdout);
	  return ERROR;
	}
      }
      if(labelPerEx > 1) labelType = MULTI_LABEL;
      dimY = MAX(dimY, labelPerEx);
      labelStat->push_back(labelPerEx);
      totalLabel += labelPerEx;
      nLabelLine++;
    }

    if(nLabelLine <= 0) {
      printf("\n%s ERROR: {CDataset} label set is empty !\n", tag);
      fflush(stdout);
      return ERROR;   
    }
  }


  // scan data (pattern) set
  dataFp.open(info->datasetFn.c_str());
  
  if(!dataFp.good()) {
    printf("\n%s ERROR: {CDataset} can not open data file <%s> !\n", tag, info->datasetFn.c_str());
    return ERROR;
  }

  // Read first line to check whether data file is in SPARSE or DENSE format
  bool done = false;
  while(!done && !dataFp.eof()) 
  {
    getline(dataFp, line);    
    TrimWhitespaces(line);
    if(IsBlankLine(line)) continue;
    if(line[0] == '#') continue;

    istringstream iss(line);
    while(!iss.eof()) {
      iss >> token;
      if(token.find("id:") != string::npos) continue;
      if(token[0] == '#') break;
      if(sscanf(token.c_str(), "%d:%lf", &tmpfidx, &tmpfval) == 2) {
        datafileType = SPARSE_FILE;
        if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
          if(info->verbosity > 0) {
            printf("%s SPARSE data file detected...", tag);
            fflush(stdout);
          }
        }
        done = true;
        break;
      }
      else {        
        datafileType = DENSE_FILE;
        if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
          if(info->verbosity > 0) {
            printf("%s DENSE data file detected...", tag);
            fflush(stdout);
          }
        }
        done = true;
        break;
      }
    }
  }
  

  // do a pre-scan base on the DS file type detected
  line = "";
  dataFp.seekg(ios::beg);

  if(datafileType == SPARSE_FILE)
  {
    datasetType = SPARSE_DS;

    for(totalExample=0, line_cnt=0; !dataFp.eof(); line_cnt++) 
    {
      getline(dataFp, line);
      TrimWhitespaces(line);
      if(IsBlankLine(line)) continue;  // blank line
      if(line[0] == '#') continue;     // comment
      istringstream iss(line);
      
      // collect information
      // query id is always in front of features !
      size_t pos = line.find("qid:");
      if(pos != string::npos) {
        sscanf(&line[pos], "qid:%d ", &curQID);      
        iss >> token; // throw away this token
      }
      
      if(preQID != curQID) {
        totalSample++;
        sampleStat->push_back(0);
        preQID = curQID;
      }
      
      
      // feature pairs
      for(featureCnt=0; !iss.eof();)
      {    
        iss >> token;
        if(IsBlankLine(token)) break;
        if(token[0] == '#') break;
        
        if(sscanf(token.c_str(), "%d:%lf", &tmpfidx, &tmpfval) != 2) 
        {
          printf("\n%s ERROR: {CDataset} invalid feature pair! (LINE: %d)\n ", tag, line_cnt+1);
          printf("%s Line : '%s'\n", tag, line.c_str());
          printf("%s Feature pair : <%s>\n", tag, token.c_str());
          return ERROR;

        }

        // check fidx order
        if(prefidx >= tmpfidx) {
          printf("\n%s ERROR: {CDataset} invalid sparse feature index order at line %d !\n", tag, line_cnt+1);
          printf("%s LINE: %s\n", tag, line.c_str());
          return ERROR;
        }

        // update statistics
        featureCnt++;
        dimX = MAX(dimX, tmpfidx);
      }

      if(featureCnt <= 0) {
        printf("\n%s ERROR: {CDataset} example #%d has no non-zero feature values!\n", tag, totalExample+1);
        return ERROR;
      }
      
      // update statistics
      if(info->bias)
        featureCnt += 1;    // for shifted hyperplane
      exampleStat->push_back(featureCnt);
      sampleStat->at(totalSample-1) += 1;
      totalFeature += featureCnt;
      totalExample++;
    }
  }
  else if(datafileType == DENSE_FILE)
  {
    datasetType = DENSE_DS;
    
    for(totalExample=0, line_cnt=0; !dataFp.eof(); line_cnt++) 
    {
      getline(dataFp, line);
      TrimWhitespaces(line);
      if(IsBlankLine(line)) continue;  // blank line
      if(line[0] == '#') continue;     // comment
      istringstream iss(line);
            
      // collect information
      // query id is always in front of feature vector!
      size_t pos = line.find("qid:");
      if(pos != string::npos) {
        sscanf(&line[pos], "qid:%d ", &curQID);      
        iss >> token; // throw away this token
      }
      
      if(preQID != curQID) {
        totalSample++;
        sampleStat->push_back(0);
        preQID = curQID;
      }
            
      // feature values      
      for(featureCnt=0, dimX=0; !iss.eof();)
      {    
        iss >> token;
        if(IsBlankLine(token)) break;
        if(token[0] == '#') break;
        
        if(sscanf(token.c_str(), "%lf", &tmpfval) != 1) 
        {
          printf("\n%s ERROR: {CDataset} invalid feature value! (LINE: %d)\n ", tag, line_cnt+1);
          printf("%s Line : '%s'\n", tag, line.c_str());
          printf("%s Feature value : <%s>\n", tag, token.c_str());
          return ERROR;
        }
        
        // update statistics
        dimX++;
        if(ABS(tmpfval) > ZERO_EPS) featureCnt++;
        
      }

      // check dimension consistency
      if(preDimX != -1 && preDimX != dimX) {
        printf("\n%s ERROR: {CDataset} feature dimension of examples #%d (%d) & #%d (%d) do not match !\n",
               tag, totalExample, preDimX, totalExample+1, dimX);
        return ERROR;
      }
      preDimX = dimX;

       // chteo: feature vector of all zero is fine for dense file ?!?
//       // check non-zero feasibility
//       if(featureCnt <= 0) {
//         printf("\n%s ERROR: {CDataset} example #%d has no non-zero feature values!\n", tag, totalExample+1);
//         return ERROR;
//       }
      
      // update statistics
      if(info->bias)
        featureCnt += 1;  // for shifted hyperplane
      exampleStat->push_back(featureCnt);
      sampleStat->at(totalSample-1) += 1;
      totalFeature += featureCnt;
      totalExample++;
    }
  }

  // for shifted hyperplane: add extra feature (dimension)
  if(info->bias)
    dimX++;
   

  if(info->programMode != PREDICTION_MODE) {
    // check whether # of data and # of label match
    if(nLabelLine != totalExample) {
      printf("\n%s ERROR: {CDataset} # of data != # of labels\n", tag);
      printf("%s # data : %d   # label : %d\n", tag, totalExample, nLabelLine);
      return ERROR;
    }
  }

  if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
    if(info->verbosity > 0) {
      printf("%s Done\n", tag);
      fflush(stdout);
    }
  }
  
  // data matrix density
  density = (double)totalFeature/dimX/totalExample;
 
  // check the density of dataset and see whether dataset
  //   should be stored in DENSE matrix format
  if(density >= DENSE_MAT_THRESHOLD)
    datasetType = DENSE_DS;
  else
    datasetType = SPARSE_DS;
    
  // finally, user decide how dataset is stored
  if(info->datasetType == FORCE_SPARSE) datasetType = SPARSE_DS;
  if(info->datasetType == FORCE_DENSE) datasetType = DENSE_DS;
  
  
  
  // display statistics
  if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) 
  {    
    printf("%s Dataset statistics :\n", tag);
    printf("%s 1. Data filename : %s\n", tag, info->datasetFn.c_str());
    printf("%s 2. Label filename : %s\n", tag, info->labelsetFn.c_str());
    printf("%s 3. Number of samples : %d\n", tag, totalSample);
    printf("%s 4. Number of examples : %d\n", tag, totalExample);    
    printf("%s 5. Number of nonzero features : %d  %s\n", tag, totalFeature, (info->bias) ? "(+ # of examples for shifted hyperplane)" : "");
    printf("%s 6. Total number of labels : %d  (average : %.2f labels per example)\n", tag, totalLabel, 1.0*totalLabel/totalExample);
    printf("%s 7. Max. number of labels per example : %d\n", tag, dimY);
    printf("%s 8. Dimension : %d  %s\n", tag, dimX, (info->bias) ? "(+ 1 for shifted hyperplane)" : "");
    printf("%s 9. Dataset density : %.4f %%\n", tag, density*100.0);
  }

  labelFp.close();
  dataFp.close();
  
  return NOERROR;
}



/** For parallel and distributed computation cases only.
 *  Called only after ScanDS().
 */
ErrorCode CDataset::BalanceLoad()
{

  int quotient  = 0;   // min # of examples on each machine
  int remainder = 0;   // remainder of examples after dataset is divided evenly
  int sample_idx  = 0;
  int example_idx = 0;
 

  // determine # of machine actually used
  // Assumption: nSample > 1 ==> sample-based loss used
  // When 1 < nSample < nNode, we use only nSample machines
  //   because sample-based loss is not decomposable any further...
  // Exception: we receive a "ignore-qid-flag" from command line arg (not implemented yet)

  partID = info->procID;
  nPart  = info->nUsedNode;
  
  startExample = 0;
  startSample  = 0;
  
  nSample      = 0;
  nExample     = 0;
  nFeature     = 0;
  nLabel       = 0;

  /*
   *  Compute nSample and determine how many machine will be used
   */
  if(totalSample == 1) {    
    info->nUsedNode = MIN(info->nNode, totalExample);
    SOME_MACHINES(info->procID, info->nUsedNode) {
      nSample = 1;    
    }
    // else, this machine is extra, hence nSample = 0    
  }
  else if(totalSample > 1 && totalSample < info->nNode) {
    info->nUsedNode = totalSample;
    SOME_MACHINES(info->procID, info->nUsedNode) {
      nSample = 1; quotient = 1;
    }
    //else, this machine is extra, hence nSample = 0; quotient = 0; remainder = 0;    
  }
  else {
    info->nUsedNode = info->nNode;
    nSample = totalSample/info->nUsedNode;
    quotient  = nSample;
    remainder = totalSample - nSample*info->nUsedNode;
    for(int i=0; i<remainder; i++) if(i == info->procID) nSample++;
  }
   
 
  /*
   *  Compute nExample, nFeature, nLabel, startExample, and startSample
   */
  SOME_MACHINES(info->procID, info->nUsedNode)
  {
    if(totalSample == 1) {
      // divide THE sample...
      nExample  = totalExample / info->nUsedNode;
      quotient  = nExample;
      remainder = totalExample - nExample*info->nUsedNode;
      for(int i=0; i<remainder; i++) if(i == info->procID) nExample++;
      
      for(int pid=0; pid < info->nUsedNode; pid++) {
        if(info->procID == pid) {
          for(int i=0; i<nExample; i++) {
            nFeature += exampleStat->at(example_idx + i);
            nLabel   += labelStat->at(example_idx + i);	
          }	                  
          break;
        }
        else {
          example_idx += quotient + (pid < remainder);
          startExample = example_idx;
        }        
      }
      
    }
    else {
      // samples are never splitted in sample-based loss cases...
      for(int pid=0; pid < info->nUsedNode; pid++) {
        if(info->procID == pid) {
          for(int i=0; i<nSample; i++) {
            int cur_nExample = sampleStat->at(sample_idx + i);

            for(int j=0; j<cur_nExample; j++) {
              nFeature += exampleStat->at(example_idx + j);
              nLabel   += labelStat->at(example_idx + j);	
              //printf("%s nLabel : %d\n",tag, nLabel);
            }
            nExample += cur_nExample;
          }	                  
          break;
        }
        else {
          int n = quotient + (pid < remainder);
          for(int i=0; i<n; i++) {
            example_idx += sampleStat->at(sample_idx);
            sample_idx++;
          }
          startExample = example_idx;
          startSample  = sample_idx;
        }        
      }
    }
  }

  printf("%s nS:%d,  nE:%d,  nF:%d,   nL:%d,    sE:%d\n",tag, nSample,nExample,nFeature, nLabel, startExample);

  return NOERROR;
}





/** Read dataset from either SPARSE or DENSE data file and store them
 *  as SPARSE dataset. For serial, parallel and distributed cases.
 */
ErrorCode CDataset::ReadSparseDS()
{
  ErrorCode ec = NOERROR;

  int      i = 0;
  int      j = 0;
  int      d_idx = 0;           // data index
  int      e_idx = 0;           // example index (reset to 0 for new sample)
  int      s_idx = -1;          // sample index (will be increased by at least once before use)
  int      line_cnt = 0;
  int      curQID = 0;          // current query id
  int      preQID = -1;         // previous query id
  int      tmpidx = 0;          // temp. index
  Real     tmpval = 0;          // temp. values
  int      featureCnt = 0;      // number of features per example
  string   line = "";
  string   token = "";
  ifstream dataFp;
  
  
  // for the sake of computing effective dimensionality of the data matrix
  vector<bool> dimMarker(dimX,false);


  // verbose a bit
  if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
    if(info->verbosity > 0) {
      string mode = "";
      if(info->compMode == SERIAL_CENTRALISED_DS)
        mode = "SERIAL";
      else if(info->compMode == PARALLEL_CENTRALISED_DS)
        mode = "PARALLEL CENTRALISED";
      else
        mode = "PARALLEL DISTRIBUTED";
      
      printf("%s Reading %s SPARSE dataset...", tag, mode.c_str());
      fflush(stdout);
    }
  }

 
  // check whether data exceed the limit
  if(nFeature > MAX_NNZ) {
    printf("\n%s ERROR: {CDataset} number of non-zero features exceeds the limit : %d !\n", tag, MAX_NNZ);
    return ERROR;
  }

  /*
   *    START READING LABELS
   */
  if(info->programMode != PREDICTION_MODE)
    ec = ReadLabel();
  
  if(ec != NOERROR) return ec;
  
  
  /*
   *    START READING DATA (FEATURE VECTORS)
   */
  
  // update info
  info->dimX = dimX;
  info->dimW = dimX;

  // allocate memory space for dataset
  row_ptr = new int[nExample+1];
  idx     = new int[nFeature];
  val     = new Real[nFeature];
  sample  = new CSample[nSample];
  
  // initialize the memory space allocated
  memset(row_ptr, 0, sizeof(int)*(nExample+1));
  memset(idx,     0, sizeof(int)*nFeature);
  memset(val,     0, sizeof(Real)*nFeature);
 
  int exIdx = 0;
  for(i=0, j=startSample; i<nSample; i++, j++) {
    sample[i].firstIdx = exIdx;
    sample[i].len      = sampleStat->at(j);
    sample[i].queryID  = 0;
    exIdx += sample[i].len;
  }

  // open feature vector file
  dataFp.open(info->datasetFn.c_str());
  if(!dataFp.good()) {
    printf("\n%s ERROR: {CDataset} can not open data file <%s> !\n", tag, info->datasetFn.c_str());
    return ERROR;
  }
  
  // skip #startExample# example lines. these lines are not belong to this machine
  for(line_cnt=0; line_cnt < startExample && !dataFp.eof(); ) {
    getline(dataFp, line);
    TrimWhitespaces(line);
    if(IsBlankLine(line)) continue;  // blank line
    if(line[0] == '#') continue;     // comment
    line_cnt++;
  }
  if(dataFp.eof()) {
    printf("\n%s ERROR: {CDataset} End of file <%s> reached before loading the content !\n", tag, info->datasetFn.c_str());
    return ERROR;
  }

  // read #nExample# feature vectors
  for(line_cnt=0, d_idx=0; line_cnt < nExample && !dataFp.eof();) {

    // set row pointer (for compressed sparse row format)
    row_ptr[line_cnt] = d_idx; 
    
    // read data line    
    getline(dataFp, line);
    TrimWhitespaces(line);
    if(IsBlankLine(line)) continue;  // blank line
    if(line[0] == '#') continue;     // comment

    istringstream issdata(line);
    
    // query id is always in front of features !
    size_t pos = line.find("qid:");
    if(pos != string::npos) {
      sscanf(&line[pos], "qid:%d ", &curQID);      
      issdata >> token; // throw away this token
    }
    
    if(preQID != curQID) {
      s_idx++;          // next sample
      e_idx = 0;        // reset
      preQID = curQID;
    }
    
    // set example pointer
    sample[s_idx].queryID = curQID;      
    
    // read a feature vector
    if(datafileType == SPARSE_FILE) 
    {
      for(featureCnt=0; !issdata.eof();) {
        issdata >> token;
        if(token[0] == '#') break;
              
        sscanf(token.c_str(), "%d:%lf", &idx[d_idx], &val[d_idx]);

        idx[d_idx] -= 1; // convert feature indexing from 1-based to 0-based
	dimMarker[idx[d_idx]] = true;   // marking this coordinate as used

        d_idx++;
        featureCnt++;
      }
    }
    else 
    { // datafileType == DENSE_FILE
      for(featureCnt=0, tmpidx=0; !issdata.eof(); tmpidx++) {
        issdata >> token;
        if(token[0] == '#') break;
         
        sscanf(token.c_str(), "%lf", &tmpval);

        if(ABS(tmpval) > ZERO_EPS) {
          idx[d_idx] = tmpidx;  // 0-based indexing
          val[d_idx] = tmpval;
	  dimMarker[idx[d_idx]] = true;   // marking this coordinate as used

          d_idx++;
          featureCnt++;
        }
      }
    }

    // add extra feature (dimension) for shifted hyperplane
    if(info->bias) {
      idx[d_idx] = dimX-1;  // 0-based indexing
      val[d_idx] = 1.0;
      dimMarker[idx[d_idx]] = true;   // marking this coordinate as used

      d_idx++;
      featureCnt++;
    }
  
    // increace counters
    line_cnt++;
    e_idx++;  
  }  

  // final row_ptr entry
  row_ptr[nExample] = nFeature;
  
  // create PETSc sparse compressed row matrix (AIJ)
  MatCreateSeqAIJWithArrays(PETSC_COMM_SELF, nExample, dimX, row_ptr, idx, val, &X);
  MatAssemblyBegin(X, MAT_FINAL_ASSEMBLY);
  MatAssemblyEnd(X, MAT_FINAL_ASSEMBLY);  

  // Find the max CInfo::k amongst all nodes
  if(info->compMode != SERIAL_CENTRALISED_DS) {
    int tmp_k = info->k;
    MPI_Allreduce(&tmp_k, &(info->k), 1, MPI_INT, MPI_MAX, BMRM_COMM_WORLD);
  }

  
  /*
   *   FINISHED READING LABEL AND FEATURE VECTORS
   */

  // verbose a bit
  if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
    if(info->verbosity > 0) {
      printf("%s Done\n", tag);
      fflush(stdout);
    }
  }

  if(info->verbosity > 1) {
      int nnz = 0;
      for(int i=0; i<dimX; i++)
	  if(dimMarker[i]) nnz++;
      cout << "Effective dimensionality of the data matrix: " << nnz << endl;
  }
 
  // clean up
  delete exampleStat; exampleStat = 0;
  delete sampleStat; sampleStat = 0;
  if(info->programMode != PREDICTION_MODE) {
    delete labelStat; 
    labelStat = 0;
  }
  return NOERROR;
}




/** Read dataset from either SPARSE or DENSE data file and store them
 *  as DENSE dataset.
 */
ErrorCode CDataset::ReadDenseDS()
{
  ErrorCode ec = NOERROR;

  int      i = 0;
  int      j = 0;
  int      d_idx = 0;           // data index
  int      e_idx = 0;           // example index (reset to 0 for new sample)
  int      s_idx = -1;          // sample index (will be increased by at least once before use)
  int      line_cnt = 0;
  int      curQID = 0;          // current query id
  int      preQID = -1;         // previous query id
  int      featureCnt = 0;      // number of features per example
  int      tmpidx = 0;
  Real     tmpval = 0; 
  string   line = "";
  string   token = "";
  ifstream dataFp;
  
  
  // for PETSc MatSetValues()
  PetscInt whichRow = 0;            
  int      *sequence = 0;
  sequence = new int[dimX];
  for(int i=0; i<dimX; i++) sequence[i] = i;


  // verbose a bit
  if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
    if(info->verbosity > 0) {
      string mode = "";
      if(info->compMode == SERIAL_CENTRALISED_DS)
	mode = "SERIAL";
      else if(info->compMode == PARALLEL_CENTRALISED_DS)
	mode = "PARALLEL CENTRALISED";
      else
	mode = "PARALLEL DISTRIBUTED";
      
      printf("%s Reading %s DENSE dataset...", tag, mode.c_str());
      fflush(stdout);
    }
  }  
  
  // for dense matrix
  nFeature = nExample * dimX;  
  

  // check whether data exceed the limit
  if(nFeature > MAX_NNZ) {
    printf("\n%s ERROR: {CDataset} number of non-zero features exceeds the limit : %d !\n", tag, MAX_NNZ);
    return ERROR;
  }

  
  
  /*
   *    START READING LABELS
   */
  if(info->programMode != PREDICTION_MODE)
    ec = ReadLabel();
  
  if(ec != NOERROR) return ec;
    
  
  /*
   *    START READING DATA (FEATURE VECTORS)
   */
  
  // update info
  info->dimX = dimX;

  
  // allocate memory space for dataset
  val    = new Real[dimX];   // used to store 1 example only
  sample = new CSample[nSample];
   
  // initialize the memory space allocated
  memset(val, 0, sizeof(Real)*dimX);

  int exIdx = 0;
  for(i=0, j=startSample; i<nSample; i++, j++) {
    //sample[i].example = new CExample[sampleStat->at(j)]; 
    sample[i].firstIdx = exIdx;
    sample[i].len      = sampleStat->at(j);
    sample[i].queryID  = 0;
    exIdx += sample[i].len;
  }

   
  // update info
  info->dimX = dimX;
  info->dimW = dimX;

  if(labelType == SINGLE_LABEL)
    info->k = 1;      // max label (integral value)
  else
    info->k = dimY;   // max number of label per example


  // open feature vector file
  dataFp.open(info->datasetFn.c_str());
  if(!dataFp.good()) {
    printf("\n%s ERROR: {CDataset} can not open data file <%s> !\n", tag, info->datasetFn.c_str());
    return ERROR;
  }
    
  // skip #startExample# example lines. these lines are not belong to this machine
  for(line_cnt=0; line_cnt < startExample && !dataFp.eof(); ) {
    getline(dataFp, line);
    TrimWhitespaces(line);
    if(IsBlankLine(line)) continue;  // blank line
    if(line[0] == '#') continue;     // comment
    line_cnt++;
  }
  if(dataFp.eof()) {
    printf("\n%s ERROR: {CDataset} End of file <%s> reached before loading the content !\n", tag, info->datasetFn.c_str());
    return ERROR;
  }
  

  // create a DENSE PETSc data matrix
  MatCreateSeqDense(PETSC_COMM_SELF, nExample, dimX, PETSC_NULL, &X);

  // read feature vectors from file depending on its format detected
  for(line_cnt=0, d_idx=0; line_cnt < nExample && !dataFp.eof();) {
    
    // read data line
    getline(dataFp, line);
    TrimWhitespaces(line);
    if(IsBlankLine(line)) continue;  // blank line
    if(line[0] == '#') continue;     // comment
  
    istringstream issdata(line);
    
    // query id is always in front of features !
    size_t pos = line.find("qid:");
    if(pos != string::npos) {
      sscanf(&line[pos], "qid:%d ", &curQID);      
      issdata >> token; // throw away this token
    }
    
    if(preQID != curQID) {
      s_idx++;          // next sample
      e_idx = 0;        // reset
      preQID = curQID;
    }
    
    // set example pointer
    sample[s_idx].queryID = curQID;      
    
    // iterate over all features
    if(datafileType == SPARSE_FILE) {
      for(featureCnt=0; !issdata.eof();) {
        issdata >> token;
        if(token[0] == '#') break;
        
        sscanf(token.c_str(), "%d:%lf", &tmpidx, &tmpval);

        for(;tmpidx-1 > featureCnt; featureCnt++, d_idx++) {
          val[featureCnt] = 0.0;
        }
        val[featureCnt] = tmpval;
        d_idx++;
        featureCnt++;
        
      }
      // fill in the rest of the entries
      for(;featureCnt < dimX-1; featureCnt++, d_idx++) val[featureCnt] = 0.0;      
    }
    else {  // DENSE_FILE
      for(featureCnt=0; !issdata.eof();) {
        issdata >> token;
        if(token[0] == '#') break;
        
        sscanf(token.c_str(), "%lf", &val[featureCnt]);
        d_idx++;
        featureCnt++;
      }
    }

    // add extra feature (dimension) for shifted hyperplane
    if(info->bias) {
      val[featureCnt] = 1.0;
      featureCnt++;
      d_idx++;
    }

    // insert the example into data matrix X
    MatSetValues(X, 1, &whichRow, dimX, sequence, val, INSERT_VALUES);
    whichRow++;

    // increace counters
    line_cnt++;
    e_idx++;  
  }  

  // assemble the data matrix after inserting values
  MatAssemblyBegin(X, MAT_FINAL_ASSEMBLY);
  MatAssemblyEnd(X, MAT_FINAL_ASSEMBLY);

  // Find the max CInfo::k amongst all nodes
  if(info->compMode != SERIAL_CENTRALISED_DS) {
    int tmp_k = info->k;
    MPI_Allreduce(&tmp_k, &(info->k), 1, MPI_INT, MPI_MAX, BMRM_COMM_WORLD);
  }
 

  /*
   *   FINISHED READING LABEL AND FEATURE VECTORS
   */

  // verbose a bit
  if(info->procID == MYROOT || info->compMode == PARALLEL_DISTRIBUTED_DS) {
    if(info->verbosity > 0) {
      printf("%s Done\n", tag);
      fflush(stdout);
    }
  }


  // clean up
  delete exampleStat; exampleStat = 0;
  delete sampleStat; sampleStat = 0;
  delete labelStat; labelStat = 0;
  delete [] sequence; sequence = 0;

  return NOERROR;
}



/** Read labels
 *
 */
ErrorCode CDataset::ReadLabel()
{
  // variables
  int          line_cnt = 0;
  int          labelCnt = 0;        // number of labels per example
  int          l_idx = 0;           // label index
  string       line;
  ifstream     labelFp;
  Real         tmp_label;
  vector<Real> tmp_labelset;

  
  // allocate memory for labelset
  if(labelType == SINGLE_LABEL) {
    Y.STL_vec.reserve(nExample);
    info->k = 1;      // max label (integral value)
  }
  else {
    Y.STL_vvec.reserve(nExample);
    info->k = dimY;   // max number of label per example
  }
    
  // open label file
  if(info->programMode != PREDICTION_MODE) {
    labelFp.open(info->labelsetFn.c_str());
    if(!labelFp.good()) {
      printf("\n%s ERROR: {CDataset} can not open label file <%s> !\n", tag, info->labelsetFn.c_str());
      return ERROR;
    }
  }
  
  // skip #startExample# example lines. these lines are not belong to this machine
  for(line_cnt=0; line_cnt < startExample && !labelFp.eof(); ) {
    getline(labelFp, line);
    TrimWhitespaces(line);
    if(IsBlankLine(line)) continue;  // blank line
    if(line[0] == '#') continue;     // comment
    line_cnt++;
  }
  if(labelFp.eof()) {
    printf("\n%s ERROR: {CDataset} End of file <%s> reached before loading the content !\n", tag, info->labelsetFn.c_str());
    return ERROR;
  }
  
  // read labels  
  for(line_cnt=0, l_idx=0; line_cnt < nExample && !labelFp.eof() ;) {
    // read label (skip lines with comments)    
    do {
      getline(labelFp, line);
      TrimWhitespaces(line);
    }while((IsBlankLine(line) || line[0] == '#') && !labelFp.eof());
    
    if(labelFp.eof() && (IsBlankLine(line) || line[0] == '#')) break;
    
    istringstream isslabel(line);
    
    if(labelType == SINGLE_LABEL) {
      isslabel >> tmp_label;
      Y.STL_vec.push_back(tmp_label);
      info->k = MAX(info->k, (int)tmp_label);
      l_idx++;
      labelCnt++;
    }
    else { // labelType == MULTI_LABEL
      tmp_labelset.resize(0);
      for(labelCnt=0; !isslabel.eof(); labelCnt++, l_idx++)  {
	isslabel >> tmp_label;
	tmp_labelset.push_back(tmp_label);
      }
      Y.STL_vvec.push_back(tmp_labelset);
    }
    
    // increace counters
    line_cnt++;
  }
  
  // create PETSc vector (this only make sense in SINGLE label cases) 
  if(labelType == SINGLE_LABEL) {
    VecCreateSeqWithArray(PETSC_COMM_SELF, nLabel, &(Y.STL_vec[0]), &Y.vec);
    VecAssemblyBegin(Y.vec);
    VecAssemblyEnd(Y.vec);
  }

  return NOERROR;
}


#endif


/*
  NOTES:
  
  1. #idx# is of type int because if it is not, we could not represent the
     solution (dense) vector, w, in most of the desktop machine memory.

  2. #idx# will be empty if dataset is in dense format.

  3. Sparse dataset has 1-based indexing feature indices

*/
