/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 16/11/2006
 * Last Updated : 15/02/2007
 */

#ifndef _DATASET_HPP_
#define _DATASET_HPP_

#include <string>
#include <vector>
#include <fstream>
#include "errorcode.hpp"
#include "common.hpp"
#include "info.hpp"

#include "petsc.h"
#include "petscmat.h"

#define SPARSE_DS 1
#define DENSE_DS  2

#define SPARSE_FILE 1
#define DENSE_FILE  2

#define MAX_NNZ 1000000000  // 1 billion

#define DENSE_MAT_THRESHOLD  0.66  //store dataset in dense matrix if density >= 66%

#define SINGLE_LABEL 1
#define MULTI_LABEL  2


using namespace std;


/** Label class. Contain some possible forms of label 
 *  e.g., array, PETSc Vec, and STL vector or vector.
 *  This class can also handle single and multi label cases. 
 */
class CLabel {
public:
  vector<Real>           STL_vec;      // for single label
  Vec                    vec;          // PETSc vec view of STL_vec
  vector<vector<Real> >  STL_vvec;     // for multi label case

  void ToFile(string filename);
};




/** Sample class. Contains sample-level annotations to the data matrix.
 */
class CSample {
public: 
  int  firstIdx;         // row index of the first example in the data matrix
  int  len;              // num of examples
  long queryID;          // query ID
};


/** Dataset class.
 *  Data can be in sparse or dense matrix format depending on its density.
 */
class CDataset {

public:

  CInfo   *info;         // learning parameters & miscellaneous parameters

  char    tag[128];      // process ID and hostname of the machine
  int     partID;        // ID of this part of sub-dataset
  int     nPart;         // num of parts of whole dataset
  int     datafileType;  // type of data file: dense of sparse ?
  int     datasetType;   // type of dataset (in memory): dense of sparse ?
  int     labelType;     // single label or multi-label?

  int     *row_ptr;      // indices of new rows (examples)
  int     *idx;          // feature indices
  Real    *val;          // feature values
  Mat     X;             // PETSc matrix view of dataset
  CLabel  Y;             // labels
  CSample  *sample;      // subset of examples having same query id
  int     nClass;        // number of class (for multi-class classification)
  int     dimY;          // dimensionality of label
  int     dimX;          // dimensionality of feature vector (+ 1 for shifted hyperplane)
  int     effectiveDimX; // the effective dimensionality of X (i.e. how many coordinates are actually used in the data matrix)
  float   density;       // dataset density

  int     totalLabel;    // total number of label (for multi-label case)
  int     totalSample;   // num of samples in whole dataset
  int     totalExample;  // num of examples in whole dataset
  int     totalFeature;  // total num of non-zero features of whole dataset

  int     nLabel;        // num of label in this sub-dataset (for parallel mode)
  int     nSample;       // num of samples in this sub-dataset (for parallel mode)  
  int     nExample;      // num of examples in this sub-dataset (for parallel mode)
  int     nFeature;      // num of non-zero features in this sub-dataset

  int     startExample;  // the # of examples to be skipped before 
                         //   the machine starts reading nExample examples

  int     startSample;   // the # of samples to be skipped before 
                         //   the machine starts reading nSample samples

  // Constructors
  CDataset();
  CDataset(CInfo *_info);

  // Destructor
  virtual ~CDataset();

protected:

  // Methods
  ErrorCode ScanDS();
  ErrorCode BalanceLoad();
  ErrorCode ReadSparseDS();
  ErrorCode ReadDenseDS();
  ErrorCode ReadLabel();


  // utilities
  bool IsBlankLine(string &line);
  void TrimWhitespaces(string &line);

private:

  vector<int> *exampleStat;  // keep number of features per example
  vector<int> *sampleStat;   // keep number of examples per sample
  vector<int> *labelStat;    // keep number of labels per example (used in parallel mode)

};

#endif


/*
  NOTES:
  
  1. #idx# is of type int.

  2. #idx# will be empty if dataset is in dense format.

*/
