/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 18/12/2006
 *
 * Last Updated :
 */

#ifndef _INPUTOUTPUT_CPP_
#define _INPUTOUTPUT_CPP_

#include "inputoutput.hpp"

using namespace std;



/** Parse command line arguments for prediction.
 *
 *  \param argc [read]  Number of arguments
 *  \param argv [read]  Arguments
 *  \param info [write] Learning and miscellaneous parameters
 */
void ParseCmdLineArgPredict(int argc, char**argv, CInfo *info)
{
  int  i = 0;
  string tmp = "";

  // default parameters
  info->learningType = 0;  // default: pick the loss function stated in model file
  info->programMode  = PRED_AND_EVAL_MODE;
  info->verbosity    = 1;
  info->outFormat    = LABEL_ONLY;

  // options
  for(i=1; i < argc; i++)
  {
    if(argv[i][0] != '-') break;   
    if(++i >= argc) PrintPredictUsageMessage();

    switch(argv[i-1][1]) 
    {
    case 's': info->learningType  = atoi(argv[i]); break;
    case 'f': info->outFormat     = atoi(argv[i]); break;
    case 'X': info->datasetFn     = string(argv[i]); break;
    case 'Y': info->labelsetFn    = string(argv[i]); break;        
    case 'M': info->modelFn       = string(argv[i]); break;        
    case 'O': tmp                 = string(argv[i]); break;        
        
    default:
      printf("ERROR: Unrecognised option <-%c> !\n", argv[i-1][1]);
      PrintPredictUsageMessage();
      exit(1);
    }
  }

  if(info->learningType < 0 || info->learningType > NUMBER_OF_LOSS)
  {
    printf("ERROR: Invalid minimisation problem !  (-s:%d)\n", info->learningType); 
    exit(1);
  }

  // dataset file    
  if(info->datasetFn == "") {
    printf("ERROR: testing data set is missing !\n");
    PrintPredictUsageMessage();
    exit(EXIT_FAILURE);
  }

  // model file
  if(info->modelFn == "") {
    printf("ERROR: model file is missing !\n");
    PrintPredictUsageMessage();
    exit(EXIT_FAILURE);    
  }
  
  // label and fval output filenames
  if(tmp == "") {
    size_t loc = info->datasetFn.rfind("/");
    if(loc != string::npos)
      tmp = info->datasetFn.substr(loc+1);
    else
      tmp = info->datasetFn;
  
    info->outputLabelFn = tmp+".prediction";
    info->outputFValFn= info->outputLabelFn+".fval";
  }
  else {
    switch(info->outFormat) {
    case LABEL_ONLY:
      info->outputLabelFn = tmp;
      break;
    case FVAL_ONLY:
      info->outputFValFn= tmp;
      break;
    case LABEL_AND_FVAL:
      info->outputLabelFn = tmp;
      info->outputFValFn= info->outputLabelFn+".fval";
      break;
    }
  }

  // program mode: 
  // PRED_AND_EVAL_MODE - do prediction and evaluate performance
  // PREDICTION_MODE    - do prediction only
  if(info->labelsetFn == "")
    info->programMode = PREDICTION_MODE;
}



/** Parse command line arguments for training.
 *
 *  \param argc [read]  Number of arguments
 *  \param argv [read]  Arguments
 *  \param info [write] Learning and miscellaneous parameters
 */
void ParseCmdLineArgTrain(int argc, char**argv, CInfo *info)
{
  int  i = 0;
  bool setGammaTol = false;
  bool setEpsilonTol = false;

  // default parameters
  info->convLog      = 0;                         // keep convergence log in files? 0:=no, 1:=yes
  info->programMode  = LEARNING_MODE;
  info->learningType = SOFT_MARGIN_LOSS;          // 1
  info->bias         = 0;                         // applied shifted hyperplane ? (default : 0 (no))
  info->k            = 1;                         // number of stacked w
  info->C            = 1;                         // slack trade-off
  info->lambda       = 1;                         // regularizer scaling factor
  info->epsilon      = 0.1;                       // for epsilon-SVR
  info->tau          = 0.1;                       // for quantile regression, etc.
  info->truncation   = 10;                        // for NDCG ranking
  info->gammaTol     = 0.01;                      // tolerance for gamma-termination criterion, default option
  info->epsilonTol   = -1;                        // tolerance for epsilon-termination criterion, switched off by default
  info->maxIter      = 4000;                      // maximum qp iteration
  info->solverType   = DAIFLETCHER;               // Dai-Fletcher PGM
  info->regularizer  = L2NORMSQ;                  // l2 norm square
  info->memSize      = MIN(400,info->maxIter);    // default: limited by # of iterations
  info->modelFn      = "";                        // model file name
  info->datasetType  = AUTOMATIC;                 // what type of dataset to store in memory ? {dense, or sparse}
                                                  //  bmrm determine this by default
  //info->compMode     = PARALLEL_CENTRALISED_DS;   // computation mode

  char hostname[128];
  gethostname(hostname, 128);
  info->hostname = string(hostname);


  // options
  for(i=1; i < argc; i++)
  {
    if(argv[i][0] != '-') break;   
    if(++i >= argc) PrintTrainUsageMessage();

    switch(argv[i-1][1]) 
    {
      case 's':	info->learningType = atoi(argv[i]); break;
      case 'r':	info->solverType   = atoi(argv[i]); break;
      case 'm': info->memSize      = atoi(argv[i]); break;
      case 'b': info->bias         = atoi(argv[i]); break;      
      case 'c':	info->C            = atof(argv[i]); break;
      case 'l':	info->lambda       = atof(argv[i]); break;
      case 'p': info->epsilon      = atof(argv[i]); break;
      case 'a': info->tau          = atof(argv[i]); break; 
      case 't': info->truncation   = atoi(argv[i]); break;
      case 'e': info->epsilonTol   = atof(argv[i]); setEpsilonTol = true; info->gammaTol = -1; break; 
      case 'g': info->gammaTol     = atof(argv[i]); setGammaTol = true; info->epsilonTol = -1; break; 
      case 'i': info->maxIter      = atoi(argv[i]); break;
      case 'v': info->verbosity    = atoi(argv[i]); break;        
      
      case 'T': info->datasetType  = atoi(argv[i]); break;
      case 'X': info->datasetFn    = string(argv[i]); break;
      case 'Y': info->labelsetFn   = string(argv[i]); break;        
      case 'D': info->dsList       = string(argv[i]); break; // not implemented yet
      case 'M': info->modelFn      = string(argv[i]); break;
      case 'L': info->convLog      = atoi(argv[i]); break;

      default:
        printf("\nERROR: Unrecognised option <-%c> !\n", argv[i-1][1]);
        PrintTrainUsageMessage();
        exit(EXIT_FAILURE);
    }
  }  

  if(i == 1) {
    printf("\nERROR: No data file and label file detected !\n");
    PrintTrainUsageMessage();
    exit(EXIT_FAILURE);
  }

  
  // arguments validation
  if(info->datasetFn == "" && info->dsList == "")
  {
    printf("ERROR: No data file or distributed data list detected !\n");
    exit(EXIT_FAILURE);
  }

  if(info->labelsetFn == "" && info->datasetFn != "")
  {
    printf("ERROR: No label file detected !\n");
    exit(EXIT_FAILURE);
  }

  if(info->datasetFn != "" && info->dsList != "")
  {
    printf("ERROR: Data file and distributed data list detected. Please specify only one of them.");
    exit(EXIT_FAILURE);
  }

  if(info->convLog < 0 || info->convLog > 1)
  {
    printf("ERROR: parameter -L  must be in {0,1}! (-L: %d)\n",info->convLog);
    exit(EXIT_FAILURE);
  }


  if(info->datasetType < 1 || info->datasetType > 3)
  {
    printf("ERROR: Data set type must be in [1,2,3].");
    exit(EXIT_FAILURE);
  }

  if(info->learningType < 1 || info->learningType > NUMBER_OF_LOSS)
  {
    printf("ERROR: Invalid minimisation problem !  (-s:%d)\n", info->learningType); 
    exit(1);
  }

  if(info->solverType < 1 || info->solverType > NUMBER_OF_SOLVER)
  {
    printf("ERROR: parameter -r must be in [1..%d] !  (-r:%d)\n", NUMBER_OF_SOLVER,info->solverType);
    exit(1);
  }

  if(info->memSize < 1)
  {
    printf("Error: parameter -m must be positive integer ! (-m%d)\n",info->memSize);
    exit(1);
  }

  if(info->C <= 0)
  {
    printf("ERROR: parameter -c must be > 0 !\n");
    exit(1);
  }

  if(info->lambda <= 0)
  {
    printf("ERROR: parameter -l must be > 0 !\n");
    exit(1);
  }

  if(info->gammaTol < 0 && info->epsilonTol < 0)
  {
    printf("ERROR: parameter -{e,g} must be > 0 !\n");
    exit(EXIT_FAILURE);
  }
  else if(setGammaTol && setEpsilonTol > 0) 
  {
    printf("ERROR: choose either -g or -e parameter !\n");
    exit(EXIT_FAILURE);
  }

  if(info->epsilon < 0)
  {
    printf("ERROR: parameter -p must be >= 0 !\n");
    exit(1);
  }

  if(info->tau < 0 || info->tau > 1)
  {
    printf("ERROR: parameter -a must be in the range [0,1] !\n");
    exit(1);
  }

  if(info->truncation < 0)
  {
    printf("ERROR: parameter -e must be >= 0 !\n");
    exit(1);
  }

  if(info->maxIter < 1)
  {
    printf("ERROR: parameter -i must be >= 1 !\n");
    exit(1);
  }

  if(info->solverType < 1 || info->solverType > 4)
  {
    printf("ERROR: parameter -r must be in [1,4] !\n");
    exit(1);
  }

  if(info->verbosity < 0)
  {
    printf("ERROR: parameter -v must be >= 0 !\n");
    exit(1);
  }

  
  // update dependent parameters
  info->memSize = MIN(info->memSize, info->maxIter);
  if(info->solverType == L1LP) info->regularizer = L1NORM;

  // distributed dataset list
  // Note: At this point, all machine has a hostname
  // Look at the file and assign data and label files accordingly
  if(info->dsList != "") 
  {
    if(info->nNode == 1) {
      printf("ERROR: serial computation does not accept distributed dataset\n");
      exit(EXIT_FAILURE);
    }
    
    info->compMode = PARALLEL_DISTRIBUTED_DS;

    // every machine checks for their corresponding dataset in the list
    // where is the list? assume the list is in the bmrm exec folder
    ifstream ifp(info->dsList.c_str());
    if(!ifp.good()) {
      printf("{proc:%2d:%s} ERROR: Can not open list file <%s> !\n",info->procID, info->hostname.c_str(), info->dsList.c_str());
      exit(EXIT_FAILURE);
    }
    
    // the format of ds list:
    // <machine_name_1> <data_file_1> <label_file_1>
    // <machine_name_*> is in lower case
    // each machine has only 1 entry in the list

    string line = "";
    string machine_name = "";
    string data_file = "";
    string label_file = "";
    
    info->datasetFn = "";
    info->labelsetFn = "";
    
    while(!ifp.eof()) 
    {
      getline(ifp, line);
      istringstream iss(line);
      iss >> machine_name >> data_file >> label_file;
      if(machine_name == info->hostname) {
	info->datasetFn = data_file;
	info->labelsetFn = label_file;
	break;
      }
    }
    
    if(info->datasetFn == "" || info->labelsetFn == "") {
      printf("{proc:%2d:%s} ERROR: Can not find data and/or label file in the list !\n",info->procID, info->hostname.c_str());
      exit(EXIT_FAILURE); 
      // we can ignore the failed node and continue the training (not implemented)
    }
  }

  // model file
  if(info->modelFn == "") {
    size_t loc = info->datasetFn.rfind("/");
    if(loc != string::npos)
      info->modelFn = info->datasetFn.substr(loc+1)+".model";
    else
      info->modelFn = info->datasetFn+".model";
  }
}

/** Print all the loss options
 *
 */
void PrintLossOptions()
{
  printf("   Binary Classification:\n");
  printf("     %2d -- Soft Margin             %d -- Squared Soft Margin\n",SOFT_MARGIN_LOSS, SQUARED_SOFT_MARGIN_LOSS);
  printf("     %2d -- Huber-hinge             %d -- Logistic\n",HUBER_HINGE_LOSS, LOGISTIC_LOSS);
  printf("     %2d -- Exponential             %d -- ROC Score (aka AUC)\n",EXPONENTIAL_LOSS, ROC_SCORE_LOSS);
  printf("     %2d -- F_beta Score\n",FBETA_SCORE_LOSS);
  printf("   Univariate Regression:\n");
  printf("     %2d -- Epsilon insensitive     %d -- Huber robust\n",ESVR_LOSS, HUBER_LOSS);
  printf("     %2d -- Least Mean Squares      %d -- Least Absolute Deviation\n",LEAST_SQUARES_LOSS, LEAST_DEVIATION_LOSS);
  printf("   Other scalar loss:\n");
  printf("     %2d -- Quantile regression\n",QUANTILE_LOSS);
  printf("     %2d -- Novelty detection (1-Class SVM)\n",NOVELTY_LOSS);
  printf("     %2d -- Poisson Regression\n",POISSON_LOSS);
  printf("   Ranking:\n");
  printf("     %2d -- NDCG (normalized discounted cummulative gain) ranking\n",NDCG_LOSS);
  printf("     %2d -- Ordinal regression\n",ORDINAL_REGRESSION_LOSS);
}




/** Print usage message for training
 */
void PrintTrainUsageMessage()
{
  printf("==============================================================\n");
  printf("=     Bundle Methods for Risk Minimisation (Version 1.0)     =\n");
  printf("==============================================================\n");
  printf("\n");
  printf("Usage: bmrm-train [options] [model_file]\n");
  printf("options:\n");
  printf("-s [1..%2d] : Loss function (default 1)\n",NUMBER_OF_LOSS);
  PrintLossOptions();
  printf("-r [1..%d] :  Inner Solver (default: 1)\n", NUMBER_OF_SOLVER);
  printf("    %d  -- Dai-Fletcher Gradient Projection Method (regularizer: 0.5*|W|^2)\n", DAIFLETCHER);
  printf("    %d  -- Quadratic Programming LOQO (regularizer: 0.5*|W|^2)\n", QPLOQO);
  printf("    %d  -- 1-norm Linear Programming (regularizer: |W|)\n", L1LP);
  printf("-m int    : Maximum gradient set size\n");
  printf("-c float  : Slack trade off parameter (default: 1)\n");
  printf("-l float  : Regularizer scaling factor (default: 1)\n");  
  printf("-p float  : Epsilon for epsilon-SVR (default: 0.1)\n");
  printf("-h [0,1]  : Use shifted hyperplane ? (default: 0 (No))\n");
  printf("-a float  : Tau for HuberHinge and QuantileRegression (default: 0.1)\n");
  printf("-t int    : Truncation for NDCG ranking (default: 10)\n");
  printf("-{g,e} float : Termination criterion (default: -g)\n");
  printf("    -g : see TeoLeSmolaViswanathan '06 (default: 0.01)\n");
  printf("    -e : see Joachims '06 (default: 0.1)\n");
  printf("-i int    : Maximum QP iteration (default 2500)\n");
  printf("-v [0..3] : Verbosity level (default 1)\n");
  printf("-T [1..3] : Choose data matrix storage type {sparse,dense} (default: 1)\n");
  printf("    1  -- BMRM determines (automatic)\n");
  printf("    2  -- Sparse (forced)\n");
  printf("    3  -- Dense (forced)\n");  
  printf("-X string : Data file\n");
  printf("-Y string : Label file\n");
  printf("-D string : Distributed dataset file list (not implemented yet)\n");
  printf("-M string : Model file (default: train_file+\".model\")\n");
  printf("-L {0,1}  : Keep convergence log in files? (default: 0:=no)\n");
  printf("\n");

}



/** Print usage message for prediction
 */
void PrintPredictUsageMessage()
{
  printf("==============================================================\n");
  printf("=     Bundle Methods for Risk Minimisation (Version 1.0)     =\n");
  printf("==============================================================\n");
  printf("\n");
  printf("Usage: bmrm-predict [options] -M model_file -X test_data_file [-Y test_label_file] [-O output_file]\n");
  printf("options:\n");
  printf("-s [0..%2d] : Loss function: (default 0 : pick loss function stated in model file)\n",NUMBER_OF_LOSS);
  PrintLossOptions();
  printf("-f [0..3] : Type of outputs: (default 1)\n");
  printf("     %d -- No prediction related output\n", NO_OUTPUT); 
  printf("     %d -- Output predicted labels only\n", LABEL_ONLY); 
  printf("     %d -- Output pre-decision function values (i.e., f := Xw) only\n", FVAL_ONLY); 
  printf("     %d -- Output predicted labels and pre-decision function values\n", LABEL_AND_FVAL); 
  printf("-X string : Data file\n");
  printf("-Y string : Label file\n");
  printf("-M string : Model file\n");
  printf("-O string : Output file\n");
  printf("     * Default prediction labels output filename <output_file> := <test_data_file>.prediction\n");
  printf("     * Default pre-decision function values output filename <output_file> := <test_data_file>.prediction.fval\n");
  printf("     * when \"-f %d\" option is used, function values will be stored in the file <output_file>.fval", LABEL_AND_FVAL);
  printf("\n");
}


/** Write model parameter into file defined in info.
 *
 *  \param w    [read] solution vector
 *  \param info [read] learning and miscllaneous parameters
 *  \param _fn  [read] optional filename which supersedes info::modelFn
 */
ErrorCode WriteModelFile(CMatVecArr* w, CInfo* info, string _fn)
{  
  ofstream ofp;
  string   fn = _fn;

  // open file
  if(fn == "")
    fn = info->modelFn;

  ofp.open(fn.c_str());
  
  if(!ofp.good()) {
    printf("\nERROR: {WriteModelFile()} unable to open file <%s> !\n", fn.c_str());
    return FILEOPENERROR;
  }

  ofp.setf(ios::fixed);

  // write:
  // 1. type of loss function
  // 2. vector dimension i.e., row and col
  // 3. shifted_hyperplane flag {0,1} for {no, yes}
  // 4. content
  ofp << info->learningType << "  #loss function ID" << endl; 
  ofp << w->row << "  #dimensionality of model parameter" << endl;
  ofp << w->col << "  #number of chunks in w i.e. number of hyperplane or the dimensionality of the label in multivariate regression" << endl;
  ofp << info->bias << "  #applied shifted hyperplane?" << endl;

  int n = w->row * w->col;
  for(int i=0; i < n; i++) {
    ofp << setprecision(12) << w->array[i] << endl;
  }

  ofp.close();
  return NOERROR;
}


/** Read model parameter from file.
 *
 *  \param w    [write] Trained model parameter
 *  \param info [r/w]   BMRM information
 *  \param _fn  [read]  Optional model filename
 */
ErrorCode ReadModelFile(CMatVecArr* &w, CInfo* info, string _fn)
{  
  ifstream ifp;
  string   fn = _fn;
  string   line;
  int      row, col, bias;
  Real     *tmp;
  Real     dummy;


  // open file
  if(fn == "")
    fn = info->modelFn;
  
  ifp.open(fn.c_str());

  if(!ifp.good()) {
    printf("\nERROR: {ReadModelFile()} unable to open file <%s> !\n", fn.c_str());
    return FILEOPENERROR;
  }


  // read:
  // 1. type of loss function
  // 2. vector dimension 
  // 3. content
  getline(ifp,line);
  sscanf(line.c_str(), "%d ", &(info->learningType));  // loss function

  getline(ifp,line);
  sscanf(line.c_str(), "%d ", &row);   // the dimensionality of w

  getline(ifp,line);
  sscanf(line.c_str(), "%d ", &col);   // the # of such w

  getline(ifp,line);
  sscanf(line.c_str(), "%d ", &bias);   // was the model trained with added feature for shifted hyperplane?
  assert(bias==1 || bias==0);

  // create w  
  // check for nasty cases: dimensionality of X in test set and model parameter do not match! 
  // if dim(X) > dim(w) we extend w, otherwise we truncate w
  
  int counter = 0;
  int actualDimW = info->dimW - info->bias;
  int actualRow  = row - bias;
  
  if(info->dimW == 0) {
    // this is not in prediction mode, so, we read model and set some parameters in info
    tmp = (Real*)calloc(row*col, sizeof(Real));
    for(int i=0; i<row*col; i++)
      ifp >> tmp[i];
    
    w = new CMatVecArr(row, col, tmp);
    
    info->dimW = row;
    info->bias = bias;
  }
  else if(actualDimW > actualRow) {
    // extend w
    cout << "WARNING: Dimensionality of model and test feature vectors do not match! ";
    cout << "Model parameter extended." << endl << endl;

    int extension = actualDimW - actualRow;
    tmp = (Real*)calloc(info->dimW*col, sizeof(Real));

    for(int i=0; i<col; i++) {
      // reading w from file
      for(int j=0; j<row-bias; j++) {
	ifp >> tmp[counter];
	counter++;
      }
      
      // extending w with zeros
      for(int p=0; p<extension; p++) {
	tmp[counter] = 0;
	counter++;
      }
      
      // insert bias as the last element of each column
      if(bias && info->bias) 
	ifp >> tmp[counter++];
      else if(bias)
	ifp >> dummy;
      else if(info->bias)
	tmp[counter++] = 0;
    }
    w = new CMatVecArr(info->dimW, col, tmp);    
  }
  else if(actualDimW < actualRow) {
    // truncate w
    cout << "WARNING: Dimensionality of model and test feature vectors do not match! ";
    cout << "Model parameter truncated." << endl << endl;;

    int truncation = actualRow - actualDimW;
    tmp = (Real*)calloc(info->dimW*col, sizeof(Real));

    for(int i=0; i<col; i++) {
      // reading w from file
      for(int j=0; j<row-bias-truncation; j++) {
	ifp >> tmp[counter];
	counter++;
      }
      
      // ignoring #truncation# values
      for(int p=0; p<truncation; p++) {
	ifp >> dummy;
      }
      
      // insert bias as the last element of each column
      if(bias && info->bias) 
	ifp >> tmp[counter++];
      else if(bias)
	ifp >> dummy;
      else if(info->bias)
	tmp[counter++] = 0;
    }
    w = new CMatVecArr(info->dimW, col, tmp);   
  }
  else {
    // dimX == dimW, so we do nothing special
    tmp = (Real*)calloc(info->dimW*col, sizeof(Real));
    for(int i=0; i<col; i++) {
      
      for(int j=0; j<row; j++) {
	ifp >> tmp[counter];
	counter++;
      }
      
      // insert bias as the last element of each column
      if(bias && info->bias) 
	ifp >> tmp[counter++];
      else if(bias)
	ifp >> dummy;
      else if(info->bias)
	tmp[counter++] = 0;
    }
    w = new CMatVecArr(info->dimW, col, tmp);
  }

  // update info
  info->k = col;


  // clean up
  if(tmp) free(tmp);
  ifp.close();

  return NOERROR;
}


#endif
