/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 12/04/2007
 *
 * Last Updated :
 */

#ifndef _FBETASCORELOSS_CPP_
#define _FBETASCORELOSS_CPP_


#include "fbetascoreloss.hpp"
#include <ext/numeric>         // for iota


/**  Constructor
 *   Assumption:
 *   1. dataset shown to constructor must be used in the optimization
 */
CFBetaScoreLoss::CFBetaScoreLoss(CInfo *info, CDataset *ds) 
  :CScalarClassificationLoss(info, ds),
   idx_pn(0),
   p(0),
   n(0),
   m_neg(0),
   m_pos(0),
   beta(1),
   beta_sq(1)
{
  // count number of -ve and +ve examples
  int m = ds->nExample;
  for(int i=0; i<m; i++) if(ds->Y.STL_vec[i] > 0)  m_pos++;
  m_neg = m - m_pos;

  // initialize indices
  idx.reserve(m);
  idx.resize(m);

  // initialize pn vector
  idx_pn.reserve(m);
  idx_pn.resize(m);
  
  // the beta parameter
  beta = info->epsilon;
  beta_sq = SQR(beta);
}



/** Label loss defined on contingency table : F_beta
 *  value = (1 + beta^2) * TP / ((1 + beta^2) * TP + FP + FN*beta^2)
 *
 *  \param TP    [read]  True positive
 *  \param TN    [read]  True negative
 *  \param FP    [read]  False positive
 *  \param FN    [read]  False negative
 *  \param value [write] Computed label loss value
 */
inline void CFBetaScoreLoss::delta(const int &TP, const int &TN, const int &FP, const int &FN, Real &value)
{
  value = 100.0 - 100.0 * (1.0 + beta_sq) * TP / ((1.0 + beta_sq) * TP + FP + FN * beta_sq);
}


/**  Compute loss. 
 *   
 *   \param m    [read]  Dimensionality of f and y (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 */
void CFBetaScoreLoss::compute_loss(int m, Real *f, Real *y, Real &loss)
{
  // reuse the MVA_g vector in CScalarLoss
  compute_loss_gradient(m,f,y,loss,MVA_g->array);
}



/**  Compute loss and gradient together. 
 *   
 *   \param m    [read]  Dimensionality of f, y and g (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 *   \param l    [write] gradient of loss w.r.t. f
 */
void CFBetaScoreLoss::compute_loss_gradient(int m, Real *f, Real *y, Real &loss, Real *l)
{
  // initialization
  register int index = 0;
  register int p_idx = 0;
  register int n_idx = m-1;
  register int i = 0;
  register int j = 0;
  Real tmp_loss = 0;
  indirect_less_than ilt(f);      // functor  
  Real sum_f_pos = 0;             // sum of f_i where y_i = 1
  Real sum_f_neg = 0;             // sum of f_i where y_i = -1
  Real prev_sum_f_pos = 0;        // changing value of sum_f_pos
  Real prev_sum_f_neg = 0;        // changing value of sum_f_neg
  int i_best = 0;
  int j_best = 0;
  Real label_loss = 0;

  loss = 0;
  memset(l,0,sizeof(Real)*m);

  // indices sorting, scale vector l, compute loss
  iota(idx.begin(), idx.end(), 0);
  sort(idx.begin(), idx.end(), ilt);  // in ascending order

  // compute i^+ and i^- index vectors --> idx_pn: first part is i^+ and the rest i^-. The cutoff is p_pos
  sum_f_pos = 0;
  sum_f_neg = 0;
  for(i=0; i<m; i++) {
    index = idx[i];
    if(y[index] > 0) { idx_pn[p_idx] = index; sum_f_pos += f[index]; p_idx++; }   // i^+ in ascending order
    else             { idx_pn[n_idx] = index; sum_f_neg += f[index]; n_idx--; }   // i^- in descending order
  }
  reverse(idx_pn.begin()+p_idx, idx_pn.end());  // because n_idx were stored in reversed order
  p = &(idx_pn[0]);       // pointer to beginning of i^+
  n = &(idx_pn[p_idx]);   // pointer to beginning of i^-
  
  //for(i=0; i<p_idx; i++)
  //printf("p[%3d] : %3d    y[p[%3d]] : %0.lf   f[p[%3d]] : %.6lf\n",i, p[i], i, y[p[i]], i, f[p[i]]);
  //for(i=0; i<n_idx; i++)
  //printf("n[%3d] : %3d    y[p[%3d]] : %0.lf   f[n[%3d]] : %.6lf\n",i, n[i], i, y[n[i]], i, f[n[i]]);
    
  // start contingency table search
  i_best = 0;
  j_best = 0;
  prev_sum_f_pos = sum_f_pos;
  prev_sum_f_neg = sum_f_neg;
  loss = -INFTY;

  for(i=0; i<m_pos; i++) {
    for(j=0; j<m_neg; j++) {
      delta(m_pos-i, m_neg-j, i, j, label_loss);
      tmp_loss = label_loss + prev_sum_f_pos - prev_sum_f_neg;
      //printf("%.6lf    %.6lf\n",label_loss, prev_sum_f_pos-prev_sum_f_neg);
      prev_sum_f_neg -= 2*f[n[j]];
      if(tmp_loss > loss) {
	loss = tmp_loss;
	i_best = i;
	j_best = j;
      }
      // since the max lable loss is 100
      //      if(prev_sum_f_pos+prev_sum_f_neg < -100.0) {
      // 	printf("early search termination, i:%d  j%d\n", i, j);
      // 	break;
      //       }
    }
    prev_sum_f_pos -= 2*f[p[i]];
    prev_sum_f_neg = sum_f_neg;
  }

  // chteo: another for loop for i=0 and j=0 ...

  // set up vector l (part 1)
  assert(p_idx == m_pos);
  for(i=0; i<m_pos; i++) {
    if(i<=i_best) l[p[i]] = -1;
    else         l[p[i]] = 1;
  }
  for(j=0; j<m_neg; j++) {
    if(j<=j_best) l[n[j]] = 1;
    else         l[n[j]] = -1;
  }
  
  cout << "i_best : " << i_best << "   j_best : " << j_best << endl;

  // compute final loss and set up vector l (part 2)
  for(i=0; i<m; i++) {
    loss -= y[i]*f[i];
    l[i] -= y[i];
  }
}

/*
     Notes:
     1. In Joachims' ICML'05, (a,b,c,d) = (true positive, false positive, false negative, true positive)

 */

#endif
