/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 13/12/2006
 *
 * Last Updated :
 */

#ifndef _LOGISTICLOSS_CPP_
#define _LOGISTICLOSS_CPP_


#include "logisticloss.hpp"

#define LN2 0.69314718055994530941  // natural logarithm

/**  Compute loss. 
 *   
 *   \param m    [read]  Dimensionality of f and y (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 */
void CLogisticLoss::compute_loss(int m, Real *f, Real *y, Real &loss)
{
  Real yf = 0;
  loss = 0;
  for(int i=0; i<m; i++) {
    yf = y[i]*f[i];
    if(ABS(yf) < ZERO_EPS)
      loss += LN2;
    else if(yf > 0)
      loss += log(1+exp(-yf));
    else // yf < 0
      loss += log(1+exp(yf)) - yf;
      
  }
}



/**  Compute loss and gradient together. 
 *   
 *   \param m    [read]  Dimensionality of f, y and g (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 *   \param g    [write] Gradient computed w.r.t. f (memory pre-allocated dense matrix)
 */
void CLogisticLoss::compute_loss_gradient(int m, Real *f, Real *y, Real &loss, Real *g)
{
  Real yf = 0;
  Real exp_yf = 0;
  loss = 0;

  for(int i=0; i<m; i++) {
    yf = y[i]*f[i];
    
    if(ABS(yf) < ZERO_EPS) {
      loss += LN2;
      g[i] = -y[i]/2.0;
    }
    else if(yf > 0) {
      exp_yf = exp(-yf);
      loss += log(1+exp_yf);
      g[i] = -y[i]*exp_yf/(1+exp_yf);
    }
    else { // yf < 0
      exp_yf = exp(yf);
      loss += log(1+exp_yf) - yf;
      g[i] = -y[i]/(1+exp_yf);
    }
  }
}

#endif
