/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 12/04/2007
 *
 * Last Updated :
 */

#ifndef _NDCGLOSS_CPP_
#define _NDCGLOSS_CPP_


#include "ndcgloss.hpp"
#include <ext/numeric>         // for iota
#include <lap.hpp>

/**  Constructor
 */
CNDCGLoss::CNDCGLoss(CInfo *info, CDataset *ds) 
  :CScalarRankingLoss(info, ds)
{
  counter = 0;
  
  num_query = 0;
  
  truncation = info->truncation;
  
  int m = ds->nSample;
  
  
  for (int i=0;i<m;i++){
    int length = ds->sample[i].len;
    int index = ds->sample[i].firstIdx;
    vector<int> indices(length);
    iota(indices.begin(), indices.end(), 0);
    indirect_greater_than igt(&(ds->Y.STL_vec[index]));
    sort(indices.begin(), indices.end(),igt);
    sort_vectors.push_back(indices);
    num_query++;
  }
  
}

void
CNDCGLoss::dcg(int m, Real *y, Real &dy)
{
  int i = 0;
  Real d = 0.0;
  for (i=0;i<truncation&&i<m;i++){
    d += (pow(2.0,get(y,i))-1)/log(i+2.0);
  }
  dy = d;
}

void
CNDCGLoss::compute_coefficients(int m, Real *y, Real *a, Real *b){
  Real dy;
  dcg(m, y, dy);
  for (int i=0;i<m;i++){
    if (i<truncation){
      a[i] = 1.0/log(i+2.0);
    }
    else{
      a[i] = 0.0;
    }
    if (dy <= 0){
      b[i] = 1;
    }else{
      b[i] = (pow(2.0,get(y,i))-1.0)/dy;    
    }
  }
}

void
CNDCGLoss::delta(int m, double *a, double *b, int *pi, double &value){
  value = 0;
  for (int i=0;i<m;i++){
    value += (a[i] - a[pi[i]])*b[i];
  }
}

void
CNDCGLoss::find_permutation(int m, Real *a, Real *b, Real *c, Real *f, int *pi){
  /* setting up C_ij */
  Real **C = (Real **)malloc(sizeof(Real*)*m);
  for (int i=0;i<m;i++){
    C[i] = (Real *)malloc(sizeof(Real)*m);
  }
  for (int i=0;i<m;i++){
    for (int j=0;j<m;j++){
      C[i][j] = -c[i]*get(f,j) + b[i]*a[j];
    }
  }
 
  int *row = (int *)malloc(sizeof(int)*m);
  
  lap(m,C,pi,row);  
  printf(".");fflush(stdout);

  free(row);
  for (int i=0;i<m;i++){
    free(C[i]);
  }
  free(C);
}


/**  Compute loss. 
 *   
 *   \param m    [read]  Dimensionality of f and y (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 */
void CNDCGLoss::compute_loss(int m, Real *f, Real *y, Real &loss)
{
  // reuse the MVA_g vector in CScalarLoss
  compute_loss_gradient(m,f,y,loss,MVA_g->array);
}



/**  Compute loss and gradient together. 
 *   
 *   \param m    [read]  Dimensionality of f, y and g (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 *   \param l    [write] The vector l used to store information for gradient computation
 */
void CNDCGLoss::compute_loss_gradient(int m, Real *f, Real *y, Real &loss, Real *l)
{

  sort_index = counter%num_query;
  current = sort_vectors[sort_index];
  
  int *pi = (int *) malloc(sizeof(int)*m);
  
  /* setting up c_i */
  Real *c = (Real *)malloc(sizeof(Real)*m);
  for (int i=0;i<m;i++){
    //c[i] = 1/((ex[i]->Y+1)*sqrt(i+1.0)); /* this is one option */
    //c[i] = (ex[i]->Y)/(i+1.0);
    //c[i] = ex[i]->Y;
    c[i] = 1.0/pow((i+1.0),1.4);
  }
  
  /* setting up a_i, b_i */ 
  Real *a = (Real *)malloc(sizeof(Real)*m);
  Real *b = (Real *)malloc(sizeof(Real)*m);
  compute_coefficients(m, y, a, b);
  
  /* find the best permutation */
  find_permutation(m, a, b, c, f, pi);
  
  /* compute the loss */
  Real value;
  delta(m, a, b, pi, value);
  loss = value;
  
  for (int i=0;i<m;i++){
    loss = loss - c[i]*(get(f,i) - get(f,pi[i]));
    l[i] = 0;
  }
  
  for (int i=0;i<m;i++){
    add(l, i, -c[i]);
    add(l, pi[i], c[i]);
  }
  free(c);
  free(a);
  free(b);
  free(pi);
  counter++;
}

double CNDCGLoss::get(Real *y, int i){
  return y[current[i]];
}

void CNDCGLoss::add(Real *y, int i, Real value){
  y[current[i]] += value;
}

#endif
