/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 12/04/2007
 *
 * Last Updated :
 */

#ifndef _ORDINALREGRESSIONLOSS_CPP_
#define _ORDINALREGRESSIONLOSS_CPP_


#include "ordinalregressionloss.hpp"
#include <ext/numeric>         // for iota


/**  Constructor
 */
COrdinalRegressionLoss::COrdinalRegressionLoss(CInfo *info, CDataset *ds) 
  :CScalarRankingLoss(info, ds)
{
  counter = 0;
  
  num_query = 0;
  
  truncation = info->truncation;
  
  int m = ds->nSample;
  
  
  for (int i=0;i<m;i++){
    /* compute the sorted vectors once */
    int length = ds->sample[i].len;
    int index = ds->sample[i].firstIdx;
    vector<int> indices(length);
    iota(indices.begin(), indices.end(), 0);
    indirect_greater_than igt(&(ds->Y.STL_vec[index]));
    sort(indices.begin(), indices.end(),igt);
    sort_vectors.push_back(indices);
    num_query++;
  }
  
}



/**  Compute loss. 
 *   
 *   \param m    [read]  Dimensionality of f and y (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 */
void COrdinalRegressionLoss::compute_loss(int m, Real *f, Real *y, Real &loss)
{
  // reuse the MVA_g vector in CScalarLoss
  compute_loss_gradient(m,f,y,loss,MVA_g->array);
}



/**  Compute loss and gradient together. 
 *   
 *   \param n    [read]  Dimensionality of f, y and g (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 *   \param g    [write] The vector g used to store information for gradient computation
 */
void COrdinalRegressionLoss::compute_loss_gradient(int m, Real *f, Real *y, Real &loss, Real *g)
{
  
  sort_index = counter%num_query;
  current = sort_vectors[sort_index];
  

  int maxy = 0;
  for (int i=0;i<m;i++){
    if (y[i]>maxy){
      maxy = (int)y[i];
    }
  }
  int n = maxy + 1;
  vector<int> u(n);
  vector<int> l(n);
  for (int i=0;i<n;i++){
    l[i] = 0;
    u[i] = 0;
  }
  for (int j=0;j<m;j++){
    for (int i=0;i<n;i++){
      int myy = (int)y[j];
      if (myy==i){
	u[i] += 1;
	break;
      }
    }
  }
  
  double M = m*m;
  for (int i=0;i<n;i++){
    M -= u[i]*u[i];
  }
  M = 0.5*M;
  if (M == 0.0){
    loss = 0;
    for (int i=0;i<m;i++){
      g[i] = 0;
    }
    
  }
  else{
    /* initialization of loss and gradient vector */
    for (int i=0;i<m;i++){
      g[i] = 0;
    }
    loss = 0;
    
    vector<int> indices(2*m);
    Real *c = (Real *)malloc(sizeof(Real)*2*m);
    for (int i=0;i<m;i++){
      c[i] = get(f,i) - 0.5;
      c[i+m] = get(f,i) + 0.5;
    }
    /* sort the indices with descending order */
    indirect_less_than ilt(c);
    iota(indices.begin(), indices.end(), 0);
    sort(indices.begin(), indices.end(), ilt);

    for (int i=0;i<2*m;i++){
      int j = indices[i]%m;
      int z = (int)get(y,j);
      //int z = (int)y[j];
      if (indices[i] < m){
	for (int k=0;k<z;k++){
	  loss = loss - (z - k)*u[k]*c[j]/M;
	  //g[j] = g[j] - (z - k)*u[k]/M;
	  add(g,j,-(z - k)*u[k]/M);
	}
	l[z] = l[z] + 1;
      }
      else{
	for (int k=z+1;k<n;k++){
	  loss = loss + (k - z)*l[k]*c[j+m]/M;
	  //g[j] = g[j] + (k - z)*l[k]/M;
	  add(g, j, (k - z)*l[k]/M);
	}
	u[z] = u[z] - 1; 
      }
      
    }
    free(c);
    
  }
  counter++;
  
}

double COrdinalRegressionLoss::get(Real *y, int i){
  return y[current[i]];
}

void COrdinalRegressionLoss::add(Real *y, int i, Real value){
  y[current[i]] += value;
}

#endif
