/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 12/04/2007
 *
 * Last Updated :
 */

#ifndef _ROCSCORELOSS_CPP_
#define _ROCSCORELOSS_CPP_


#include "rocscoreloss.hpp"
#include <ext/numeric>         // for iota


/**  Constructor
 */
CROCScoreLoss::CROCScoreLoss(CInfo *info, CDataset *ds) 
 :CScalarClassificationLoss(info, ds),
  m_neg(0),
  m_pos(0),
  l_neg(0),
  l_pos(0),
  scalingfactor(1)
{
  // count number of -ve and +ve examples
  int m = ds->nExample;
  for(int i=0; i<m; i++) if(ds->Y.STL_vec[i] > 0)  m_pos++;
  m_neg = m - m_pos;

  // initialize indices
  idx.reserve(m);
  idx.resize(m);

  // used in loss & grad computation
  //   multiply with m to compensate the division by m done automatically in 
  //   CScalarLoss::ComputeLoss() and ComputeLossGradient()
  scalingfactor = (double)m/(double)m_neg/(double)m_pos;
}


/**  Compute loss. 
 *   
 *   \param m    [read]  Dimensionality of f and y (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 */
void CROCScoreLoss::compute_loss(int m, Real *f, Real *y, Real &loss)
{
  // reuse the MVA_g vector in CScalarLoss
  compute_loss_gradient(m,f,y,loss,MVA_g->array);
}



/**  Compute loss and gradient together. 
 *   
 *   \param m    [read]  Dimensionality of f, y and g (# of examples)
 *   \param f    [read]  f := X w, where X is data matrix and w is current solution
 *   \param y    [read]  labels
 *   \param loss [write] Loss value computed
 *   \param l    [write] The vector l used to store information for gradient computation
 */
void CROCScoreLoss::compute_loss_gradient(int m, Real *f, Real *y, Real &loss, Real *l)
{
  // initialization
  register int i = 0;
  register int index = 0;
  indirect_less_than ilt(f);          // functor

  loss = 0;
  l_pos = 0;
  l_neg = m_neg;
  memset(l,0,sizeof(Real)*m);
  for(i=0; i<m; i++) f[i] -= 0.5*y[i];

  // indices sorting, scale vector l, compute loss
  iota(idx.begin(), idx.end(), 0);
  sort(idx.begin(), idx.end(), ilt);  // in ascending order
  
  for(i=0; i<m; i++) {
    index = idx[i];
    if(y[index] < 0) {
      l[index] = l_pos;
      l_neg -= 1;
    }
    else {
      l[index] = -l_neg;
      l_pos += 1;
    }
    l[index] *= scalingfactor;
    loss += l[index] * f[index];
  }
}

#endif
