/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 10/04/2007
 *
 * Last Updated :
 */

#ifndef _SCALARCLASSIFICATIONLOSS_CPP_
#define _SCALARCLASSIFICATIONLOSS_CPP_

#include "scalarclassificationloss.hpp"
#include <cassert>
#include <ext/numeric>






/** Prediction function.
 *
 *  \param w     [read]  Model parameter
 *  \param ds    [read]  Dataset
 *  \param ybar  [write] Predicted best labels (memory pre-allocated)
 *  \param arg_f [write] arg_f := X*w, where X is data matrix (memory pre-allocated)
 */
inline void CScalarClassificationLoss::predict(CMatVecArr *w, CDataset *ds, Real *ybar, Real *arg_f)
{
  Real *f = arg_f;
  if(f == 0) f = MVA_f->array;   // reuse CScalarLoss::MVA_f
  int m = ds->nExample;

  x_dot_w(w, ds, f);
  for(int i=0; i<m; i++) 
    ybar[i] = SGN(f[i]);
}


/** Evaluate performance of a model w.r.t. this loss function (and/or others)
 *  Display the performance measures here.
 *  
 *  \param m [read]  Number of examples
 *  \param f [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 *  \param y [read]  True labels
 *
 */
inline void CScalarClassificationLoss::eval_performance(int m, Real *f, Real *y)
{
  vector<string> metric;
  vector<double> score;

  eval_performance(m,f,y,metric,score);
 
  printf("\n");
  for(size_t i=0; i<metric.size(); i++)
    printf("%24s : %.2lf\n", metric[i].c_str(), score[i]);
  printf("\n");
  
  /* notes:
     1. precision and recall are only good measures when the classes are balanced
   */
}


/** Evaluate performance of a model w.r.t. this loss function (and/or others)
 *
 *  \param m      [read]  Number of examples
 *  \param f      [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 *  \param y      [read]  True labels
 *  \param metric [write] Type of metric applied in the evaluation (optional)
 *  \param score  [write] The score for the corresponding metric in #metric# (optinal)
 *
 */
inline void CScalarClassificationLoss::eval_performance(int m, Real *f, Real *y, vector<string> &metric, vector<double> &score)
{
  assert(m > 0);
  assert(f != 0);
  assert(y != 0);

  int TP = 0;    // true positive
  int TN = 0;    // true negative
  int FP = 0;    // false positive
  int FN = 0;    // false negative

  // compute TP, TN, FP, FN (i.e. contingency table)
  for(int i=0; i<m; i++) {
    Real predict = SGN(f[i]);
    Real desired = y[i];

    if(predict > 0 && desired > 0)        TP++;
    else if(predict > 0 && desired <= 0)  FP++;
    else if(predict <= 0 && desired <= 0) TN++;
    else                                  FN++;
  }
  
  // performance measures
  double total = TP+TN+FP+FN;
  double precision = (TP>0)? 100.0*TP/(TP+FP) : 0.0;
  double recall = (TP>2) ? 100.0*TP/(TP+FN) : 0.0;
  double fone = ((precision+recall) > 0) ? 2.0*precision*recall/(precision+recall) : 0.0;
  
  /// for auc : START
  double auc = 0;
  vector<int> idx(m);
  iota(idx.begin(), idx.end(), 0);

  indirect_less_than ilt(f);
  sort(idx.begin(), idx.end(), ilt);  // ascending order
  
  //// count swapped pairs i.e., misclassified points
  int n_pos = 0, n_neg = 0, sum = 0;
  for(int i=0; i<m; i++) {
    if(y[idx[i]] < 0) {
      sum += n_pos;
      n_neg++;
    }
    else
      n_pos++;
  }
  auc = 100.0 - 100.0*sum/n_neg/n_pos;
  /// END

  metric.push_back("Class. accuracy (%)");
  score.push_back(100.0*(TP+TN)/(total));

  metric.push_back("Precision (%)");
  score.push_back(precision);

  metric.push_back("Recall (%)");
  score.push_back(recall);
  
  metric.push_back("F1 score");
  score.push_back(fone);

  metric.push_back("Area under ROC (%)");
  score.push_back(auc);

  metric.push_back("True Positive");
  score.push_back(TP);
  
  metric.push_back("True Negative");
  score.push_back(TN);

  metric.push_back("False Positive");
  score.push_back(FP);
  
  metric.push_back("False Negative");
  score.push_back(FN);


  

  /* notes:
     1. precision and recall are only good measures when the classes are balanced
   */
}


/** Implementation of parent's prediction method.
 *  (see predict() for details)
 *
 *  \param w     [read]  Model parameter
 *  \param ds    [read]  Dataset
 *  \param ybar  [write] Predicted best labels
 *  \param arg_f [write] arg_f := X*w, where X is data matrix (memory pre-allocated)
 */
ErrorCode
CScalarClassificationLoss::Predict(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *arg_f)
{
  ybar->STL_vec.resize(ds->nExample);
  predict(w, ds, &(ybar->STL_vec[0]), arg_f->array);
  return NOERROR;
}


/** Implementation of parent's performance evaluation method.
 *  (see eval_performance() for details)
 *
 *  \param w      [read]  Model parameter
 *  \param ds     [read]  Dataset 
 *  \param ybar   [read]  Predicted labels (post-decision function values)
 *  \param f      [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 *  \param metric [write] Type of metric applied in the evaluation (optional)
 *  \param score  [write] The score for the corresponding metric in #metric# (optinal)
 */
ErrorCode
CScalarClassificationLoss::EvalPerformance(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *f, vector<string> &metric, vector<double> &score)
{
  eval_performance(ds->nExample, f->array, &(ds->Y.STL_vec[0]), metric, score);
  return NOERROR;
}


/** Implementation of parent's performance evaluation method.
 *  (see eval_performance() for details)
 *
 *  \param w    [read]  Model parameter
 *  \param ds   [read]  Dataset 
 *  \param ybar [read]  Predicted labels (post-decision function values)
 *  \param f    [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 */
ErrorCode
CScalarClassificationLoss::EvalPerformance(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *f)
{
  eval_performance(ds->nExample, f->array, &(ds->Y.STL_vec[0]));
  return NOERROR;
}


#endif
