/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 13/12/2006
 *
 * Last Updated :
 */

#ifndef _SCALARLOSS_CPP_
#define _SCALARLOSS_CPP_

#include "scalarloss.hpp"


CScalarLoss::CScalarLoss(CInfo *info, CDataset *ds) : CLoss(info, ds)
{
  // create MVA_f := Xw
  MVA_f = new CMatVecArr(ds->nExample, 1);

  // create MVA_g := \partial_f loss
  MVA_g = new CMatVecArr(ds->nExample, 1);


  //MatCreateSeqAIJ(PETSC_COMM_SELF, MVA_g->col, MVA_g->row, MVA_g->col*MVA_g->row, PETSC_NULL, &sparse_gT);
}


CScalarLoss::~CScalarLoss()
{
  if(MVA_f) { delete MVA_f; MVA_f = 0; }
  if(MVA_g) { delete MVA_g; MVA_g = 0; }
  //MatDestroy(sparse_gT);
}


/**  Compute loss. 
 *   (Used PETSc matrix operation to speed up the dot-products.)
 *   
 *   \param w    [read]  Current solution
 *   \param ds   [read]  Dataset
 *   \param loss [write] Loss value computed
 */
ErrorCode
CScalarLoss::ComputeLoss(CMatVecArr *w, CDataset *ds, Real &loss)
{
  // In this case, dataset contains 1 sample which, in turn, contains many examples
  // We make use of the PETSc matrix view of the dataset to do <w,x_i> \forall i
  //   in a more efficient way.
  
  // F := X*w, where X is a matrix of rows of examples.
  ierr = MatMult(ds->X, w->vec, MVA_f->vec); CHKERRQ(ierr);
    
  // now we have f, we can compute the loss.
  // some arguments are redundant but we leave it there so that the context is clearer.
  compute_loss(ds->nExample, MVA_f->array, &(ds->Y.STL_vec[0]), loss);

  // scaling by the number of examples in the WHOLE dataset
  loss /= ds->totalExample;

  return NOERROR;
}



/**  Compute loss and gradient together. 
 *   (Used PETSc matrix operation to speed up the dot-products.)
 *   
 *   \param w    [read]  Current solution
 *   \param ds   [read]  Dataset
 *   \param loss [write] Loss value computed
 *   \param grad [write] Gradient computed (memory pre-allocated dense matrix)
 */

ErrorCode
CScalarLoss::ComputeLossGradient(CMatVecArr *w, CDataset *ds, Real &loss, CMatVecArr *grad)
{
  // In this case, dataset contains 1 sample which, in turn, contains many examples
  // We make use of the PETSc matrix view of the dataset to do <w,x_i> \forall i
  //   in a more efficient way.
    
  // f := X*w, where X is a matrix of rows of examples.
  // Note that dense PETSc matrix is stored in column-major-order,
  //   hence, the data matrix is transposed. Sparse PETSc matrix is fine.

  rightmult.Start();
  ierr = MatMult(ds->X, w->vec, MVA_f->vec); CHKERRQ(ierr);
  rightmult.Stop();

  lossgradient.Start();
  // now we have f. we can compute the loss and gradient w.r.t. f
  compute_loss_gradient(ds->nExample, MVA_f->array, &(ds->Y.STL_vec[0]), loss, MVA_g->array);
  lossgradient.Stop();

  // gradient grad = g'*X

  leftmult.Start();
    
  //
  //  [250407] chteo: BUGBUG
  //  MatMatMult requires A and B to be of compatible MatType! But we have g as dense
  //  while ds-> as sparse in most of the cases...
  //
  //MatConvert(MVA_g->matT, MATSEQAIJ, MAT_REUSE_MATRIX, &sparse_gT);
  //MatCopy(MVA_g->matT, sparse_gT, DIFFERENT_NONZERO_PATTERN);
  //MatMatMult(sparse_gT, ds->X, MAT_INITIAL_MATRIX, ds->dimX, &grad->mat); 

 
  MatMultTranspose(ds->X, MVA_g->vec, grad->vec);  
  leftmult.Stop();

  if(_info->verbosity > 1) {
    printf("left matmult  : %lf\n",leftmult.Total());
    printf("right matmult : %lf\n",rightmult.Total());
    printf("loss & grad   : %lf\n",lossgradient.Total());
  }

  // scaling by the number of examples in the WHOLE dataset
  loss /= ds->totalExample; 
  VecScale(grad->vec, 1.0/ds->totalExample);

  return NOERROR;
}


/**  Perform matrix vector multiplication. For the given dataset _ds, and solution w.
 *
 *   \param w  [read]  Solution (i.e. parameter)
 *   \param ds [read]  Dataset
 *   \param f  [write] Xw =: f (memory pre-allocated)
 */
void CScalarLoss::x_dot_w(CMatVecArr *w, CDataset *ds, Real *f)
{
  MatMult(ds->X, w->vec, MVA_f->vec);
  if(f != MVA_f->array)
    memcpy(f, MVA_f->array, sizeof(Real)*MVA_f->row);
}


#endif
