/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 10/04/2007
 *
 * Last Updated :
 */

#ifndef _SCALARRANKINGLOSS_CPP_
#define _SCALARRANKINGLOSS_CPP_

#include "scalarrankingloss.hpp"
#include <cassert>
#include <ext/numeric>
#include <statistics.hpp>


inline void CScalarRankingLoss::eval_ndcg(int m, Real *f, Real *y, double &ndcg)
{
  int truncation = 10;
  vector<int> indices1(m);
  vector<int> indices2(m);
  iota(indices1.begin(), indices1.end(), 0);
  iota(indices2.begin(), indices2.end(), 0);
  
  indirect_greater_than igt1(y);
  sort(indices1.begin(), indices1.end(), igt1);  // descending order
  
  indirect_greater_than igt2(f);
  sort(indices2.begin(), indices2.end(), igt2);  // descending order
  
  double dy = 0;
  double df = 0;
  
  for (int i=0;i<m&&i<truncation;i++){
    dy += (pow(2.0, y[indices1[i]])-1)/log(i+2.0);
    df += (pow(2.0, y[indices2[i]])-1)/log(i+2.0);
  }
  if (dy <= 0){
    ndcg = 1.0;
  }
  else{
    ndcg = 1.0*df/dy;
  }

}


/** Prediction function.
 *
 *  \param w     [read]  Model parameter
 *  \param ds    [read]  Dataset
 *  \param ybar  [write] Predicted best labels (memory pre-allocated)
 *  \param arg_f [write] arg_f := X*w, where X is data matrix (memory pre-allocated)
 */
inline void CScalarRankingLoss::predict(CMatVecArr *w, CDataset *ds, Real *ybar, Real *arg_f)
{
  Real *f = arg_f;
  if(f == 0) f = MVA_f->array;   // reuse CScalarLoss::MVA_f
  int m = ds->nExample;

  x_dot_w(w, ds, f);
  for(int i=0; i<m; i++) 
    ybar[i] = SGN(f[i]);
}




/** Implementation of parent's prediction method.
 *  (see predict() for details)
 *
 *  \param w     [read]  Model parameter
 *  \param ds    [read]  Dataset
 *  \param ybar  [write] Predicted best labels
 *  \param arg_f [write] arg_f := X*w, where X is data matrix (memory pre-allocated)
 */
ErrorCode
CScalarRankingLoss::Predict(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *arg_f)
{
  ybar->STL_vec.reserve(ds->nExample);
  ybar->STL_vec.resize(ds->nExample);
  predict(w, ds, &(ybar->STL_vec[0]), arg_f->array);
  return NOERROR;
}


/** Implementation of parent's performance evaluation method.
 *  (see eval_performance() for details)
 *
 *  \param w      [read]  Model parameter
 *  \param ds     [read]  Dataset 
 *  \param ybar   [read]  Predicted labels (post-decision function values)
 *  \param f      [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 *  \param metric [write] Type of metric applied in the evaluation (optional)
 *  \param score  [write] The score for the corresponding metric in #metric# (optinal)
 */
ErrorCode
CScalarRankingLoss::EvalPerformance(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *f, vector<string> &metric, vector<double> &score)
{
  double ndcg;
  //double mrr;
  //double wta;
  CStatistics ndcg_stats = CStatistics();
  //CStatistics mrr_stats = CStatistics();
  //CStatistics wta_stats = CStatistics();
  int m = ds->nSample;
  
  for (int i=0;i<m;i++){
    int length = ds->sample[i].len;
    int index = ds->sample[i].firstIdx;
    eval_ndcg(length, &(f->array[index]), &(ds->Y.STL_vec[index]), ndcg);
    //eval_mrr(ds->sample[i]->len, f->array[ds->sample[i]->firstIdx], ybar, mrr);
    //eval_wta(ds->sample[i]->len, f->array[ds->sample[i]->firstIdx], ybar, mrr);
    
    ndcg_stats.Add(ndcg);
    //mrr_stats.Add(mrr);
    //wta_stats.Add(wta);
  }
  //metric.push_back("ndcg");
  //score.push_back(ndcg_stats.GetMean());
  printf("ndcg mean : %.2f\n",100*ndcg_stats.GetMean());
  printf("ndcg std  : %.2f\n",100*ndcg_stats.GetStandardDeviation());
  
  return NOERROR;
}


/** Implementation of parent's performance evaluation method.
 *  (see eval_performance() for details)
 *
 *  \param w    [read]  Model parameter
 *  \param ds   [read]  Dataset 
 *  \param ybar [read]  Predicted labels (post-decision function values)
 *  \param f    [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 */
ErrorCode
CScalarRankingLoss::EvalPerformance(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *f)
{
  double ndcg;
  //double mrr;
  //double wta;
  CStatistics ndcg_stats = CStatistics();
  //CStatistics mrr_stats = CStatistics();
  //CStatistics wta_stats = CStatistics();
  
  int m = ds->nSample;
  for (int i=0;i<m;i++){
    int length = ds->sample[i].len;
    int index = ds->sample[i].firstIdx;
    eval_ndcg(length, &(f->array[index]), &(ds->Y.STL_vec[index]), ndcg);
    //eval_mrr(ds->sample[i]->len, f->array[ds->sample[i]->firstIdx], ybar, mrr);
    //eval_wta(ds->sample[i]->len, f->array[ds->sample[i]->firstIdx], ybar, mrr);
    
    ndcg_stats.Add(ndcg);
    //mrr_stats.Add(mrr);
    //wta_stats.Add(wta);
  }
  //metric.push_back("ndcg");
  //score.push_back(ndcg_stats.GetMean());
  printf("ndcg mean : %.2f\n",100*ndcg_stats.GetMean());
  printf("ndcg std  : %.2f\n",100*ndcg_stats.GetStandardDeviation());
  return NOERROR;
}


ErrorCode
CScalarRankingLoss::ComputeLoss(CMatVecArr *w, CDataset *ds, Real &loss){
        // In this case, dataset contains 1 sample which, in turn, contains many examples
  // We make use of the PETSc matrix view of the dataset to do <w,x_i> \forall i
  //   in a more efficient way.
    
  // f := X*w, where X is a matrix of rows of examples.
  // Note that dense PETSc matrix is stored in column-major-order,
  //   hence, the data matrix is transposed. Sparse PETSc matrix is fine.

  rightmult.Start();
  ierr = MatMult(ds->X, w->vec, MVA_f->vec); CHKERRQ(ierr);
  rightmult.Stop();

  lossgradient.Start();
  // now we have f. we can compute the loss and gradient w.r.t. f
  Real value = 0;
  loss = 0;
  int m = ds->nSample;
  for (int i=0;i<m;i++){
    int length = ds->sample[i].len;
    int index = ds->sample[i].firstIdx;
    compute_loss(length, &(MVA_f->array[index]), 
		 &(ds->Y.STL_vec[index]), value);
    loss += value;
  }

  return NOERROR;
}

ErrorCode
CScalarRankingLoss::ComputeLossGradient(CMatVecArr *w, CDataset *ds, Real &loss, CMatVecArr *grad)
{
  // In this case, dataset contains 1 sample which, in turn, contains many examples
  // We make use of the PETSc matrix view of the dataset to do <w,x_i> \forall i
  //   in a more efficient way.
    
  // f := X*w, where X is a matrix of rows of examples.
  // Note that dense PETSc matrix is stored in column-major-order,
  //   hence, the data matrix is transposed. Sparse PETSc matrix is fine.

  rightmult.Start();
  ierr = MatMult(ds->X, w->vec, MVA_f->vec); CHKERRQ(ierr);
  rightmult.Stop();

  lossgradient.Start();
  // now we have f. we can compute the loss and gradient w.r.t. f
  Real value = 0;
  loss = 0;
  int m = ds->nSample;
  //printf("number of samples: %d\n",m);
  lossgradient.Start();
  for (int i=0;i<m;i++){
    int length = ds->sample[i].len;
    //printf("number of docs: %d\n",length);
    int index = ds->sample[i].firstIdx;
    compute_loss_gradient(length, &(MVA_f->array[index]), 
			  &(ds->Y.STL_vec[index]), value, &(MVA_g->array[index]));
    loss += value;
  }
  lossgradient.Stop();
  leftmult.Start();  
  MatMultTranspose(ds->X, MVA_g->vec , grad->vec);  
  leftmult.Stop();

  // gradient grad = g'*X


  if(_info->verbosity > 1) {
    printf("left matmult  : %lf\n",leftmult.Total());
    printf("right matmult : %lf\n",rightmult.Total());
    printf("loss & grad   : %lf\n",lossgradient.Total());
  }
  // scaling
  //loss /= ds->nExample; 
  //VecScale(grad->vec, 1.0/ds->nExample);
  return NOERROR;
}
#endif
