/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 10/04/2007
 *
 * Last Updated :
 */

#ifndef _SCALARREGRESSIONLOSS_CPP_
#define _SCALARREGRESSIONLOSS_CPP_

#include "scalarregressionloss.hpp"
#include <cassert>
#include <cmath>


/** Prediction function.
 *
 *  \param w     [read]  Model parameter
 *  \param ds    [read]  Dataset
 *  \param ybar  [write] Predicted best labels (memory pre-allocated)
 *  \param arg_f [write] arg_f := X*w, where X is data matrix (memory pre-allocated)
 */
inline void CScalarRegressionLoss::predict(CMatVecArr *w, CDataset *ds, Real *ybar, Real *arg_f)
{
  assert(ybar != 0);
  
  Real *f = arg_f;
  if(f == 0) f = MVA_f->array;   // reuse CScalarLoss::MVA_f
  int m = ds->nExample;

  x_dot_w(w, ds, f);
  
  // both ybar and f are the same !
  memcpy(ybar, f, sizeof(Real)*m);
}


/** Evaluate performance of a model w.r.t. this loss function (and/or others)
 *  Display the performance measures here.
 *  
 *  \param m [read]  Number of examples
 *  \param f [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 *  \param y [read]  True labels
 *
 */
inline void CScalarRegressionLoss::eval_performance(int m, Real *f, Real *y)
{
  vector<string> metric;
  vector<double> score;

  eval_performance(m,f,y,metric,score);
 
  printf("\n");
  for(size_t i=0; i<metric.size(); i++)
    printf("%24s : %lf\n", metric[i].c_str(), score[i]);
  printf("\n");
  
  /* notes:
     1. precision and recall are only good measures when the classes are balanced
   */
}


/** Evaluate performance of a model w.r.t. this loss function (and/or others)
 *
 *  \param m      [read]  Number of examples
 *  \param f      [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 *  \param y      [read]  True labels
 *  \param metric [write] Type of metric applied in the evaluation (optional)
 *  \param score  [write] The score for the corresponding metric in #metric# (optinal)
 *
 */
inline void CScalarRegressionLoss::eval_performance(int m, Real *f, Real *y, vector<string> &metric, vector<double> &score)
{
  assert(m > 0);
  assert(f != 0);
  assert(y != 0);

  // compute MSE and RMSE
  double tmp = 0;
  double MSE = 0;
  
  for(int i=0; i<m; i++) {
    tmp = f[i]-y[i];
    MSE += SQR(tmp);
  }
  MSE /= m;
  
  metric.push_back("Mean Squared Error");
  score.push_back(MSE);

  metric.push_back("Root Mean Squared Error");
  score.push_back(sqrt(MSE));

  /* notes:
     1. precision and recall are only good measures when the classes are balanced
   */
}


/** Implementation of parent's prediction method.
 *  (see predict() for details)
 *
 *  \param w     [read]  Model parameter
 *  \param ds    [read]  Dataset
 *  \param ybar  [write] Predicted best labels
 *  \param arg_f [write] arg_f := X*w, where X is data matrix (memory pre-allocated)
 */
ErrorCode
CScalarRegressionLoss::Predict(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *arg_f)
{
  ybar->STL_vec.reserve(ds->nExample);
  ybar->STL_vec.resize(ds->nExample);
  predict(w, ds, &(ybar->STL_vec[0]), arg_f->array);
  return NOERROR;
}


/** Implementation of parent's performance evaluation method.
 *  (see eval_performance() for details)
 *
 *  \param w      [read]  Model parameter
 *  \param ds     [read]  Dataset 
 *  \param ybar   [read]  Predicted labels (post-decision function values)
 *  \param f      [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 *  \param metric [write] Type of metric applied in the evaluation (optional)
 *  \param score  [write] The score for the corresponding metric in #metric# (optinal)
 */
ErrorCode
CScalarRegressionLoss::EvalPerformance(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *f, vector<string> &metric, vector<double> &score)
{
  eval_performance(ds->nExample, f->array, &(ds->Y.STL_vec[0]), metric, score);
  return NOERROR;
}


/** Implementation of parent's performance evaluation method.
 *  (see eval_performance() for details)
 *
 *  \param w    [read]  Model parameter
 *  \param ds   [read]  Dataset 
 *  \param ybar [read]  Predicted labels (post-decision function values)
 *  \param f    [read]  Pre-Decision function values f := Xw, where X is data matrix and w is model parameter
 */
ErrorCode
CScalarRegressionLoss::EvalPerformance(CMatVecArr *w, CDataset *ds, CLabel *ybar, CMatVecArr *f)
{
  eval_performance(ds->nExample, f->array, &(ds->Y.STL_vec[0]));
  return NOERROR;
}


#endif
