/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        

 * Created      : 20/04/2007
 * Last Updated :
 */

#ifndef _CG_CPP_
#define _CG_CPP_


#include "cg.hpp"

using namespace std;



CCG::CCG(CInfo* _info, Real _tol)
  :max_mem_size(_info->memSize)
{
  cout << "CG solver instantiated!" << endl;

  // inherited from CInnerSolver
  iter         = 0;
  info         = _info;
  tol          = _tol;
  verbosity    = info->verbosity;
  n            = 0;                       // dimension of solution (increasing)
  m            = 1;                       // number of eq or ineq constraint  (fixed)

  aSet.resize(max_mem_size);              // prepare space for gradients
  bSet.resize(max_mem_size);              // prepare space for offsets  
  timeStamp.resize(max_mem_size, -1);

 
  // allocate mem for PETSc vec and mat
  X_old_array = (Real*)calloc(max_mem_size+1,sizeof(Real));
  X_array     = (Real*)calloc(max_mem_size+1,sizeof(Real));
  B_array     = (Real*)calloc(max_mem_size+1,sizeof(Real));
  A_array     = (Real*)calloc(SQR(max_mem_size+1),sizeof(Real));

  // construct PETSc vec and mat
  // note that calloc initialize all values of the memory to zero
  VecCreateSeqWithArray(PETSC_COMM_SELF, max_mem_size+1, X_old_array, &X_old); 
  VecCreateSeqWithArray(PETSC_COMM_SELF, max_mem_size+1, X_array, &X); 
  VecCreateSeqWithArray(PETSC_COMM_SELF, max_mem_size+1, B_array, &B); 
  MatCreateSeqDense(PETSC_COMM_SELF, max_mem_size+1, max_mem_size+1, A_array, &A); 
  
  VecAssemblyBegin(X_old); VecAssemblyEnd(X_old);
  VecAssemblyBegin(X); VecAssemblyEnd(X);
  VecAssemblyBegin(B); VecAssemblyEnd(B);
  MatAssemblyBegin(A, MAT_FINAL_ASSEMBLY); MatAssemblyEnd(A, MAT_FINAL_ASSEMBLY);

  // take care of inherited member variables
  Q = 0;             // already a part of A_array
  x = X_old_array+1;     // point to the x-part of X_array
  f = B_array+1;     // point to the f-part pf B_array
  l = 0;             // we will enforce lower bound constaint explicitly
  u = 0;             // we dont have upper bound constraint
  a = 0;             // already a part of A_array
  b = 0;             // already a part of B_array   
  B_array[0] = info->C;

  // some temporaries for MatSetValues
  values  = (Real*)calloc(max_mem_size+1,sizeof(Real));
  indices = (PetscInt*)calloc(max_mem_size+1,sizeof(PetscInt));
  for(int i=0; i<max_mem_size+1; i++)
    indices[i] = i;

  
  // KSP linear solver
  KSPCreate(PETSC_COMM_SELF, &ksp);
  //KSPSetInitialGuessNonzero(ksp, PETSC_TRUE); 
  KSPSetOperators(ksp,A,A,SAME_NONZERO_PATTERN); 
  
  KSPSetTolerances(ksp, 
                   PETSC_DEFAULT,   // Relative tolerance
                   tol,             // Absolute tolerance
                   PETSC_DEFAULT,   // Div. Tolerance
                   PETSC_DEFAULT);  // Max Itr
  
  
  KSPSetType(ksp,KSPSYMMLQ);  //' solver:SYMMLQ
}

/** Destructor
 */
CCG::~CCG()
{
  if(X_old_array) free(X_old_array);
  if(X_array)     free(X_array);
  if(B_array)     free(B_array);
  if(A_array)     free(A_array);
  if(values)      free(values);
  if(indices)     free(indices);

  KSPDestroy(ksp);   

  for(int i=0; i<n; i++)
    delete aSet[i];

}


/** Solve the QP by calling PETSc KSP method : Conjugate Gradient
 */
ErrorCode
CCG::Solve()
{ 
  PetscInt itr;
  //Real sum = 0;

  // call CG
  VecZeroEntries(X);
  ierr = KSPSolve(ksp,B,X); CHKERRQ(ierr); 

  if(verbosity > 1) {
    ierr = KSPGetIterationNumber(ksp,&itr);CHKERRQ(ierr);
    cout << "# CG iteration: " << itr << endl;
  } 
  
  // enforce x >= 0
  // Real t = 1;
  // svnvish: BUGBUG
  Real t = INFTY;
  for(int i=1; i<n+1; i++) {    
    if(X_array[i] < 0) {    
      // t = MIN(t, X_old_array[i]/(X_old_array[i]-X_array[i]));
      // svnvish: BUGBUG
      t = MIN(t, X_array[i]);
      
    }
    if(verbosity > 0) {
      cout << i-1 << ":  " << X_array[i] << "  " << X_old_array[i] << endl;
      //cout << "t:" << t << endl;
    }
  }
 //  for(int i=0; i<n+1; i++) {
//     // note that i starts from zero!
//     X_array[i] = X_old_array[i] + t*(X_array[i]-X_old_array[i]);
//     if(i>=1 && X_array[i] < ZERO_EPS) {
//       X_array[i] = 0;
//       //cout << "setting "<<i<<"-th time to -1000" << endl;
//       //timeStamp[i];  // must be replaced in next iter!
//       //B_array[i] = 0;
//     }
//   }

  cout << "t:" << t << endl;

  if(verbosity > 3) {
    //cout << "t:" << t << endl;
    //cout << "A:" << endl;
    //MatView(A, PETSC_VIEWER_STDOUT_SELF);
    cout << "X:" << endl;
    VecView(X, PETSC_VIEWER_STDOUT_SELF);
    //cout << "B:" << endl;
    //VecView(B, PETSC_VIEWER_STDOUT_SELF);
  }
  
  // copy final X into X_old
  //VecCopy(X, X_old);
  

  return NOERROR;
}


/**  Update matrix and vectors needed in optimisation round.
 *
 *   \param a [read] Gradient
 *   \param b [read] Offset 
 */
ErrorCode 
CCG::Update(CMatVecArr* a, Real b)
{
  int how_old = 40;                  // constraint not used in the last #how_old# iteration will be replaced by new one.
  int idx = -1;                      // default: look for i such that x[i] is smallest
  int first_new_slot = -1;
  int oldest_slot = 0;
  int smallest_slot = 0;
  
  iter++;                            // another column generated

  // deciding which slot to store the latest gradient
  for(int i=0; i<max_mem_size; i++) {
    if(timeStamp[i] == -1 && first_new_slot == -1) first_new_slot = i;
    if(timeStamp[i] < timeStamp[oldest_slot] && timeStamp[i] >= 0) oldest_slot = i; 
    //if(timeStamp[i] < timeStamp[oldest_slot]) oldest_slot = i; 
    if(x[i] < x[smallest_slot]) smallest_slot = i;
  }
  if(timeStamp[oldest_slot] < iter - how_old) 
    idx = oldest_slot;
  else if(first_new_slot != -1) {
    idx = first_new_slot;
    timeStamp[idx] = iter;
    n++;
  }
  else 
    idx = smallest_slot;
  
  if(verbosity > 1) cout << "new column idx: " << idx << "   aSet size:" << aSet.size() << endl;


  // store new column at #idx#
  if(aSet[idx]) delete aSet[idx];
  aSet[idx] = a;                   // store the gradient
  bSet[idx] = b;                   // store the offset  

  // compute dot-products (if we store aSet as matrix, this can be computed much faster)
  for(int col=0; col<n; col++) {
    // the first element of values is always 1 (the element of a in the QP formulation above)
    VecDot(aSet[idx]->vec, aSet[col]->vec, &values[1+col]);   
    values[1+col] = 1.0/info->lambda*values[1+col];
  }
  values[0] = 1;

  // update the idx-th row and col of H
  PetscInt index = idx+1;  // first row and col of H belongs to the vector a in the QP formulation above
  ierr = MatSetValues(A, 1, &index, n+1, indices, values, INSERT_VALUES);
  ierr = MatSetValues(A, n+1, indices, 1, &index, values, INSERT_VALUES);
  MatAssemblyBegin(A, MAT_FINAL_ASSEMBLY); MatAssemblyEnd(A, MAT_FINAL_ASSEMBLY);

  // update vectors Y and C (well, indirecly through the x and f arrays :)
  x[idx] = 0;
  f[idx] = bSet[idx];

  return NOERROR;
}


/** Return solution.
 *
 *  \param w      [write] solution vector (preallocated)
 *  \param xi     [write] lower bound of empirical risk, R_emp[w]
 *  \param objval [write] objective value
 */
ErrorCode 
CCG::GetSolution(CMatVecArr* &w, Real &xi, Real &objval)
{
  assert(w->row > 0 && w->col > 0);
  assert(x != 0);
  Real factor = 1.0/info->lambda;


  // compute objective value (explicitly)
  // n := dimensionality of x (the solution of QP)
  // A_array := augmented hessian matrix
  // f := linear part of obj
  objval = 0;
  Real tmp = 0;
  Real fx = 0;
  int dim = max_mem_size+1;
  for(int i=0; i<n; i++) {
    tmp = 0;
    for(int j=0; j<n; j++)
      tmp += A_array[(j+1) + (i+1)*dim]*x[j];
    objval += x[i]*tmp;
    fx -= f[i]*x[i];   // note that we omitted the -ve sign in vector f due to the linear system formulation
  }


  // since the dual of minimization problem is maximization
  // and we solved the  -maximization version,
  // the objective value should be -objval
  objval = -0.5*objval - fx;

  // Compute new w
  VecZeroEntries(w->vec);  
  for(int i=0; i<n; i++)
    VecAXPY(w->vec, -x[i], aSet[i]->vec);
  VecScale(w->vec, factor);

  // compute xi
  Real normsq = 0;
  VecDot(w->vec, w->vec, &normsq);
  xi = (objval - info->lambda*0.5*normsq)/info->C;

  // update time-stamp of constraints
  for(int i=0; i<n; i++)
    if(x[i] > 0) timeStamp[i] = iter;

  // constraint set size
  if(verbosity > 1)
    cout << "qp_size: " << n << endl;


  // display nonzero terms
  if(verbosity > 2) {
    int howmany=0;
    cout << "nz: ";
    for(int i=0; i<n; i++) {
      if(ABS(x[i])> ZERO_EPS) {
        cout << i << "..";
        howmany++;
      }
    }
    cout << "   ("<<howmany<<")"<<endl;
  }

  return NOERROR;
}

#endif
