/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        

 * Created      : 20/04/2007
 * Last Updated :
 *
 */

#ifndef _CG_HPP_
#define _CG_HPP_

#include "errorcode.hpp"
#include "common.hpp"
#include "info.hpp"
#include "matvecarr.hpp"
#include "solver.hpp"
#include <iostream>
#include <cassert>

#include "petsc.h"
#include "petscvec.h"
#include "petscksp.h"


/* 
   Solve:
   
   (P) min  x'Qx + f'x
        x
       s.t. a'x = b      (1)
            x >= 0       (2)

   Lagrangian of (P), L(x,s) := x'Qx + f'x + s(a'x-b)
                      s.t. x >= 0

   At optimality,
   \partial_s L = 0  => a'x = b        (3)
   \partial_x L = 0  => Qx + sa = -f   (4)

 
   Ignoring (2) at the moment and packing (3) & (4) into a linear system,
   
      A      X  =  B

   +      + + +   +  +
   | 0  a'| |s| = | b|
   | a  Q | |x|   |-f|
   +      + + +   +  +

   where a,x,f := column vector
             Q := matrix
             s := lagrange multiplier

   Solving the linear system at gives us a feasible solution (s,x).
   Now, we enforce (2) by finding a t \in [0,1] such that x := x_old + t*(x - x_old)  >= 0,
   where x_old is the initial value of x. Similar adjustment for s. Note that t is found
   w.r.t. x only.
   
*/

class CCG : public CSolver 
{
public:

  CCG(){}
  CCG(CInfo* _info, Real _tol=1e-8);
  
  virtual ~CCG();

  ErrorCode Solve();
  ErrorCode Update(CMatVecArr* a, Real b);
  ErrorCode GetSolution(CMatVecArr* &w, Real &xi, Real &objval);

private:
  int   max_mem_size;
  Real *values;
  int  *indices;
  
  // we allocate memory for all PETSc vec and mat.
  Real *X_old_array;
  Real *X_array;
  Real *B_array;
  Real *A_array;

  // PETSc error flag holder
  PetscInt ierr;

  // the PETSc vec and mat
  Vec X_old;
  Vec X;
  Vec B;
  Mat A;  

  // PETSC KSP solver
  KSP ksp;
};

#endif
