/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 30/04/2007
 * Last Updated :
 * 
 * Purpose      : Linear Program solver using COIN-OR Clp for cutting plane method
 *
 */

#ifndef _CUTTINGPLANE_CPP_
#define _CUTTINGPLANE_CPP_


#include "cuttingplane.hpp"
#include <vector>

using namespace std;

// Clp headers
//#include "CoinBuild.hpp"


/*
    minimize    \xi
     \xi,w

    subject to  <a_i,w> - \xi <= -b_i  \forall i \in [1..k]

*/


CCuttingPlane::CCuttingPlane(CInfo* _info,  Real _tol)
  :newRowIndices(0),
   newRowElements(0)
{
  // parameters
  iter = 0;
  info = _info;
  verbosity = info->verbosity;
  n = info->dimW;
  m = 0;
  x = 0;
  Q = 0;
  f = 0;
  a = 0;
  b = 0;
  l = 0;
  u = 0;

  
  // build simplex model
  sim = new ClpSimplex();
  sim->setLogLevel(0);
  sim->resize(0, 1+n);

  // set objective coefficients    
  sim->setObjCoeff(0, 1.0/info->lambda);
  for(int i=1; i<n+1; i++) 
    sim->setObjCoeff(i, 0);
  
  // set lower bounds for variables (columns)
  sim->setColumnBounds(0, 0.0, COIN_DBL_MAX);
  for (int i=1; i<n+1; i++)
    //sim->setColumnBounds(i, -COIN_DBL_MAX, COIN_DBL_MAX);
    sim->setColumnBounds(i, -1, 1);
  
  // something useful in constraint matrix update phase
  newRowIndices = (int*)calloc(1+n, sizeof(int));
  newRowElements = (Real*)calloc(1+n, sizeof(Real));    
  
  for(int i=0; i<n+1; i++)
    newRowIndices[i] = i;
  
  newRowElements[0] = -1;
  // Update() set the rest of the elements in newRowElements[]


  //   printf("num cols : %d\n", sim->numberColumns());
  //   printf("num rows : %d\n", sim->numberRows());
}


CCuttingPlane::~CCuttingPlane()
{
  if(sim) delete sim;
  if(newRowIndices) free(newRowIndices);
  if(newRowElements) free(newRowElements);
  
}

  
ErrorCode 
CCuttingPlane::Solve()
{
  //sim->dual();
  sim->primal();

  return NOERROR;
}


/**  Update matrix and vectors needed in optimisation round.
 *
 *   \param a [read] Gradient
 *   \param b [read] Offset 
 */
ErrorCode 
CCuttingPlane::Update(CMatVecArr* a, Real b)
{
  iter++;
  
   // delete inactive gradients when constraint matrix reaches the limit (i.e. info->memSize)
//   if(m >= info->memSize) {
//     int delrows[20] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
//     sim->deleteRows(20, delrows);
//     m -= 20;
//   }
   
  // addRow
  memcpy(newRowElements+1, a->array, a->row*a->col*sizeof(Real));
  sim->addRow(1+n, newRowIndices, newRowElements, -COIN_DBL_MAX, -b);
  m++;
  
  //printf("num cols : %d\n", sim->numberColumns());
  //printf("num rows : %d\n", sim->numberRows());
  
  // free the gradient
  delete a;

  return NOERROR;
}



/** Return solution.
 *
 *  \param w      [write] solution vector (preallocated)
 *  \param xi     [write] lower bound of empirical risk, R_emp[w]
 *  \param regval [write] value of regularization term
 *  \param objval [write] objective value
 */
ErrorCode 
CCuttingPlane::GetSolution(CMatVecArr* &w, Real &xi, Real &regval, Real &objval)
{
  x = sim->primalColumnSolution(); 
  assert(x);  
  assert(w->row > 0 && w->col > 0);
  
  memset(w->array, 0, sizeof(Real)*w->row*w->col);
  memcpy(w->array, x+1, w->row*w->col*sizeof(Real));

  // compute some important values
  xi = x[0];
  objval = sim->objectiveValue();
  regval = 0;
    
  Real sum = 0;
  for(int i=0; i<n; i++)
    sum += x[i+1];

  cout << "sum : " << sum << endl;
  cout << "objval : " << objval << endl;

  return NOERROR;
}

#endif
