/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 16/11/2006
 * Last Updated :
 */

#ifndef _DAIFLETCHERPGM_CPP_
#define _DAIFLETCHERPGM_CPP_

#include <iostream>
#include <cmath>
#include <cstdlib>
#include <vector>

#include "daifletcherpgm.hpp"


using namespace std;



CDaiFletcherPGM::CDaiFletcherPGM(CInfo* _info, Real _tol)
  :maxProjIter(200), 
   maxPGMIter(20000),
   increment(50),
   max_mem_size(_info->memSize),  // max num of constraints
   cur_mem_size(50),
   ipt(0),
   ipt2(0),
   uv(0),
   g(0),
   y(0),
   tempv(0),
   d(0),
   Qd(0),
   t(0),
   xplus(0),
   tplus(0),
   sk(0),
   yk(0)
{
  // inherited from CInnerSolver
  iter         = 0;
  info         = _info;
  tol          = _tol;
  verbosity    = info->verbosity;
  n            = 0;                       // dimension of solution (increasing)
  m            = 1;                       // number of eq or ineq constraint  (fixed)
  
  // for precision related
  if(sizeof(Real) == sizeof(double)) {
    EPS_SV = 1e-15;
    eps    = 1e-20;
  }
  else {
    EPS_SV = 1e-7;
    eps    = 1e-8;
  }

  // pre-allocate memory for gradient, offset and time-stamps
  aSet.resize(max_mem_size);                  
  bSet.resize(max_mem_size);                  
  timeStamp.resize(max_mem_size, -1);


  // allocate maximum memory needed
  Q = (Real*)calloc(max_mem_size*max_mem_size, sizeof(Real));
  x = (Real*)calloc(max_mem_size,    sizeof(Real));
  f = (Real*)calloc(max_mem_size,    sizeof(Real));
  l = (Real*)calloc(max_mem_size,    sizeof(Real));
  u = (Real*)calloc(max_mem_size,    sizeof(Real));
  a = (Real*)calloc(max_mem_size,    sizeof(Real));
  b = new Real(info->C); 
  
  // initializations
  for(int i=0; i<max_mem_size; i++) {
    l[i] = 0;
    u[i] = INFTY;
    a[i] = 1;      
  }

  // pre-allocate mem needed by solver
  ReallocMem(cur_mem_size);

}

CDaiFletcherPGM::~CDaiFletcherPGM()
{
  // free for parent
  if(x)     free(x);
  if(Q)     free(Q);
  if(f)     free(f);
  if(l)     free(l);
  if(u)     free(u);
  if(a)     free(a);
  if(b)     delete b;
  
  // free mem allocated
  if(ipt)   free(ipt);
  if(ipt2)  free(ipt2);
  if(uv)    free(uv);
  if(g)     free(g);
  if(y)     free(y);
  if(tempv) free(tempv);
  if(d)     free(d);
  if(Qd)    free(Qd);
  if(t)     free(t);
  if(xplus) free(xplus);
  if(tplus) free(tplus);
  if(sk)    free(sk);
  if(yk)    free(yk);

  for(int i=0; i<n; i++)
    delete aSet[i];

}

/** Reallocate memory.
 */
void CDaiFletcherPGM::ReallocMem(int n)
{
  ipt   = (int  *)realloc(ipt,   n*sizeof(int));
  ipt2  = (int  *)realloc(ipt2,  n*sizeof(int));
  uv    = (int  *)realloc(uv,    n*sizeof(int));
  g     = (Real *)realloc(g,     n*sizeof(Real));
  y     = (Real *)realloc(y,     n*sizeof(Real));
  tempv = (Real *)realloc(tempv, n*sizeof(Real));
  d     = (Real *)realloc(d,     n*sizeof(Real));
  Qd    = (Real *)realloc(Qd,    n*sizeof(Real));
  t     = (Real *)realloc(t,     n*sizeof(Real));
  xplus = (Real *)realloc(xplus, n*sizeof(Real));
  tplus = (Real *)realloc(tplus, n*sizeof(Real));
  sk    = (Real *)realloc(sk,    n*sizeof(Real));
  yk    = (Real *)realloc(yk,    n*sizeof(Real));

  cur_mem_size = n;
}

// Real CDaiFletcherPGM::MinDiff(Real *l, Real *u, int len)
// {
//   register int i;
//   Real diff = u[0] - l[0];

//   for(i=1; i<len; i++)
//     if(u[i]-l[i] < diff) 
//       diff = u[i]-l[i];

//   return diff;
// }


// Real CDaiFletcherPGM::MaxElem(Real *a, int len)
// {
//   register int i;
//   Real maxelem = a[0];
 
//   for(i=1; i<len; i++)
//     if(a[i] > maxelem)
//       maxelem = a[i];

//   return maxelem;
// }



/** Piecewise linear monotone target function for the Dai-Fletcher projector.
 *
 *  
 */
Real CDaiFletcherPGM::ProjectR(Real *x1, int n, Real lambda, Real *a1, 
                                 Real b, Real *c, Real *l, Real *u)
{
  register int i;
  Real r = 0.0;

  for (i = 0; i < n; i++)
  {
    x1[i] = -c[i] + lambda*a1[i];
    if (x1[i] > u[i]) 
      x1[i] = u[i];
    else if(x1[i] < l[i]) 
      x1[i] = l[i];
    r += a1[i]*x1[i];
  }

  return (r - b);
}




/** Modified Dai-Fletcher QP projector solves the problem:
 *
 *      minimise  0.5*x'*x + c'*x
 *      subj to   a'*x <= b
 *                l <= x <= u  
 *
 *  \param c The point to be projected onto feasible set
 */ 
int CDaiFletcherPGM::ProjectDF(int n, Real *a, Real b, Real *c, Real *l, 
                               Real *u, Real *x, Real &lam_ext)
{
  Real lambda, lambdal, lambdau, dlambda, lambda_new, tol_lam;
  Real r, rl, ru, s, tol_r;
  int    innerIter;

  tol_lam = 1.0e-15;
  tol_r   = 1.0e-15;
  lambda  = lam_ext = 0;
  dlambda = 0.5;
  innerIter = 1;

  /*  \phi(x;lambda) := 0.5*x'*x + c'*x - lambda*(a'*x-b)
   *
   *  Optimality conditions for \phi:
   *
   *  1. lambda   <= 0
   *  2. r        <= 0
   *  3. r*lambda == 0
   */

  
  // Bracketing Phase
  r = ProjectR(x, n, lambda, a, b, c, l, u);

  // chteo: equality constraint
  //if (fabs(r) < tol_r) return 0;

  // chteo: inequality constraint
  if (r < tol_r) return 0;

  
  if (r < 0.0)
  {
    lambdal = lambda;
    rl      = r;
    lambda  = lambda + dlambda;
    r       = ProjectR(x, n, lambda, a, b, c, l, u);
    while (r < 0.0 && dlambda < INFTY)
    {
      lambdal = lambda;
      s       = rl/r - 1.0;
      if (s < 0.1) s = 0.1;
      dlambda = dlambda + dlambda/s;
      lambda  = lambda + dlambda;
      rl      = r;
      r       = ProjectR(x, n, lambda, a, b, c, l, u);
    }
    lambdau = lambda;
    ru      = r;
  }
  else
  {
    lambdau = lambda;
    ru      = r;
    lambda  = lambda - dlambda;
    r       = ProjectR(x, n, lambda, a, b, c, l, u);
    while (r > 0.0 && dlambda > -INFTY)
    {
      lambdau = lambda;
      s       = ru/r - 1.0;
      if (s < 0.1) s = 0.1;
      dlambda = dlambda + dlambda/s;
      lambda  = lambda - dlambda;
      ru      = r;
      r       = ProjectR(x, n, lambda, a, b, c, l, u);
    }
    lambdal = lambda;
    rl      = r;
  }

  if(ABS(dlambda) > INFTY) {
    printf("ERROR: (DaiFletcherPGM) Infeasible problem!\n");
    exit(0);
  }

  // Secant Phase
  s       = 1.0 - rl/ru;
  dlambda = dlambda/s;
  lambda  = lambdau - dlambda;
  r       = ProjectR(x, n, lambda, a, b, c, l, u);

  while (   fabs(r) > tol_r 
         && dlambda > tol_lam * (1.0 + fabs(lambda)) 
         && innerIter    < maxProjIter                    )
  {
    innerIter++;
    if (r > 0.0)
    {
      if (s <= 2.0)
      {
        lambdau = lambda;
        ru      = r;
        s       = 1.0 - rl/ru;
        dlambda = (lambdau - lambdal) / s;
        lambda  = lambdau - dlambda;
      }
      else
      {
        s          = ru/r-1.0;
        if (s < 0.1) s = 0.1;
        dlambda    = (lambdau - lambda) / s;
        lambda_new = 0.75*lambdal + 0.25*lambda;
        if (lambda_new < (lambda - dlambda))
          lambda_new = lambda - dlambda;
        lambdau    = lambda;
        ru         = r;
        lambda     = lambda_new;
        s          = (lambdau - lambdal) / (lambdau - lambda);
      }
    }
    else
    {
      if (s >= 2.0)
      {
        lambdal = lambda;
        rl      = r;
        s       = 1.0 - rl/ru;
        dlambda = (lambdau - lambdal) / s;
        lambda  = lambdau - dlambda;
      }
      else
      {
        s          = rl/r - 1.0;
        if (s < 0.1) s = 0.1;
        dlambda    = (lambda-lambdal) / s;
        lambda_new = 0.75*lambdau + 0.25*lambda;
        if (lambda_new > (lambda + dlambda))
          lambda_new = lambda + dlambda;
        lambdal    = lambda;
        rl         = r;
        lambda     = lambda_new;
        s          = (lambdau - lambdal) / (lambdau-lambda);
      }
    }
    r = ProjectR(x, n, lambda, a, b, c, l, u);
  }
  
  lam_ext = lambda;
  if(innerIter >= maxProjIter)
    printf("WARNING: {DaiFletcherPGM} Projector exits after max iterations (%d): %d\n", maxProjIter, innerIter);
  
  return (innerIter);
}


ErrorCode CDaiFletcherPGM::Solve()
{
  int    i, j, innerIter, it, luv, info, lscount = 0, projcount = 0;
  Real gd, max, ak, bk, akold, bkold, lamnew, alpha, kktlam, lam_ext;
  Real DELTAsv, ProdDELTAsv;
  Real c;


  /*** variables for the adaptive nonmonotone linesearch ***/
  int    L, llast;
  Real fr, fbest, fv, fc, fv0;
  
  // check if pre-allocated memory is enough
  if(n > cur_mem_size) ReallocMem(n+increment);

  c = INFTY; //c = MaxElem(ub,n);
  DELTAsv = EPS_SV;// * c;
  if (tol <= 1.0e-5 || n <= 20)
    ProdDELTAsv = 0.0F;
  else
    ProdDELTAsv = EPS_SV;// * c;
  
  for (i = 0; i < n; i++) {
    tempv[i] = -x[i];
  }
  
  lam_ext = 0.0;
  
  projcount += ProjectDF(n, a, *b, tempv, l, u, x, lam_ext);

  // g = Q*x + f;
  // SparseProd(n, t, Q, x, ipt);
  {
    int    it;
    Real *tempQ;
    
    it = 0;
    for (i = 0; i < n; i++)
      if (fabs(x[i]) > ProdDELTAsv)
        ipt[it++] = i;
    
    memset(t, 0, n*sizeof(Real));
    for (i = 0; i < it; i++)
    {
      tempQ = Q + ipt[i] * n;
      for (j = 0; j < n; j++)
        t[j] += (tempQ[j]*x[ipt[i]]);
    }
  }
  
  for (i = 0; i < n; i++) 
  {
    g[i] = t[i] + f[i];
    y[i] = g[i] - x[i];
  }

  projcount += ProjectDF(n, a, *b, y, l, u, tempv, lam_ext);

  max = alpha_min;
  for (i = 0; i < n; i++)
  {
    y[i] = tempv[i] - x[i];
    if (fabs(y[i]) > max)
      max = fabs(y[i]);
  }

  if (max < tol*1e-3)
  {
    lscount = 0;
    innerIter    = 0;
    goto Clean;    
  }

  alpha = 1.0 / max;

  fv0   = 0.0;
  for (i = 0; i < n; i++)
    fv0 += x[i] * (0.5*t[i] + f[i]);

  /*** adaptive nonmonotone linesearch ***/
  L     = 2;
  fr    = alpha_max;
  fbest = fv0;
  fc    = fv0;
  llast = 0;
  akold = bkold = 0.0;

  for (innerIter = 1; innerIter <= maxPGMIter; innerIter++)
  {
    for (i = 0; i < n; i++)
      tempv[i] = alpha*g[i] - x[i];

    projcount += ProjectDF(n, a, *b, tempv, l, u, y, lam_ext);

    gd = 0.0;
    for (i = 0; i < n; i++)
    {
      d[i] = y[i] - x[i];
      gd  += d[i] * g[i];
    }
    
    /* compute Qd = Q*d  or  Qd = Q*y - t depending on their sparsity */
    {
      int    i, it, it2;
      Real *tempQ;
      
      it = it2 = 0;
      for (i = 0; i < n; i++)
        if (fabs(d[i]) > (ProdDELTAsv*1.0e-2))
          ipt[it++]   = i;
      for (i = 0; i < n; i++)
        if (fabs(y[i]) > ProdDELTAsv)
          ipt2[it2++] = i;
      
      memset(Qd, 0, n*sizeof(Real));
      if (it < it2) // compute Qd = Q*d
      {
        for (i = 0; i < it; i++)
        {
          tempQ = Q + ipt[i]*n;
          for (j = 0; j < n; j++)
            Qd[j] += (tempQ[j] * d[ipt[i]]);
        }
      }
      else          // compute Qd = Q*y-t
      {
        for (i = 0; i < it2; i++)
        {
          tempQ = Q + ipt2[i]*n;
          for (j = 0; j < n; j++)
            Qd[j] += (tempQ[j] * y[ipt2[i]]);
        }
        for (j = 0; j < n; j++)
          Qd[j] -= t[j];
      }
    }

    ak = 0.0;
    for (i = 0; i < n; i++)
      ak += d[i] * d[i];
    
    bk = 0.0;
    for (i = 0; i < n; i++)
      bk += d[i]*Qd[i];
    
    if (bk > eps*ak && gd < 0.0)    // ak is normd
      lamnew = -gd/bk;
    else
      lamnew = 1.0;
    
    
    fv = 0.0;
    for (i = 0; i < n; i++)
    {
      xplus[i] = x[i] + d[i];
      tplus[i] = t[i] + Qd[i];
      fv      += xplus[i] * (0.5*tplus[i] + f[i]);
    }

    if ((innerIter == 1 && fv >= fv0) || (innerIter > 1 && fv >= fr))
    {
      lscount++;
      fv = 0.0;
      for (i = 0; i < n; i++)
      {
        xplus[i] = x[i] + lamnew*d[i];
        tplus[i] = t[i] + lamnew*Qd[i];
        fv      += xplus[i] * (0.5*tplus[i] + f[i]);
      }
    }
    
    for (i = 0; i < n; i++)
    {
      sk[i] = xplus[i] - x[i];
      yk[i] = tplus[i] - t[i];
      x[i]  = xplus[i];
      t[i]  = tplus[i];
      g[i]  = t[i] + f[i];
    }

    // update the line search control parameters
    
    if (fv < fbest)
    {
      fbest = fv;
      fc    = fv;
      llast = 0;
    }
    else
    {
      fc = (fc > fv ? fc : fv);
      llast++;
      if (llast == L)
      {
        fr    = fc;
        fc    = fv;
        llast = 0;
      }
    }

    ak = bk = 0.0;
    for (i = 0; i < n; i++)
    {
      ak += sk[i] * sk[i];
      bk += sk[i] * yk[i];
    }

    if (bk <= eps*ak)
      alpha = alpha_max;
    else
    {
      if (bkold < eps*akold)
        alpha = ak/bk;
      else
        alpha = (akold+ak)/(bkold+bk);
      
      if (alpha > alpha_max)
        alpha = alpha_max;
      else if (alpha < alpha_min)
        alpha = alpha_min;
    }

    akold = ak;
    bkold = bk;
    

    /*** stopping criterion based on KKT conditions ***/
    // chteo: at optimal, gradient of lagrangian w.r.t. x is zero

    bk = 0.0;
    for (i = 0; i < n; i++)
      bk +=  x[i] * x[i];
    

    if (sqrt(ak) < tol*10 * sqrt(bk))
    {
      it     = 0;
      luv    = 0;
      kktlam = 0.0;
      for (i = 0; i < n; i++)
      {
        // chteo: x[i] is active hence lagrange multipliers for box constraints
        //        are zero. The lagrange multiplier for ineq. const. is then 
        //        defined as below

        if ((x[i] > DELTAsv) && (x[i] < c-DELTAsv))
        {
          ipt[it++] = i;
          kktlam    = kktlam - a[i]*g[i];
        }
        else
          uv[luv++] = i;
      }
    
      if (it == 0)
      {
        if (sqrt(ak) < tol*0.5 * sqrt(bk))
          goto Clean;
      }
      else
      {
        
        kktlam = kktlam/it;
        info   = 1;
        for (i = 0; i < it; i++) {
          if ( fabs(a[ipt[i]] * g[ipt[i]] + kktlam) > tol )
          {
            info = 0;
            break;
          }
        }
        if (info == 1)
          for (i = 0; i < luv; i++)
            if (x[uv[i]] <= DELTAsv)
            {
              //chteo: x[i] == lower bound, hence, lagrange multiplier (say, beta) for lower bound may
              //       not be zero. So, the gradient without beta is > 0
              if (g[uv[i]] + kktlam*a[uv[i]] < -tol)
              {
                info = 0;
                break;
              }
            }
            else
            {
              //chteo: x[i] == upper bound, hence, lagrange multiplier (say, eta) for upper bound may
              //       not be zero. So, the gradient without eta is < 0
              if (g[uv[i]] + kktlam*a[uv[i]] > tol)
              {
                info = 0;
                break;
              }
            }
        
        if (info == 1)
          goto Clean;
      }
    }
  }
    
  printf("\nERROR: Dai-Fletcher PGM exceeded maxPGMIter: %d iter. "\
         "Try lower tolerence (current:%g) or smaller C parameter.\n", 
         maxPGMIter, tol);
  
 Clean:

  if(verbosity > 1)
    printf("Gradient projection count: %d\n",innerIter);

  if(verbosity > 2) {
    for(int i=0; i<n; i++)
      cout << "x["<<i<<"]: " << x[i] << endl;
  }

  return NOERROR;
}

/**  Update matrix and vectors needed in optimisation round.
 *
 *   \param a [read] Gradient
 *   \param b [read] Offset 
 */
ErrorCode 
CDaiFletcherPGM::Update(CMatVecArr* a, Real b)
{
  int how_old = 40;       // constraint not used in the last #how_old# iteration will be replaced by new one.
  int idx = -1;                      // default: look for i such that x[i] is smallest
  int first_new_slot = -1;
  int oldest_slot = 0;
  int smallest_slot = 0;
  
  iter++;                            // another column generated

  // deciding which slot to store the latest gradient
  for(int i=0; i<max_mem_size; i++) {
    if(timeStamp[i] == -1 && first_new_slot == -1) first_new_slot = i;
    if(timeStamp[i] < timeStamp[oldest_slot] && timeStamp[i] >= 0) oldest_slot = i; 
    if(x[i] < x[smallest_slot]) smallest_slot = i;
  }
  if(timeStamp[oldest_slot] < iter - how_old) 
    idx = oldest_slot;
  else if(first_new_slot != -1) {
    idx = first_new_slot;
    n++;
  }
  else 
    idx = smallest_slot;
  
  if(verbosity > 1) cout << "new column idx: " << idx << "   aSet size:" << aSet.size() << endl;


  // store new column at #idx#
  if(aSet[idx]) delete aSet[idx];
  aSet[idx] = a;                   // store the gradient
  bSet[idx] = b;                   // store the offset  


  // use new slot
  if(timeStamp[idx] == -1)
  {
    // chteo: if this is not active after immediate optimization, we still make to make it current since it is new!
    timeStamp[idx] = iter;  

    // step 1: insert new row into Q
    int lastRow = (n-1)*n;
    for(int col=0; col<n; col++) {
      Real value = 0;
      VecDot(aSet[idx]->vec, aSet[col]->vec, &value);
      Q[lastRow+col] = 1.0/info->lambda*value;
    }

    // step 2: adjust memory area to fit a new column
    int prevDim = n-1;
    int size    = sizeof(Real)*prevDim;
    for(int row=n-2; row >= 1; row--)
    {
      Real *src  = Q + (row*prevDim);
      Real *dest = Q + (row*prevDim)+ row;
      memmove(dest, src, size);
    }

    // step 3: insert new column (at the right-most-side of the matrix)
    for(int i=0; i<n-1; i++)
      Q[i*n + n-1] = Q[lastRow + i];
    
    // update vectors
    x[n-1] = 0;
    f[n-1] = -bSet[n-1];
  }
  // reuse slot
  else 
  { 
    // This gradient is new!
    timeStamp[idx] = iter;  
   
    // insert new row into Q[idx,:]
    int theRow = idx*n;
    for(int col=0; col<n; col++) {
      Real value = 0;
      VecDot(aSet[idx]->vec, aSet[col]->vec, &value);
      Q[theRow+col] = 1.0/info->lambda*value;
    }
    
    // insert new column into Q[:,idx]
    for(int i=0; i<n; i++)
      Q[i*n + idx] = Q[theRow + i];
    
    // update vectors      
    x[idx] = 0;
    f[idx] = -bSet[idx];
  }

#ifdef DEBUG
  // check Q matrix !    
  MatrixCorrectnessCheck();
#endif

  return NOERROR;
}


/** Return solution.
 *
 *  \param w      [write] solution vector (preallocated)
 *  \param xi     [write] lower bound of empirical risk, R_emp[w]
 *  \param regval [write] value of regularization term
 *  \param objval [write] objective value
 */
ErrorCode 
CDaiFletcherPGM::GetSolution(CMatVecArr* &w, Real &xi, Real &regval, Real &objval)
{
  assert(w->row > 0 && w->col > 0);
  assert(x != 0);
  Real factor = 1.0/info->lambda;


  // compute objective value (explicitly)
  // n := dimensionality of x (the solution of QP)
  // Q := hessian matrix
  // f := linear part of obj
  objval = 0;
  Real tmp = 0;
  Real fx = 0;
  for(int i=0; i<n; i++) {
    tmp = 0;
    for(int j=0; j<n; j++)
      tmp += Q[j + i*n]*x[j];
    objval += x[i]*tmp;
    fx += f[i]*x[i];
  }

  // since the dual of minimization problem is maximization
  // and we solved the  -maximization version,
  // the objective value should be -objval
  objval = -0.5*objval - fx;

  
  // Compute new w
  VecZeroEntries(w->vec);  
  for(int i=0; i<n; i++)
    if(x[i] > ZERO_EPS)
      VecAXPY(w->vec, -x[i], aSet[i]->vec);
  VecScale(w->vec, factor);

  // compute xi
  //VecDot(w->vec, w->vec, &regval);
  VecNorm(w->vec, NORM_2, &regval);
  regval = regval * regval * info->lambda * 0.5; 
  xi = (objval - regval)/info->C;


  // update time-stamp of constraints
  for(int i=0; i<n; i++)
    if(x[i] > 0) timeStamp[i] = iter;

  // constraint set size
  if(verbosity > 1)
    cout << "qp_size: " << n << endl;


  // display nonzero terms
  if(verbosity > 2) {
    int howmany=0;
    cout << "nz: ";
    for(int i=0; i<n; i++) {
      if(ABS(x[i])> ZERO_EPS) {
        cout << i << "..";
        howmany++;
      }
    }
    cout << "   ("<<howmany<<")"<<endl;
  }

  return NOERROR;
}


#ifdef DEBUG
/** Check if the Q matrix update is correct
 */
void
CDaiFletcherPGM::MatrixCorrectnessCheck() 
{
    cout << " in matrix correctness check! " << endl;
  Real dot = 0;
  Real *correctmat = new Real[n*n];
  memset(correctmat, 0, sizeof(Real)*n*n);
  
  for(int i=0; i<n; i++)
  {
    for(int j=i; j<n; j++)
    {
      //cout << "i:" << i << "   j:" << j << endl;
      dot = 0;
      VecDot(aSet[i]->vec, aSet[j]->vec, &dot);
      correctmat[i*n + j] = 1.0/info->lambda*dot;
      correctmat[j*n + i] = correctmat[i*n+j];
    }
  }
  
  for(int i=0; i<n*n; i++)
  {
    if(fabs(correctmat[i] - Q[i]) > 1e-15) {
      cout << "residual : " << fabs(correctmat[i] - Q[i]) << endl;
      cout << "matrix update correctness check : i = " << i << endl;
      cout << "ERROR: rank-1 updated matrix is incorrect ! " << endl;
      exit(-1);
    }
  }
}
#endif

#endif
