/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 16/11/2006
 * Last Updated :
 *
 * Notes :
 *
 * Dai-Fletcher Projected Gradient method for SVM [1].
 * Modified from the GPDT software [2] for inequality constraints.
 *
 * References:
 *
 *   [1] Y. H. Dai and R. Fletcher,
 *       New algorithms for singly linearly constrained quadratic programs
 *       subject to lower and upper bounds, 
 *       Math. Program., 2006.
 *
 *   [2] L. Zanni, T. Serafini, and G. Zanghirati,
 *       Parallel Software for Training Large Scale Support Vector Machines
 *       on Multiprocessor Systems,
 *       JMLR 7, 2006.
 */

#ifndef _DAIFLETCHERPGM_HPP_
#define _DAIFLETCHERPGM_HPP_

#include "errorcode.hpp"
#include "common.hpp"
#include "info.hpp"
#include "matvecarr.hpp"
#include "solver.hpp"
#include <iostream>
#include <cassert>

// from PETSc
#include "petsc.h"
#include "petscvec.h"

#define alpha_min 1e-10
#define alpha_max 1e10


class CDaiFletcherPGM : public CSolver 
{
public:

  CDaiFletcherPGM(){}
  CDaiFletcherPGM(CInfo* _info, Real _tol=1e-8);
  
  virtual ~CDaiFletcherPGM();

  ErrorCode Solve();
  ErrorCode Update(CMatVecArr* a, Real b);
  ErrorCode GetSolution(CMatVecArr* &w, Real &xi, Real &regval, Real &objval);

private: 
  Real EPS_SV, eps;  // for precision related
  int maxProjIter;
  int maxPGMIter;

  // some arrays used in Solve() can be preallocated in constructor
  int increment;
  int max_mem_size;
  int cur_mem_size;
  int *ipt, *ipt2, *uv;
  Real *g, *y, *tempv, *d, *Qd, *t, *xplus, *tplus, *sk, *yk;

     
  int ProjectDF(int n, Real *a, Real b, Real *c, Real *l, 
                Real *u, Real *x, Real &lam_ext);

  Real ProjectR(Real *x, int n, Real lambda, Real *a, 
                  Real b, Real *c, Real *l, Real *u);

  void ReallocMem(int n);

#ifdef DEBUG
  void MatrixCorrectnessCheck();
#endif

};

#endif
