/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 07/02/2007
 * Last Updated :
 * 
 * Purpose      : L1-norm Linear Program solver using COIN-OR Clp 
 *
 */

#ifndef _L1CLP_CPP_
#define _L1CLP_CPP_


#include "l1clp.hpp"
#include <vector>

using namespace std;

// Clp headers
#include "CoinBuild.hpp"


/*
    L1-Norm linear program:

    minimize   :  d'x_bar
     x_bar
     
    subject to : | -1_cm  A_mn  -A_mn | |\xi|    | 0|
                                        | v | <= | 0| 
                                        | w |    |-B|


    where      : d := [C; lambda; lambda] (length: [1,n,n])
                 C is a positive scalar
                 lambda is a column vector of size n
                 x_bar := [\xi; u; v]
		 u,v column vectors of size n
                 1_c# is a column vector of 1 of size #,
                 A_#* is a dense matrix of size #-by-*, and
                 m << n (m starts from 1 and increase by 1 after 
                 each lp optimization round.)

*/


CL1Clp::CL1Clp(CInfo* _info,  Real _tol)
  :newRowIndices(0),
   newRowElements(0)
{
  // chteo: _tol is not used at the moment. we use default tol. in Clp

  // parameters
  iter = 0;
  info = _info;
  verbosity = info->verbosity;
  n = info->dimW;
  m = 0;
  x = 0;
  Q = 0;
  f = 0;
  a = 0;
  b = 0;
  l = 0;
  u = 0;

  
  // build simplex model
  sim = new ClpSimplex();
  sim->setLogLevel(0);
  sim->resize(0, 1+2*n);

  // set objective coefficients    
  sim->setObjCoeff(0, 1.0);
  for(int i=1; i<=n+n; i++) 
    sim->setObjCoeff(i,info->lambda);
  
  // set bounds for variables (i.e., columns as in COIN Clp terminology)
  for (int i=0; i<n+n+1; i++)
    sim->setColumnBounds(i, 0.0, COIN_DBL_MAX);
  

  
  // something useful in constraint matrix update phase
  newRowIndices = (int*)calloc(2*n+1, sizeof(int));
  newRowElements = (double*)calloc(2*n+1, sizeof(double));    
  
  for(int i=0; i<n+n+1; i++)
    newRowIndices[i] = i;
  
  newRowElements[0] = -1.0;
  // Update() set the rest of the elements in newRowElements[]
    
}


CL1Clp::~CL1Clp()
{
  if(sim) delete sim;
  if(newRowIndices) free(newRowIndices);
  if(newRowElements) free(newRowElements);
  
}

  
ErrorCode 
CL1Clp::Solve()
{
  sim->dual();
  //sim->primal();

  return NOERROR;
}


/**  Update matrix and vectors needed in optimisation round.
 *
 *   \param a [read] Gradient
 *   \param b [read] Offset 
 */
ErrorCode 
CL1Clp::Update(CMatVecArr* a, Real b)
{
  iter++;
 
  // addRow
  memcpy(newRowElements+1, a->array, a->row*a->col*sizeof(Real));
  memcpy(newRowElements+1+n, a->array, a->row*a->col*sizeof(Real));
  for(int i=1+n; i<=n+n; i++) newRowElements[i] *= -1;

  sim->addRow(1+n+n, newRowIndices, newRowElements, -COIN_DBL_MAX, -b);
  m++;
  
  //printf("num cols : %d\n", sim->numberColumns());
  //printf("num rows : %d\n", sim->numberRows());
  
  // free the gradient
  delete a;

  return NOERROR;
}



/** Return solution.
 *
 *  \param w      [write] solution vector (preallocated)
 *  \param xi     [write] lower bound of empirical risk, R_emp[w]
 *  \param regval [write] value of regularization term
 *  \param objval [write] objective value
 */
ErrorCode 
CL1Clp::GetSolution(CMatVecArr* &w, Real &xi, Real &regval, Real &objval)
{
  x = sim->primalColumnSolution(); 
  assert(x);  
  assert(w->row > 0 && w->col > 0);
  
  // construct new w
  int length = w->row*w->col;
  for(int i=0; i<length; i++) {
    w->array[i] = x[1+i] - x[1+n+i];   // w := u-v
  }

  // compute xi, objective value and regularizer value
  xi = x[0];
  objval = sim->objectiveValue();
  regval = 0;
  for(int i=0; i<length; i++)
    regval += ABS(w->array[i]);
  regval *= info->lambda;
  
  return NOERROR;
}

#endif
