/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Alex J. Smola (Alex.Smola@nicta.com.au)
 *                Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 24/01/2007
 * Last Updated :
 * 
 * Purpose      : solves quadratic programming problem
 *
 * Note         : adapted from pr_loqo
 */

#ifndef _PRLOQO_CPP
#define _PRLOQO_CPP

#include <cmath>
#include <cassert>
#include "prloqo.hpp"



#define PREDICTOR 1
#define CORRECTOR 2



CprLOQO::CprLOQO(CInfo* _info, Real _tol)
  :sigfig_max(-log10(_tol)),
   counter_max(5000),
   increment(50),
   max_mem_size(_info->memSize),  // max num of constraints
   cur_mem_n(1),
   cur_mem_m(1),
   b_local(0),
   primal(0),
   dual(0),
   workspace(0),
   diag_h_x(0),
   h_x(0),
   h_y(0),
   c_x(0),
   c_y(0),
   h_dot_x(0),
   rho(0),
   nu(0),
   tau(0),
   alpha(0),
   beta(0),
   sigma(0),
   gamma_z(0),
   gamma_w(0),
   gamma_s(0),
   gamma_q(0),
   hat_nu(0),
   hat_tau(0),
   hat_alpha(0),
   hat_beta(0),
   delta_x(0),
   delta_y(0),
   delta_s(0),
   delta_z(0),
   delta_g(0),
   delta_t(0),
   delta_w(0),
   delta_p(0),
   delta_q(0),
   delta_v(0),
   d(0),
   e(0)
{
  // inherited from CInnerSolver
  iter      = 0;
  info      = _info;
  tol       = _tol;
  verbosity = info->verbosity;
  n         = 0;                   // dimension of solution (increasing)
  m         = 1;                   // number of eq or ineq constraint  (fixed)

  aSet.resize(max_mem_size);                // prepare space for gradients
  bSet.resize(max_mem_size);                // prepare space for offsets
  //for(int i=0; i<max_mem_size; i++) aSet[i] = 0;

  timeStamp.resize(max_mem_size, -1);

  //n = info->memSize;    // in dual: solution dimension
  //m = 1;                // in dual: # of eq or ineq constraint

  // allocate memory for parent's member variables
  //x = 0;          // x is the first part of #primal#
  x = (Real*)calloc(max_mem_size,    sizeof(Real));
  Q = (Real*)calloc(max_mem_size*max_mem_size, sizeof(Real));
  f = (Real*)calloc(max_mem_size,    sizeof(Real));
  l = (Real*)calloc(max_mem_size,    sizeof(Real));
  u = (Real*)calloc(max_mem_size,    sizeof(Real));
  a = (Real*)calloc(max_mem_size,    sizeof(Real));
  b = new Real(info->C);        
  r = -100000;    // the #r# variable in LOQO

  // initializations
  for(int i=0; i<max_mem_size; i++) {
    l[i] = 0;
    u[i] = INFTY;
    a[i] = 1;      
  }
    

  // pre-allocate mem needed by solver
  ReallocPrimalMem(cur_mem_n);
  ReallocDualMem(cur_mem_m);
}



CprLOQO::~CprLOQO()
{
  // free for parent
  if(Q)         free(Q);
  if(x)         free(x);
  if(f)         free(f);
  if(l)         free(l);
  if(u)         free(u);
  if(a)         free(a);
  if(b)         delete b;

  // free memory
  if(b_local)   free(b_local);
  if(primal)    free(primal);
  if(dual)      free(dual);
  if(workspace) free(workspace);
  if(diag_h_x)  free(diag_h_x);
  if(h_x)       free(h_x);
  if(h_y)       free(h_y);
  if(c_x)       free(c_x);
  if(c_y)       free(c_y);
  if(h_dot_x)   free(h_dot_x);
  if(rho)       free(rho);
  if(nu)        free(nu);
  if(tau)       free(tau);
  if(alpha)     free(alpha);
  if(beta)      free(beta);
  if(sigma)     free(sigma);
  if(gamma_z)   free(gamma_z);
  if(gamma_w)   free(gamma_w);
  if(gamma_s)   free(gamma_s);
  if(gamma_q)   free(gamma_q);
  if(hat_nu)    free(hat_nu);
  if(hat_tau)   free(hat_tau);
  if(hat_alpha) free(hat_alpha);
  if(hat_beta)  free(hat_beta);
  if(delta_x)   free(delta_x);
  if(delta_y)   free(delta_y);
  if(delta_s)   free(delta_s);
  if(delta_z)   free(delta_z);
  if(delta_g)   free(delta_g);
  if(delta_t)   free(delta_t);
  if(delta_w)   free(delta_w);
  if(delta_p)   free(delta_p);
  if(delta_q)   free(delta_q);
  if(delta_v)   free(delta_v);
  if(d)         free(d);
  if(e)         free(e);
 
  for(int i=0; i<n; i++)
    delete aSet[i];
}

/** Allocate internal memory dual problem related variables
 *
 */
void CprLOQO::ReallocDualMem(int m)
{
  int n = cur_mem_n;

  b_local   = (Real*)realloc(b_local,   m*sizeof(Real));
  dual      = (Real*)realloc(dual,      (5*m)*sizeof(Real));
  workspace = (Real*)realloc(workspace, (n*(m+2)+2*m)*sizeof(Real));
  h_y       = (Real*)realloc(h_y,       m*m*sizeof(Real));
  c_y       = (Real*)realloc(c_y,       m*sizeof(Real));
  rho       = (Real*)realloc(rho,       m*sizeof(Real));
  alpha     = (Real*)realloc(alpha,     m*sizeof(Real));
  beta      = (Real*)realloc(beta,      m*sizeof(Real));
  gamma_w   = (Real*)realloc(gamma_w,   m*sizeof(Real));
  gamma_q   = (Real*)realloc(gamma_q,   m*sizeof(Real));
  hat_alpha = (Real*)realloc(hat_alpha, m*sizeof(Real));
  hat_beta  = (Real*)realloc(hat_beta,  m*sizeof(Real));
  delta_y   = (Real*)realloc(delta_y,   m*sizeof(Real));
  delta_w   = (Real*)realloc(delta_w,   m*sizeof(Real));
  delta_p   = (Real*)realloc(delta_p,   m*sizeof(Real));
  delta_q   = (Real*)realloc(delta_q,   m*sizeof(Real));
  delta_v   = (Real*)realloc(delta_v,   m*sizeof(Real));
  e         = (Real*)realloc(e,         m*sizeof(Real));

  cur_mem_m = m;
}

/** Allocate internal memory primal problem related variables
 *
 */
void CprLOQO::ReallocPrimalMem(int n)
{
  int m = cur_mem_m;

  h_x       = (Real*)realloc(h_x,       n*n*sizeof(Real));
  primal    = (Real*)realloc(primal,    4*n*sizeof(Real));
  workspace = (Real*)realloc(workspace, (n*(m+2)+2*m)*sizeof(Real));
  diag_h_x  = (Real*)realloc(diag_h_x,  n*sizeof(Real));
  c_x       = (Real*)realloc(c_x,       n*sizeof(Real));
  h_dot_x   = (Real*)realloc(h_dot_x,   n*sizeof(Real));
  nu        = (Real*)realloc(nu,        n*sizeof(Real));
  tau       = (Real*)realloc(tau,       n*sizeof(Real));
  sigma     = (Real*)realloc(sigma,     n*sizeof(Real));
  gamma_z   = (Real*)realloc(gamma_z,   n*sizeof(Real));
  gamma_s   = (Real*)realloc(gamma_s,   n*sizeof(Real));
  hat_nu    = (Real*)realloc(hat_nu,    n*sizeof(Real));
  hat_tau   = (Real*)realloc(hat_tau,   n*sizeof(Real));
  delta_x   = (Real*)realloc(delta_x,   n*sizeof(Real));
  delta_s   = (Real*)realloc(delta_s,   n*sizeof(Real));
  delta_z   = (Real*)realloc(delta_z,   n*sizeof(Real));
  delta_g   = (Real*)realloc(delta_g,   n*sizeof(Real));
  delta_t   = (Real*)realloc(delta_t,   n*sizeof(Real));
  d         = (Real*)realloc(d,         n*sizeof(Real));

  cur_mem_n = n;
}


// /** Utility functions
//  */
// void printmat(Real a[], int row, int col)
// {
//   printf("matrix: %d x %d\n",row,col);
//   for(int i=0; i<row; i++)
//   {
//     for(int j=0; j<col; j++)
//       printf("% .3e  ",a[i*col + j]);
//     printf("\n");
//   }   
// }

// void printvec(Real a[], int n)
// {
//   printf("vector: %d x 1\n",n);
//   for(int i=0; i<n; i++)
//     printf("%2d:%.8f\n",i,a[i]);
// }


/*****************************************************************
   taken from numerical recipes and modified to accept pointers
   moreover numerical recipes code seems to be buggy (at least the
   ones on the web)

   cholesky solver and backsubstitution
   leaves upper right triangle intact (rows first order)
   ***************************************************************/

void CprLOQO::choldc(Real a[], int n, Real p[])
{
  int i, j, k;
  Real sum;

  //printf("in choldc...\n");

  for (i = 0; i < n; i++){
    for (j = i; j < n; j++) {
      sum=a[n*i + j];
      for (k=i-1; k>=0; k--) sum -= a[n*i + k]*a[n*j + k];
      if (i == j) {
        if (sum <= 0.0) {
          
          printf("choldc failed, matrix not positive definite!\n");
          printf("sum : %.16f\n",sum);
          //printmat(a,n,n);
          exit(0);
        }
        p[i]=sqrt(sum);
      } 
      else 
        a[n*j + i] = sum/p[i];
    }
  }
}


/** solve linear system using cholesky dc output
 *
 */
void CprLOQO::cholsb(Real a[], int n, Real p[], Real b[], Real x[])
{
  int i, k;
  Real sum;

  for (i=0; i<n; i++) {
    sum=b[i];
    for (k=i-1; k>=0; k--) sum -= a[n*i + k]*x[k];
    x[i]=sum/p[i];
  }

  for (i=n-1; i>=0; i--) {
    sum=x[i];
    for (k=i+1; k<n; k++) sum -= a[n*k + i]*x[k];
    x[i]=sum/p[i];
  }
}

/*****************************************************************
  sometimes we only need the forward or backward pass of the
  backsubstitution, hence we provide these two routines separately 
  ***************************************************************/

void CprLOQO::chol_forward(Real a[], int n, Real p[], Real b[], Real x[])
{
  int i, k;
  Real sum;

  for (i=0; i<n; i++) {
    sum=b[i];
    for (k=i-1; k>=0; k--) sum -= a[n*i + k]*x[k];
    x[i]=sum/p[i];
  }
}

void CprLOQO::chol_backward(Real a[], int n, Real p[], Real b[], Real x[])
{
  int i, k;
  Real sum;

  for (i=n-1; i>=0; i--) {
    sum=b[i];
    for (k=i+1; k<n; k++) sum -= a[n*k + i]*x[k];
    x[i]=sum/p[i];
  }
}

/*****************************************************************
  solves the system | -H_x A' | |x_x| = |c_x|
                    |  A   H_y| |x_y|   |c_y|

  with H_x (and H_y) positive (semidefinite) matrices
  and n, m the respective sizes of H_x and H_y

  for variables see pg. 48 of notebook or do the calculations on a
  sheet of paper again

  predictor solves the whole thing, corrector assues that H_x didn't
  change and relies on the results of the predictor. therefore do
  _not_ modify workspace

  if you want to speed tune anything in the code here's the right
  place to do so: about 95% of the time is being spent in
  here. something like an iterative refinement would be nice,
  especially when switching from double to single precision. if you
  have a fast parallel cholesky use it instead of the numrec
  implementations.

  side effects: 1. changes H_y (but this is just the unit matrix 
                   or zero anyway in our case)
                2. changes H_x (due to choldc)
  ***************************************************************/

void CprLOQO::solve_reduced(int n, int m, Real h_x[], Real h_y[], 
                            Real a[], Real x_x[], Real x_y[],
                            Real c_x[], Real c_y[],
                            Real workspace[], int step)
{
  int i,j,k;

  Real *p_x;
  Real *p_y;
  Real *t_a;
  Real *t_c;
  Real *t_y;

  p_x = workspace;		/* together n + m + n*m + n + m = n*(m+2)+2*m */
  p_y = p_x + n;
  t_a = p_y + m;
  t_c = t_a + n*m;
  t_y = t_c + n;

  if (step == PREDICTOR) {
    choldc(h_x, n, p_x);	/* do cholesky decomposition */

    for (i=0; i<m; i++)         /* forward pass for A' */
      chol_forward(h_x, n, p_x, a+i*n, t_a+i*n);
				
    for (i=0; i<m; i++)         /* compute (h_y + a h_x^-1A') */
      for (j=i; j<m; j++)
        for (k=0; k<n; k++) 
          h_y[m*i + j] += t_a[n*j + k] * t_a[n*i + k];
				
    choldc(h_y, m, p_y);	/* and cholesky decomposition */
  }
  
  chol_forward(h_x, n, p_x, c_x, t_c);   /* forward pass for c */

  for (i=0; i<m; i++) {		/* and solve for x_y */
    t_y[i] = c_y[i];
    for (j=0; j<n; j++)
      t_y[i] += t_a[i*n + j] * t_c[j];
  }

  cholsb(h_y, m, p_y, t_y, x_y);

  for (i=0; i<n; i++) {		/* finally solve for x_x */
    t_c[i] = -t_c[i];
    for (j=0; j<m; j++)
      t_c[i] += t_a[j*n + i] * x_y[j];
  }

  chol_backward(h_x, n, p_x, t_c, x_x);
}

/*****************************************************************
  matrix vector multiplication (symmetric matrix but only one triangle
  given). computes m*x = y
  no need to tune it as it's only of O(n^2) but cholesky is of
  O(n^3). so don't waste your time _here_ although it isn't very
  elegant. 
  ***************************************************************/

void CprLOQO::matrix_vector(int n, Real m[], Real x[], Real y[])
{
  int i, j;

  for (i=0; i<n; i++) {
    y[i] = m[(n+1) * i] * x[i];

    for (j=0; j<i; j++)
      y[i] += m[i + n*j] * x[j];

    for (j=i+1; j<n; j++) 
      y[i] += m[n*i + j] * x[j]; 
  }
}

/*****************************************************************
  call only this routine; this is the only one you're interested in
  for doing quadratical optimization

  the restart feature exists but it may not be of much use due to the
  fact that an initial setting, although close but not very close the
  the actual solution will result in very good starting diagnostics
  (primal and dual feasibility and small infeasibility gap) but incur
  later stalling of the optimizer afterwards as we have to enforce
  positivity of the slacks.
  ***************************************************************/
ErrorCode CprLOQO::Solve()
{

  // check if pre allocated mem is enough
  if(n > cur_mem_n)
    ReallocPrimalMem(n+increment);
  if(m > cur_mem_m)
    ReallocDualMem(m+increment);


  /* the knobs to be tuned ... */
  Real margin = -0.97;	/* we will go up to 95% of the
                               distance between old variables and zero */
  Real bound = 10;		/* preset value for the start. small
                               values give good initial
                               feasibility but may result in slow
                               convergence afterwards: we're too
                               close to zero */

  // bridge between wrapper and original pr_loqo impl
  Real *c = f;
  
  // h_x get changed after each predictor-corrector pair call
  // alternatively, copy the upper triangle back to the lower one
  memcpy(h_x, Q, n*n*sizeof(Real));

  // adjust range of constraint
  Real base = (r<0) ? -r : r;
  for(int i=0; i<m; i++) {
    b_local[i] = base + b[i];
    //printf("b_local:%.8f\n",b_local[i]);
  }

  /* from the header - pointers into primal and dual */
  //Real *x;  // using member variable #x#
  Real *y;
  Real *g;
  Real *z;
  Real *s;
  Real *t;

  // for ineq constraint
  Real *v;
  Real *w;
  Real *p;  
  Real *q;


  /* auxiliary variables */
  Real b_plus_1;
  Real c_plus_1;

  Real x_h_x;
  Real primal_inf;
  Real dual_inf;

  Real sigfig;
  Real primal_obj, dual_obj;
  Real mu;
  Real steplength = 0;
  int counter = 0;

  ErrorCode status = 0;

  int i,j;

  // chteo: dont mix x with other primal variables anymore
  //x = primal;			/* n */
  //g = x + n;			/* n */
  

  g = primal;
  t = g + n;			/* n */
  z = t + n;			/* n */
  s = z + n;			/* n */

  y = dual;			    /* m */
  w = y + m;		    /* m */
  p = w + m;		    /* m */
  q = p + m;		    /* m */
  v = q + m;		    /* m */
  
  

  /* initial settings */
  c_plus_1 = 0;
  for(i=0; i<n; i++)
    c_plus_1 += c[i]*c[i];
  c_plus_1 = sqrt(c_plus_1) + 1;

  b_plus_1 = 0;
  for(i=0; i<m; i++)
    b_plus_1 += b[i]*b[i];
  b_plus_1 = sqrt(b_plus_1) + 1;


  /* get diagonal terms */
  for (i=0; i<n; i++) diag_h_x[i] = h_x[(n+1)*i]; 
  
  // chteo: no more restart, use default start settings
  for (i=0; i<m; i++)
    for (j=i; j<m; j++)
      h_y[i*m + j] = (i==j) ? 1 : 0;
  
  for (i=0; i<n; i++) {
    c_x[i] = c[i];
    h_x[(n+1)*i] += 1;
  }
  
  for (i=0; i<m; i++)
    c_y[i] = b_local[i];
   
  /* and solve the system [-H_x A'; A H_y] [x, y] = [c_x; c_y] */
  solve_reduced(n, m, h_x, h_y, a, x, y, c_x, c_y, workspace,
                PREDICTOR);

  
  /* initialize the other variables (primal) */
  for (i=0; i<n; i++) {
    g[i] = MAX(ABS(x[i] - l[i]), bound);
    z[i] = MAX(ABS(x[i]), bound);
    t[i] = MAX(ABS(u[i] - x[i]), bound); 
    s[i] = MAX(ABS(x[i]), bound); 
  }

  // chteo:
  /* initialize the other variables (dual) */
  for (i=0; i<m; i++) {    
    v[i] = MAX(ABS(y[i]), bound);
    w[i] = MAX(ABS(y[i]), bound);
    q[i] = MAX(ABS(y[i]), bound); 
    p[i] = MAX(ABS(b_local[i] - w[i]), bound);  // the roles of r and b are swapped
  }

  mu = 0;
  for(i=0; i<n; i++)
    mu += z[i] * g[i] + s[i] * t[i];
  for(i=0; i<m; i++)
    mu += v[i] * w[i] + p[i] * q[i];
  mu = mu / (2*(m+n));
  sigfig = 0;
  steplength = 1;

  /* the main loop */
  if (verbosity >= STATUS) {
    printf("counter | pri_inf  | dual_inf  | pri_obj   | dual_obj  | ");
    printf("sigfig | alpha  | nu \n");
    printf("-------------------------------------------------------");
    printf("---------------------------\n");
  }
  
  while (status == 0) {
    /* predictor */
    
    /* put back original diagonal values */
    for (i=0; i<n; i++) 
      h_x[(n+1) * i] = diag_h_x[i];

    matrix_vector(n, h_x, x, h_dot_x); /* compute h_dot_x = h_x * x */

    for (i=0; i<m; i++) {
      rho[i] = r + w[i];
      for (j=0; j<n; j++)
        rho[i] -= a[n*i + j] * x[j];
    }

    for (i=0; i<n; i++) {
      nu[i] = l[i] - x[i] + g[i];
      tau[i] = u[i] - x[i] - t[i];

      sigma[i] = c[i] - z[i] + s[i] + h_dot_x[i];
      for (j=0; j<m; j++)
        sigma[i] -= a[n*j + i] * y[j];

      gamma_z[i] = - z[i];
      gamma_s[i] = - s[i];
    }

    for(i=0; i<m; i++) {
      alpha[i] = b_local[i] - w[i] - p[i];  // the roles of r and b swapped 
      beta[i]  = y[i] + q[i] - v[i];

      gamma_w[i] = - w[i];
      gamma_q[i] = - q[i];
    }

    /* instrumentation */
    x_h_x = 0;
    primal_inf = 0;
    dual_inf = 0;
    
    for (i=0; i<n; i++) {
      x_h_x += h_dot_x[i] * x[i];
      primal_inf += SQR(tau[i]);
      primal_inf += SQR(nu[i]);
      dual_inf += SQR(sigma[i]);
    }

    for (i=0; i<m; i++) {
      primal_inf += SQR(rho[i]);
      primal_inf += SQR(alpha[i]);
      dual_inf += SQR(beta[i]);
    }

    primal_inf = sqrt(primal_inf)/b_plus_1;
    dual_inf = sqrt(dual_inf)/c_plus_1;

    primal_obj = 0.5 * x_h_x;
    dual_obj = -0.5 * x_h_x;
    for (i=0; i<n; i++) {
      primal_obj += c[i] * x[i];
      dual_obj += l[i] * z[i] - u[i] * s[i];
    }
    for (i=0; i<m; i++) {
      dual_obj += r * y[i];     // the roles of r and b swapped
      dual_obj -= b_local[i] * q[i];
    }

    sigfig = log10(ABS(primal_obj) + 1) -
             log10(ABS(primal_obj - dual_obj));
    sigfig = MAX(sigfig, 0);
		   
    /* the diagnostics - after we computed our results we will
       analyze them */

    if (counter > counter_max) status = ITERATION_LIMIT;
    //if (sigfig  > sigfig_max)  status = OPTIMAL_SOLUTION;
    if (sigfig  > -log10(tol))  status = OPTIMAL_SOLUTION;
    if (primal_inf > 10e100)   status = PRIMAL_INFEASIBLE;
    if (dual_inf > 10e100)     status = DUAL_INFEASIBLE;
    if ((primal_inf > 10e100) & (dual_inf > 10e100)) status = PRIMAL_AND_DUAL_INFEASIBLE;
    if (ABS(primal_obj) > 10e100) status = PRIMAL_UNBOUNDED;
    if (ABS(dual_obj) > 10e100) status = DUAL_UNBOUNDED;

    /* write some nice routine to enforce the time limit if you
       _really_ want, however it's quite useless as you can compute
       the time from the maximum number of iterations as every
       iteration costs one cholesky decomposition plus a couple of 
       backsubstitutions */

    /* generate report */
    if ((verbosity >= FLOOD) | ((verbosity == STATUS) & (status != 0)))
      printf("%7i | %.2e | %.2e | % .2e | % .2e | %6.3f | %.4f | %.2e\n",
             counter, primal_inf, dual_inf, primal_obj, dual_obj,
             sigfig, steplength, mu);

    counter++;

    if (status == 0) {		
      /* 
         we may keep on going, otherwise it'll cost one loop extra plus a
         messed up main diagonal of h_x 
      */

      /* intermediate variables (the ones with hat) */
      for (i=0; i<n; i++) {
        hat_nu[i]    = nu[i] + g[i] * gamma_z[i] / z[i];
        hat_tau[i]   = tau[i] - t[i] * gamma_s[i] / s[i];
        /* diagonal terms */
        d[i] = z[i] / g[i] + s[i] / t[i];
      }
      for(i=0; i<m; i++) {
        hat_beta[i]  = beta[i] - v[i] * gamma_w[i] / w[i];
        hat_alpha[i] = alpha[i] - p[i] * gamma_q[i] / q[i];
      }
      
      /* initialization before the cholesky solver */
      for (i=0; i<n; i++) {
        h_x[(n+1)*i] = diag_h_x[i] + d[i];
        c_x[i] = sigma[i] - z[i] * hat_nu[i] / g[i] - 
          s[i] * hat_tau[i] / t[i];
      }           
      
      memset(h_y, 0, m*m*sizeof(Real));
      for (i=0; i<m; i++) {
        c_y[i] = rho[i];
        e[i] = 1 / (v[i] / w[i] + q[i] / p[i]);
        h_y[(m+1)*i] = e[i];
      }
      
          
      /* and do it */
      solve_reduced(n, m, h_x, h_y, a, delta_x, delta_y, c_x, c_y, workspace,
                    PREDICTOR);
      
      // chteo: for PRIMAL related var
      for(i=0; i<n; i++) {
        /* backsubstitution */
        delta_s[i] = s[i] * (delta_x[i] - hat_tau[i]) / t[i];
        delta_z[i] = z[i] * (hat_nu[i] - delta_x[i]) / g[i];        
        delta_g[i] = g[i] * (gamma_z[i] - delta_z[i]) / z[i];
        delta_t[i] = t[i] * (gamma_s[i] - delta_s[i]) / s[i];
        
        /* central path (corrector) */
        gamma_z[i] = mu / g[i] - z[i] - delta_z[i] * delta_g[i] / g[i];
        gamma_s[i] = mu / t[i] - s[i] - delta_s[i] * delta_t[i] / t[i];
        
        /* (some more intermediate variables) the hat variables */
        hat_nu[i] = nu[i] + g[i] * gamma_z[i] / z[i];
        hat_tau[i] = tau[i] - t[i] * gamma_s[i] / s[i];
        
        /* initialization before the cholesky */
        c_x[i] = sigma[i] - z[i] * hat_nu[i] / g[i] - s[i] * hat_tau[i] / t[i];
      }

      // chteo: for DUAL related var
      for(i=0; i<m; i++) {
        /* backsubstitution */
        delta_w[i] = - e[i] * (hat_beta[i] - q[i] * hat_alpha[i] / p[i] + delta_y[i]);
        delta_q[i] = q[i] * (delta_w[i] - hat_alpha[i]) / p[i];

        delta_v[i] = v[i] * (gamma_w[i] - delta_w[i]) / w[i];
        delta_p[i] = p[i] * (gamma_q[i] - delta_q[i]) / q[i];


        /* central path (corrector) */
        gamma_w[i] = mu / v[i] - w[i] - delta_w[i] * delta_v[i] / v[i];
        gamma_q[i] = mu / p[i] - q[i] - delta_q[i] * delta_p[i] / p[i];

        /* (some more intermediate variables) the hat variables */
        hat_alpha[i] = alpha[i] - p[i] * gamma_q[i] / q[i];
        hat_beta[i]  = beta[i] - v[i] * gamma_w[i] / w[i];

        /* initialization before the cholesky */
        c_y[i] = rho[i] - e[i] * (hat_beta[i] - q[i] * hat_alpha[i] / p[i]);
      }

    
      /* and do it */
      solve_reduced(n, m, h_x, h_y, a, delta_x, delta_y, c_x, c_y, workspace,
                    CORRECTOR);
      
      // chteo: for primal related var
      for (i=0; i<n; i++) {
        /* backsubstitution */
        delta_s[i] = s[i] * (delta_x[i] - hat_tau[i]) / t[i];
        delta_z[i] = z[i] * (hat_nu[i] - delta_x[i]) / g[i];        
        delta_g[i] = g[i] * (gamma_z[i] - delta_z[i]) / z[i];
        delta_t[i] = t[i] * (gamma_s[i] - delta_s[i]) / s[i];
      }
      
      // chteo: for dual related var
      for(i=0; i<m; i++) {
        /* backsubstitution */
        delta_w[i] = - e[i] * (hat_beta[i] - q[i] * hat_alpha[i] / p[i] + delta_y[i]);
        delta_q[i] = q[i] * (delta_w[i] - hat_alpha[i]) / p[i];
        delta_v[i] = v[i] * (gamma_w[i] - delta_w[i]) / w[i];
        delta_p[i] = p[i] * (gamma_q[i] - delta_q[i]) / q[i];
      }

      steplength = -1;
      for (i=0; i<n; i++) {
        steplength = MIN(steplength, delta_g[i]/g[i]);
        steplength = MIN(steplength, delta_t[i]/t[i]);
        steplength = MIN(steplength, delta_s[i]/s[i]);
        steplength = MIN(steplength, delta_z[i]/z[i]);
      }
      for (i=0; i<m; i++) {
        steplength = MIN(steplength, delta_w[i]/w[i]);
        steplength = MIN(steplength, delta_p[i]/p[i]);
        steplength = MIN(steplength, delta_q[i]/q[i]);
        steplength = MIN(steplength, delta_v[i]/v[i]);
      }  
      steplength = margin / steplength;
      
      /* compute mu */
      mu = 0;
      for(i=0; i<n; i++)
        mu += z[i] * g[i] + s[i] * t[i];
      for(i=0; i<m; i++)
        mu += v[i] * w[i] + p[i] * q[i];            
      mu = mu / (2*(m+n));
      mu = mu * SQR((steplength - 1) / (steplength + 10));
      
      for (i=0; i<n; i++) {
        x[i] += steplength * delta_x[i];
        g[i] += steplength * delta_g[i];
        t[i] += steplength * delta_t[i];
        z[i] += steplength * delta_z[i];
        s[i] += steplength * delta_s[i];
      }

      for (i=0; i<m; i++) {
        y[i] += steplength * delta_y[i];
        w[i] += steplength * delta_w[i];
        p[i] += steplength * delta_p[i];
        q[i] += steplength * delta_q[i];
        v[i] += steplength * delta_v[i];
      }
    }
  }
  if ((status == 1) && (verbosity >= STATUS)) {
    printf("----------------------------------------------------------------------------------\n");
    printf("optimization converged\n");
  }

  // chteo:
  // h_x get changed after each predictor-corrector pair call
  // alternatively, copy the upper triangle back to the lower one
  memcpy(h_x, Q, n*n*sizeof(Real));


  return status;
}


/**  Update matrix and vectors needed in optimisation round.
 *
 *   \param a [read] Gradient
 *   \param b [read] Offset 
 */
ErrorCode 
CprLOQO::Update(CMatVecArr* a, Real b)
{
  int how_old = 40;                  // constraint not used in the last #how_old# iteration will be replaced by new one.
  int idx = -1;                      // default: look for i such that x[i] is smallest
  int first_new_slot = -1;
  int oldest_slot = 0;
  int smallest_slot = 0;
  
  iter++;                            // another column generated

  // deciding which slot to store the latest gradient
  for(int i=0; i<max_mem_size; i++) {
    if(timeStamp[i] == -1 && first_new_slot == -1) first_new_slot = i;
    if(timeStamp[i] < timeStamp[oldest_slot] && timeStamp[i] >= 0) oldest_slot = i; 
    if(x[i] < x[smallest_slot]) smallest_slot = i;
  }
  if(timeStamp[oldest_slot] < iter - how_old) 
    idx = oldest_slot;
  else if(first_new_slot != -1) {
    idx = first_new_slot;
    n++;
  }
  else 
    idx = smallest_slot;
  
  if(verbosity > 1) cout << "new column idx: " << idx << "   aSet size:" << aSet.size() << endl;

  // store new column at #idx#
  if(aSet[idx]) delete aSet[idx];
  aSet[idx] = a;                   // store the gradient
  bSet[idx] = b;                   // store the offset  


  // use new slot
  if(timeStamp[idx] == -1)
  {
    // chteo: if this is not active after immediate optimization, we still make to make it current since it is new!
    timeStamp[idx] = iter;  
    
    // insert new row into Q
    int lastRow = (n-1)*n;
    for(int col=0; col<n; col++) {
      PetscScalar value = 0;
      VecDot(aSet[idx]->vec, aSet[col]->vec, &value);
      Q[lastRow+col] = 1.0/info->lambda*value;
    }

    // adjust memory area to fit a new column into Q
    int prevDim = n-1;
    int size    = sizeof(Real)*prevDim;
    for(int row=n-2; row >= 1; row--)
    {
      Real *src  = Q + (row*prevDim);
      Real *dest = Q + (row*prevDim) + row;
      memmove(dest, src, size);
    }

    // insert new column (at the right-most-side of the matrix)
    for(int i=0; i<n-1; i++)
      Q[i*n + n-1] = Q[lastRow + i];
       

    // update vectors
    f[n-1] = -bSet[n-1];
  }
  else 
  {      
    // chteo: if this is not active after immediate optimization, we still make to make it current since it is new!
    timeStamp[idx] = iter;  
    
    // insert new row into Q[idx,:]
    int theRow = idx*n;
    for(int col=0; col<n; col++) {
      PetscScalar value = 0;
      VecDot(aSet[idx]->vec, aSet[col]->vec, &value);
      Q[theRow+col] = 1.0/info->lambda*value;
    }
    
    // insert new column into Q[:,idx]
    for(int i=0; i<n; i++)
      Q[i*n + idx] = Q[theRow + i];
    
    // update vectors      
    f[idx] = -bSet[idx];
  }
  

#ifdef DEBUG
  // check matrix !
  MatrixCorrectnessCheck();
#endif

  return NOERROR;
}


/** Return solution.
 *
 *  \param w      [write] solution vector (preallocated)
 *  \param xi     [write] lower bound of empirical risk, R_emp[w]
 *  \param regval [write] value of regularization term
 *  \param objval [write] objective value
 */
ErrorCode 
CprLOQO::GetSolution(CMatVecArr* &w, Real &xi, Real &regval, Real &objval)
{
  assert(w->row > 0 && w->col > 0);
  assert(x != 0);
  Real factor = 1.0/info->lambda;


  objval = 0;
  Real tmp = 0;
  Real fx = 0;
  for(int i=0; i<n; i++) {
    tmp = 0;
    for(int j=0; j<n; j++)
      tmp += Q[j + i*n]*x[j];
    objval += x[i]*tmp;
    fx += f[i]*x[i];
  }

  // since the dual of minimization problem is maximization
  // and we solved the  -maximization version,
  // the objective value should be -objval
  objval = -0.5*objval - fx;


  // Compute new w
  VecZeroEntries(w->vec);  
  for(int i=0; i<n; i++)
    VecAXPY(w->vec, -x[i], aSet[i]->vec);
  VecScale(w->vec, factor);

  // compute xi
  VecDot(w->vec, w->vec, &regval);
  regval *= (info->lambda/2); 
  xi = (objval - regval)/info->C;

  // update time-stamp of constraints
  for(int i=0; i<n; i++)
    if(x[i] > 0) timeStamp[i] = iter;

  // constraint set size
  if(verbosity > 1)
    cout << "qp_size: " << n << endl;

  // display nonzero terms
  if(verbosity > 2) {
    int howmany=0;
    cout << "nz: ";
    for(int i=0; i<n; i++) {
      if(ABS(x[i]) > ZERO_EPS) {
        cout << i << "..";
        howmany++;
      }
    }
    cout << "   ("<<howmany<<")"<<endl;
  }

  return NOERROR;
}
 

#ifdef DEBUG
/** Check if the Q matrix update is correct
 */
void
CprLOQO::MatrixCorrectnessCheck()
{

  PetscScalar dot = 0;
  Real *correctmat = new Real[n*n];
  memset(correctmat, 0, sizeof(Real)*n*n);
  
  for(int i=0; i<n; i++)
  {
    for(int j=i; j<n; j++)
    {
      //cout << "i:" << i << "   j:" << j << endl;
      dot = 0;
      VecDot(aSet[i]->vec, aSet[j]->vec, &dot);
      correctmat[i*n + j] = 1.0/info->lambda*dot;
      correctmat[j*n + i] = correctmat[i*n+j];
    }
  }

  for(int i=0; i<n*n; i++)
  {
    //cout << i << "  Q:"<<Q[i]<< "    cm:"<<correctmat[i] << endl;
    if(fabs(correctmat[i] - Q[i]) > 1e-15) {
      cout << "residual : " << fabs(correctmat[i] - Q[i]) << endl;
      cout << "matrix update correctness check : i = " << i << endl;
      cout << "ERROR: rank-1 updated matrix is incorrect ! " << endl;
      exit(-1);
    }
  }
}
#endif

#endif
