/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Alex J. Smola (Alex.Smola@nicta.com.au)
 *                Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 24/01/2007
 * Last Updated :
 * 
 * Purpose      : solves quadratic programming problem for pattern recognition
 *                and regression estimation for support vectors
 *
 * Note         : adapted from pr_loqo
 */

#ifndef _PRLOQO_HPP_
#define _PRLOQO_HPP_

#include "errorcode.hpp"
#include "common.hpp"
#include "info.hpp"
#include "matvecarr.hpp"
#include "solver.hpp"

// from PETSc
#include "petsc.h"
#include "petscvec.h"


/* 
 * solve the quadratic programming problem
 *
 * minimize    c' * x + 1/2 x' * H * x
 * subject to  r <= A*x <= r+b
 *             l <= x <= u
 *
 *  for a documentation see R. Vanderbei, LOQO: an Interior Point Code
 *                          for Quadratic Programming
 *
 *  NOTE: the role of r and b are swapped to maintain
 *        consistency across different solvers.
 */

/*
 * n   : number of primal variables
 * m   : number of constraints (typically 1)
 * h_x : dot product matrix (n.n)
 * a   : constraint matrix (n.m)
 * b   : constant term (m)
 * l   : lower bound (n)
 * u   : upper bound (m)
 *
 * primal : workspace for primal variables, has to be of size 3 n
 *
 *  x = primal;			n
 *  g = x + n;			n
 *  t = g + n;			n
 *
 * dual : workspace for dual variables, has to be of size m + 2 n
 *
 *  y = dual;			m
 *  z = y + m;			n
 *  s = z + n;			n
 *
 * verb       : verbosity level
 * sigfig_max : number of significant digits
 * counter_max: stopping criterion
 * restart    : 1 if restart desired
 *
 */

class CprLOQO : public CSolver
{
public:

  CprLOQO() {}
  CprLOQO(CInfo *_info, Real _tol=1e-8);
  
  virtual ~CprLOQO();
  
  ErrorCode Solve();
  ErrorCode Update(CMatVecArr* a, Real b);
  ErrorCode GetSolution(CMatVecArr* &w, Real &xi, Real &regval, Real &objval);


private:
  Real sigfig_max;   // := -log10(tolerance)
  int    counter_max;
  
  int    increment;  
  int    max_mem_size;
  int    cur_mem_n;  // mem size for # of primal var
  int    cur_mem_m;  // mem size for # of dual var

  Real *b_local;   // modified constraint range

  // mem used in Solve()
  Real *primal;    // workspace for primal var
  Real *dual;      // workspace for dual var
  Real *workspace;
  Real *diag_h_x;
  Real *h_x;
  Real *h_y;
  Real *c_x;
  Real *c_y;
  Real *h_dot_x;
  Real *rho;
  Real *nu;
  Real *tau;
  Real *alpha;
  Real *beta;
  Real *sigma;
  Real *gamma_z;
  Real *gamma_w;  
  Real *gamma_s;  
  Real *gamma_q;  
  Real *hat_nu;
  Real *hat_tau;
  Real *hat_alpha;
  Real *hat_beta;
  Real *delta_x;
  Real *delta_y;
  Real *delta_s;
  Real *delta_z;
  Real *delta_g;
  Real *delta_t;
  Real *delta_w;
  Real *delta_p;
  Real *delta_q;
  Real *delta_v;
  Real *d;
  Real *e;

  void choldc(Real a[], int n, Real p[]);
  void cholsb(Real a[], int n, Real p[], Real b[], Real x[]);
  void chol_forward(Real a[], int n, Real p[], Real b[], Real x[]);
  void chol_backward(Real a[], int n, Real p[], Real b[], Real x[]);
  void matrix_vector(int n, Real m[], Real x[], Real y[]);
  void solve_reduced(int n, int m, Real h_x[], Real h_y[], 
                     Real a[], Real x_x[], Real x_y[],
                     Real c_x[], Real c_y[],
                     Real workspace[], int step);
  
  void ReallocPrimalMem(int n);
  void ReallocDualMem(int m);

#ifdef DEBUG
  void MatrixCorrectnessCheck();
#endif

};

#endif
