/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo (ChoonHui.Teo@anu.edu.au)
 * Created      : 16/11/2006
 * Last Updated :
 */

#ifndef _SOLVER_HPP_
#define _SOLVER_HPP_

#include "errorcode.hpp"
#include "common.hpp"
#include "info.hpp"
#include "matvecarr.hpp"
#include <iostream>
#include <vector>

using namespace std;


/** Abstract class for inner solver which solve the following type of QP:
 *
 *  minimise  0.5*x'*Q*x + f'*x
 *     x
 *
 *  subj to   l <= x <= u
 *            a'*x <= b
 *
 *  where     a, x, lb, and ub \in R^n
 *            b \in R
 *            Q is an empty matrix for linear program
 *
 */
class CSolver 
{
public:
  
  int  iter;        // number of times solver is called
  int  verbosity;   // program verbosity
  Real tol;         // solution precision

  int  n;           // dimension of solution
  int  m;           // number of constraints
  Real *x;          // solution
  Real *f;          // linear part of obj
  Real *Q;          // quad part of obj
  Real *a;          // constraint matrix
  Real *b;          // constraint rhs
  Real *l;          // solution lower bound
  Real *u;          // solution upper bound
  Real r;           // for interior point solver

  CInfo  *info;       // learning parameter & miscellaneous parameter

  vector<CMatVecArr*> aSet;       // gradient set
  vector<Real>        bSet;       // offset set
  vector<int>         timeStamp;  // timeStamp[i] := the last iter when x[i] > 0


  // Constructor
  CSolver()
    : iter(0), verbosity(1), tol(1e-10),
      n(0), m(0), x(0), f(0), Q(0), a(0), b(0), l(0), u(0), r(0),
      info(0)
  {}
  
  // Destructor
  virtual ~CSolver(){}
  
  // Solve the problem
  virtual ErrorCode Solve()=0;

  // Update internal data before Solve() calls
  virtual ErrorCode Update(CMatVecArr* a, Real b)=0;

  // Get solution and lower bound of empirical risk
  virtual ErrorCode GetSolution(CMatVecArr* &w, Real &xi, Real &regval, Real &objval)=0;
  
};

#endif


