/* Copyright (c) 2006, National ICT Australia
 * All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Authors      : Choon Hui Teo       (ChoonHui.Teo@nicta.com.au)
 *                Quoc V. Le          (Quoc.Le@nicta.com.au)
 *                Alex Smola          (Alex.Smola@nicta.com.au)
 *                S.V.N. Vishwanathan (SVN.Vishwanathan@nicta.com.au)        
 *
 * Created      : 18/12/2006
 *
 * Last Updated :
 */

#include <iostream>
#include <string>
#include <vector>
#include <cstdlib>
#include <iomanip>
#include <fstream>

#include "core_header.hpp"
#include "loss.hpp"

#include "petsc.h"
#include "petscmat.h"
#include "petscvec.h"


// loss functions
CLoss* SelectLoss(CInfo *info, CDataset *ds);



/* ===== START tao stuff ===== */
#include "tao.h"
static  char help[] = "tao bmrm\n";
/* ===== END TAO stuff ===== */


using namespace std;

/* ===== START tao stuff ===== */
typedef struct {
  int       verbosity;
  int       n;              // dimension
  double    C;              // margin-loss trade off
  CDataset *ds;            // dataset
  CLoss     *lossFunction;  // loss function
  double    lambda;         // coef of norm-2 regularizer

} AppCtx;

int Grad(TAO_APPLICATION taoapp, Vec X, double *f, Vec G, void *ptr);

/* ===== END TAO stuff ===== */

/** get rid of the trailing whitespaces
 */
void TrimWhitespaces(string &line)
{
  int cnt = line.size();
  while(isspace(line[cnt-1])) cnt--;
  line = line.substr(0, cnt);
}

/** Check if str is blank
 */
bool isBlankLine(string &str)
{
  int n = str.size();
  for(int i=0; i<n; i++)
    if(!isspace(str[i]))
      return false;
  return true;
}


int main(int argc, char **argv)
{
  CTimer totaltime;           /* total time */
  CTimer solvertime;          /* time to solve the problem only (exlucding data loading) */
  CInfo myinfo;

  totaltime.Start();
  ParseCmdLineArgTrain(argc, argv, &myinfo);
  

  
     
  /* ===== START TAO stuff ===== */
  int                info;                  /* used to check for functions returning nonzeros */
  PetscScalar        zero=0.0;
  Vec                x;                     /* solution vector */
  TAO_SOLVER         tao;                   /* TAO_SOLVER solver context */
  TAO_APPLICATION    taoapp;                /* TAO application context */
  int                size,rank;             /* number of processes running */
  TaoTerminateReason reason;
  AppCtx             user;                  /* user-defined application context */

  int(*Gradient)(TAO_APPLICATION,Vec,double*,Vec,void*);
   
  Gradient = &(Grad);

  /* Initialize TAO and PETSc */
  PetscInitialize(&argc,&argv,(char *)0,help);
  TaoInitialize(&argc,&argv,(char *)0,help);

  info = MPI_Comm_size(PETSC_COMM_WORLD,&size); CHKERRQ(info);
  info = MPI_Comm_rank(PETSC_COMM_WORLD,&rank); CHKERRQ(info);

  CDataset ds(&myinfo);
  CLoss *lossFunction = SelectLoss(&myinfo, &ds);


  if (size >1) {
    if (rank == 0)
      PetscPrintf(PETSC_COMM_SELF,"This example is intended for single processor use!\n");
    SETERRQ(1,"Incorrect number of processors");
  }


  /* Initialize problem parameters */
  user.n = myinfo.dimW; 
  user.ds = &ds;
  user.lossFunction = lossFunction;
  user.C = myinfo.C;
  user.verbosity = myinfo.verbosity;
  user.lambda = myinfo.lambda;


  /* Allocate vectors for the solution and gradient */
  info = VecCreateSeq(PETSC_COMM_SELF,user.n,&x); CHKERRQ(info);
  
  /* Create TAO solver with desired solution method */
  info = TaoCreate(PETSC_COMM_SELF,"tao_lmvm",&tao); CHKERRQ(info);
  info = TaoApplicationCreate(PETSC_COMM_SELF,&taoapp); CHKERRQ(info);
  
     
  // set lmvm storage size
  info = TaoLMVMSetSize(tao, myinfo.memSize); CHKERRQ(info);

  //info = TaoCreateArmijoLineSearch(tao); CHKERRQ(info);
  //info = TaoCreateUnitLineSearch(tao); CHKERRQ(info);
  
  
  /* Set solution vec and an initial guess */
  info = VecSet(x, zero); CHKERRQ(info);
  info = TaoAppSetInitialSolutionVec(taoapp,x); CHKERRQ(info); 

  /* Set routines for function, gradient, hessian evaluation */
  info = TaoAppSetObjectiveAndGradientRoutine(taoapp,Gradient,(void *)&user); 
  CHKERRQ(info);

  /* Check for TAO command line options */
  info = TaoSetOptions(taoapp,tao); CHKERRQ(info);

  // termination parameters
  info = TaoSetTolerances(tao,          // solver
                          myinfo.epsilonTol,     // fatol : abs. conv. tol.
                          myinfo.epsilonTol,     // frtol : rel. conv. tol.
                          0,            // catol : abs. constr. err. tol.
                          0             // crtol : rel. constr. err. tol.
                          ); CHKERRQ(info);
  
  info = TaoSetMaximumIterates(tao, myinfo.maxIter); CHKERRQ(info);

  solvertime.Start();
  /* SOLVE THE APPLICATION */
  info = TaoSolveApplication(taoapp,tao); CHKERRQ(info);
  TaoView(tao);

  // solution
  PetscScalar *solution;
  info = VecGetArray(x,&solution); CHKERRQ(info);

  // convert X to CsparseVec w
  CMatVecArr w(user.n, 1, solution);
  WriteModelFile(&w, &myinfo);
  //w.Assign((double*)solution, user.n);
  //w.ToFile(misc.modelFn+".tao");


  /* Get termination information */
  info = TaoGetTerminationReason(tao,&reason); CHKERRQ(info);
  if (reason <= 0)
    PetscPrintf(MPI_COMM_WORLD,"Try a different TAO method, adjust some parameters, or check the function evaluation routines\n");


  /* Free TAO data structures */
  info = TaoDestroy(tao); CHKERRQ(info);
  info = TaoAppDestroy(taoapp); CHKERRQ(info);

  /* Free PETSc data structures */
  info = VecDestroy(x); CHKERRQ(info);



  /* ===== END Tao stuff =====  */
  solvertime.Stop();
  totaltime.Stop();
  printf("-------------------------------------\n");
  printf("Total time: %3.2f seconds\n",totaltime.Total());
  printf("Solver time: %lf seconds\n", solvertime.Total());
  printf("-------------------------------------\n");

  //training accuracy
  CLabel ybar;  // prediction
  CMatVecArr *f = new CMatVecArr(ds.nExample, myinfo.k);  // the matrix Xw
  lossFunction->Predict(&w, &ds, &ybar, f);
  lossFunction->EvalPerformance(&w, &ds, &ybar, f);


  /* Finalize TAO */
  TaoFinalize();
  PetscFinalize();

  delete lossFunction;
  return 0;
}


int Grad(TAO_APPLICATION taoapp, Vec W, double *f, Vec G, void *ptr)
{
  static int iteration = 0;

  AppCtx *user = (AppCtx *) ptr;
  int info,nn=user->n/2;
  CDataset *ds = user->ds;
  CLoss *lossFunction = user->lossFunction;
  PetscScalar *x, *g;
  //double w_normSq = 0;
  

  /* Get pointers to vector data */
  info = VecGetArray(W,&x); CHKERRQ(info);
  info = VecGetArray(G,&g); CHKERRQ(info);


  // convert W to CDenseVec w
  //CDenseVec w(user->n);
  //w.Assign((double*)x, user->n);
  CMatVecArr w(user->n, 1, x);

  PetscInt size;
  VecGetSize(W, &size);

  // gradient vector
  CMatVecArr Grad(user->n, 1);
  //CDenseVec grad_i(user->n);
  //CDenseVec Grad(user->n);
  

  // loss value
  double Loss = 0;

  // compute loss value and loss gradient here
  Loss = 0;
  lossFunction->ComputeLossGradient(&w, ds, Loss, &Grad);
  iteration++;

  Loss = user->C * Loss;
  printf("[%4d]  Loss : %.12f\n", iteration, Loss);

  // compute objective value and objective gradient here
  //w_normSq = 0.5*w.NormSq();
  VecScale(Grad.vec, user->C);
  //Grad.Mult(user->C);

  VecAXPY(Grad.vec, user->lambda, w.vec);
  //Grad.Add(&w, user->lambda);      // norm-2 regularization term
  
  VecDot(w.vec, w.vec, f);
  *f = user->lambda*0.5*(*f) + Loss;
  //*f = user->lambda*0.5*w.NormSq() + Loss;
  
  
  memcpy(g, Grad.array, user->n*sizeof(double));
  //Grad.ToArray((double*)g, user->n);

  /* Restore vectors */
  info = VecRestoreArray(W,&x); CHKERRQ(info);
  info = VecRestoreArray(G,&g); CHKERRQ(info);

  info = PetscLogFlops(nn*15); CHKERRQ(info);
  return 0;
  
}

